// Update mixture properties
rho = fluid.rho();
U = fluid.U();

// Determine the phase density for buoyancy term
#include "pU/updatePhaseDensity.H"

// Store velocity and flux of previous iteration
forAll(fluid.movingPhases(), movingPhasei)
{
    fluid.movingPhases()[movingPhasei].U()().storePrevIter();
    fluid.movingPhases()[movingPhasei].phi()().storePrevIter();
}

// Update drag and virtual mass coefficient table and interfacial forces tables
fluid.momentumTransfer();

// Face volume fractions
forAll(phases, phasei)
{
    const phaseModel& phase = phases[phasei];
    const volScalarField& alpha = phase;

    alphafs.set(phasei, fvc::interpolate(alpha));
}

// Time derivative equation
#include "pU/ddtUEqn.H"

// Momentum transfer by mass transfer equation
#include "pU/dmdtUEqn.H"

// Convection, diffusion equation
#include "pU/HUEqn.H"
#include "pU/stabiliseHUEqn.H"

// fvModels for momentum equation
#include "pU/fvModelsEqn.H"

forAll(fluid.movingPhases(), movingPhasei)
{
    phaseModel& phase = fluid.movingPhases()[movingPhasei];

    HUEqns[phase.index()].relax(1.0);
    //ddtUEqns[phase.index()].relax(1.0);
    //dmdtUEqns[phase.index()].relax(1.0);
    //fvModelsEqns[phase.index()].relax(1.0);
}

// rHUs, rHUfs, rAs, rAfs
forAll(fluid.movingPhases(), movingPhasei)
{
    const phaseModel& phase = fluid.movingPhases()[movingPhasei];

    rHUs.set(phase.index(), 1.0/HUEqns[phase.index()].A());
    rHUfs.set(phase.index(), fvc::interpolate(rHUs[phase.index()]));

    // Diagonal coefficients for partial elimination
    rAs.set
    (
        phase.index(),
        1.0
       /(
            HUEqns[phase.index()].A()
          + ddtUEqns[phase.index()].A()
          + fvModelsEqns[phase.index()].A()
        )
    );

    // Face diagonal coefficients for partial elimination
    rAfs.set
    (
        phase.index(),
        1.0
       /(
            fvc::interpolate(HUEqns[phase.index()].A())
          + fvc::interpolate(ddtUEqns[phase.index()].A())
          + fvc::interpolate(fvModelsEqns[phase.index()].A())
        )
    );
}
fluid.fillFields("rHUs", dimless, rHUs);
fluid.fillFields("rHUfs", dimless, rHUfs);
fluid.fillFields("rAs", dimTime/dimDensity, rAs);
fluid.fillFields("rAfs", dimTime/dimDensity, rAfs);

// Implicit and pressure part of the drag
#include "pU/dragDiag.H"

// Implicit part of the time derivative of the virtual mass
#include "pU/vmDiag.H"

// Momentum predictor
if (pimple.momentumPredictor())
{
    // Correct p_rgh for consistency with p
    p_rgh = p - rho*gh;

    // Explicit part of the drag terms
    #include "pU/dragSource.H"

    // Explicit part of the time derivative of the virtual mass
    #include "pU/vmSource.H"

    // Add lift, wall lubrication and turbulent dispersion forces
    #include "pU/forcesSourcef.H"

    // Volumetric buoyancy contribution
    #include "pU/buoyancySource.H"

    // Mixture density gradient times gravitational acceleration
    const surfaceScalarField ghSnGradRho
    (
        "ghSnGradRho",
        ghf*fvc::snGrad(rho)*mesh.magSf()
    );

    // rAUfs, alpharAUfs, phigs WITHOUT field relaxation of UAlpha
    forAll(fluid.movingPhases(), movingPhasei)
    {
        phaseModel& phase = fluid.movingPhases()[movingPhasei];
        volVectorField& UAlpha = phase.URef();

        rAUfs.set
        (
            phase.index(),
            1.0
           /(
                1.0
              + rHUfs[phase.index()]
               *(
                    fvc::interpolate(ddtUEqns[phase.index()].A())
                  + fvc::interpolate(dmdtUEqns[phase.index()].A())
                  + fvc::interpolate(fvModelsEqns[phase.index()].A())
                  + dragDiagf[phase.index()]
                  + vmDiagf[phase.index()]
                )
            )
        );

        // Phase diagonal coefficients
        alpharAUfs.set
        (
            phase.index(),
            rHUfs[phase.index()]*rAUfs[phase.index()]
           *(alphafs[phase.index()] + dragPCoefff[phase.index()])
        );

        // Combined buoyancy and force fluxes
        phigs.set
        (
            phase.index(),
            alpharAUfs[phase.index()]
           *(
              - ghSnGradRho
              + fluid.surfaceTension(phase)*mesh.magSf()
            )
          - rHUfs[phase.index()]*rAUfs[phase.index()]*mesh.magSf()
           *phase.turbulentNormalStress()
        );

        // Momentum predictor equation
        fvVectorMatrix UEqns
        (
            ddtUEqns[phase.index()]
          + dmdtUEqns[phase.index()]
          + HUEqns[phase.index()]
          + fvm::Sp(dragDiag[phase.index()], UAlpha)
          + fvm::Sp(vmDiag[phase.index()], UAlpha)
          + fvModelsEqns[phase.index()]
            ==
            dragSource[phase.index()]
          + buoyancySource[phase.index()]
          + (
                ddtUEqns[phase.index()].A()
              + dmdtUEqns[phase.index()].A()
              + HUEqns[phase.index()].A()
              + dragDiag[phase.index()]
              + fvModelsEqns[phase.index()].A()
            )
           *fvc::reconstruct
            (
                phigs[phase.index()]
              - alpharAUfs[phase.index()]*fvc::snGrad(p_rgh)*mesh.magSf()
              + rHUfs[phase.index()]*rAUfs[phase.index()]
               *forcesSourcef[phase.index()]
            )
          + vmSource[phase.index()]
        );

        UEqns.relax();
        fvConstraints.constrain(UEqns);
        UEqns.solve();
        fvConstraints.constrain(UAlpha);
    }
}

// Construct pressure "diffusivity"
surfaceScalarField rAUf
(
    IOobject
    (
        "rAUf",
        runTime.timeName(),
        mesh
    ),
    mesh,
    dimensionedScalar(dimensionSet(-1, 3, 1, 0, 0), Zero)
);

// rAUfs, rHAUfs, alpharAUfs, rAUf WITH field relaxation of UAlpha
initZeroPtrList(rAUfs, mesh, dimless, "rAUfs");
initZeroPtrList(rHAUfs, mesh, dimTime/dimDensity, "rHAUfs");
initZeroPtrList(alpharAUfs, mesh, dimTime/dimDensity, "alpharAUfs");
forAll(fluid.movingPhases(), movingPhasei)
{
    const phaseModel& phase = fluid.movingPhases()[movingPhasei];

    rAUfs.set
    (
        phase.index(),
        1.0
       /(
            1.0/max(relaxU[phase.index()], small)
          + rHUfs[phase.index()]
           *(
                fvc::interpolate(ddtUEqns[phase.index()].A())
              + fvc::interpolate(dmdtUEqns[phase.index()].A())
              + fvc::interpolate(fvModelsEqns[phase.index()].A())
              + dragDiagf[phase.index()]
              + vmDiagf[phase.index()]
            )
        )
    );

    rHAUfs.set(phase.index(), rHUfs[phase.index()]*rAUfs[phase.index()]);

    // Phase diagonal coefficients
    alpharAUfs.set
    (
        phase.index(),
        rHUfs[phase.index()]*rAUfs[phase.index()]
       *(alphafs[phase.index()] + dragPCoefff[phase.index()])
    );

    // Pressure "diffusivity"
    rAUf += alphafs[phase.index()]*alpharAUfs[phase.index()];
}

rAUf = mag(rAUf);
