rho = fluid.rho();

// Correct p_rgh for consistency with p and the updated densities due to
// updated temperature
p_rgh = p - rho*gh;

// Determine the phase density for buoyancy term
#include "pU/updatePhaseDensity.H"

// Correct fixed-flux BCs to be consistent with the velocity BCs
fluid.correctBoundaryFlux();

{
    // Mixture density gradient times gravitational acceleration
    const surfaceScalarField ghSnGradRho
    (
        "ghSnGradRho",
        ghf*fvc::snGrad(rho)*mesh.magSf()
    );

    // phigs WITH field relaxation of UAlpha
    forAll(fluid.movingPhases(), movingPhasei)
    {
        const phaseModel& phase = fluid.movingPhases()[movingPhasei];

        // Combined buoyancy and force fluxes
        phigs.set
        (
            phase.index(),
            alpharAUfs[phase.index()]
           *(
              - ghSnGradRho
              + fluid.surfaceTension(phase)*mesh.magSf()
            )
          - rHUfs[phase.index()]*rAUfs[phase.index()]
           *phase.turbulentNormalStress()*mesh.magSf()
        );
    }
}

// Update explicit drag terms for PISO loop
#include "pU/dragSource.H"
#include "pU/dragSourcef.H"

// Explicit part of the time derivative of the virtual mass for PISO loop
#include "pU/vmSource.H"
#include "pU/vmSourcef.H"

// Update lift, wall lubrication and turbulent dispersion forces for PISO loop
#include "pU/forcesSourcef.H"

// Buoyancy for dispersed phases and parallel to underresolved interfaces
#include "pU/buoyancySource.H"
#include "pU/buoyancySourcef.H"

// Construct Total predicted flux
surfaceScalarField phiHbyA
(
    IOobject
    (
        "phiHbyA",
        runTime.timeName(),
        mesh,
        IOobject::NO_READ,
        IOobject::AUTO_WRITE
    ),
    mesh,
    dimensionedScalar(dimFlux, Zero)
);

forAll(fluid.movingPhases(), movingPhasei)
{
    const phaseModel& phase = fluid.movingPhases()[movingPhasei];
    const volVectorField& UAlpha = phase.U();

    // Pseudo convected velocities
    HbyAs.set
    (
        phase.index(),
        constrainHbyA
        (
            rHUs[phase.index()]*HUEqns[phase.index()].H(),
            UAlpha,
            p_rgh
        )
    );

    // Predicted fluxes for each phase
    phiHbyAs.set
    (
        phase.index(),
        constrainPhiHbyA
        (
            rAUfs[phase.index()]
           *(
                fvc::flux(HbyAs[phase.index()])
              + rHUfs[phase.index()]
               *(
                    ddtPhiCoeffs[phase.index()]
                  + dmdtPhiCoeffs[phase.index()]
                  + buoyancySourcef[phase.index()]
                  + forcesSourcef[phase.index()]
                  + fvc::flux(fvModelsEqns[phase.index()].H())
                )
              + (1 - relaxU[phase.index()])/max(relaxU[phase.index()], small)
               *MRF.absolute(phases[phase.index()].phi()().prevIter())
              + dragSourcef[phase.index()]
              + vmSourcef[phase.index()]
            ),
            UAlpha,
            p_rgh
        )
      + phigs[phase.index()]
    );

    phiHbyA += alphafs[phase.index()]*phiHbyAs[phase.index()];
}

MRF.makeRelative(phiHbyA);

// Update the fixedFluxPressure BCs to ensure flux consistency
{
    surfaceScalarField::Boundary phib
    (
        surfaceScalarField::Internal::null(),
        phi.boundaryField()
    );
    phib = 0;

    forAll(phases, phasei)
    {
        const phaseModel& phase = phases[phasei];
        phib += alphafs[phasei].boundaryField()*phase.phi()().boundaryField();
    }

    setSnGrad<fixedFluxPressureFvPatchScalarField>
    (
        p_rgh.boundaryFieldRef(),
        (
            phiHbyA.boundaryField() - phib
        )/(mesh.magSf().boundaryField()*rAUf.boundaryField())
    );
}

// Compressible pressure equations
#include "pEqnComps.H"

// Cache p prior to solve for density update
volScalarField p_rgh_0(p_rgh);

// Iterate over the pressure equation to correct for non-orthogonality
while (pimple.correctNonOrthogonal())
{
    // Construct the transport part of the pressure equation
    fvScalarMatrix pEqnIncomp
    (
        fvc::div(phiHbyA)
      - fvm::laplacian(rAUf, p_rgh)
    );

    // Solve
    {
        fvScalarMatrix pEqn(pEqnIncomp);

        forAll(phases, phasei)
        {
            pEqn += pEqnComps[phasei];
        }

        if (fluid.incompressible())
        {
                pEqn.setReference
                (
                    pressureReference.refCell(),
                    pressureReference.refValue()
                );
        }

        pEqn.solve();
    }

    // Correct fluxes and velocities on last non-orthogonal iteration
    if (pimple.finalNonOrthogonalIter())
    {
        phi = phiHbyA + pEqnIncomp.flux();

        surfaceScalarField mSfGradp("mSfGradp", pEqnIncomp.flux()/rAUf);

        forAll(fluid.movingPhases(), movingPhasei)
        {
            phaseModel& phase = fluid.movingPhases()[movingPhasei];

            phase.phiRef() =
                phiHbyAs[phase.index()]
              + alpharAUfs[phase.index()]*mSfGradp;

            // Set the phase dilatation rate
            phase.divU(-pEqnComps[phase.index()] & p_rgh);
        }

        // Optionally relax pressure for velocity correction
        p_rgh.relax();

        mSfGradp = pEqnIncomp.flux()/rAUf;

        // Momentum corrector
        forAll(fluid.movingPhases(), movingPhasei)
        {
            phaseModel& phase = fluid.movingPhases()[movingPhasei];
            volVectorField& UAlpha = phase.URef();

            if (faceMomentum)
            {
                UAlpha = fvc::reconstruct(phase.phi());
            }
            else
            {
                UAlpha =
                    1.0
                   /(
                        ddtUEqns[phase.index()].A()
                      + dmdtUEqns[phase.index()].A()
                      + HUEqns[phase.index()].A()
                       /max(relaxU[phase.index()], small)
                      + dragDiag[phase.index()]
                      + vmDiag[phase.index()]
                      + fvModelsEqns[phase.index()].A()
                    )
                   *(
                        HUEqns[phase.index()].A()*HbyAs[phase.index()]
                      + ddtUEqns[phase.index()].H()
                      + dmdtUEqns[phase.index()].H()
                      + dragSource[phase.index()]
                      + buoyancySource[phase.index()]
                      + (1 - relaxU[phase.index()])
                       /max(relaxU[phase.index()], small)
                       *HUEqns[phase.index()].A()*UAlpha.prevIter()
                      + vmSource[phase.index()]
                      + fvModelsEqns[phase.index()].H()
                    )
                  + fvc::reconstruct
                    (
                        phigs[phase.index()]
                      + alpharAUfs[phase.index()]*mSfGradp
                      + rHUfs[phase.index()]*rAUfs[phase.index()]
                       *forcesSourcef[phase.index()]
                    );
            }

            UAlpha.correctBoundaryConditions();
            fvConstraints.constrain(UAlpha);

            MRF.makeRelative(phase.phiRef());
        }
    }
}

// Update the mixture properties and static pressure
U = fluid.U();
p = p_rgh + rho*gh;
fvConstraints.constrain(p);

// Account for static pressure reference
if (p_rgh.needReference() && fluid.incompressible())
{
    p += dimensionedScalar
    (
        "p",
        p.dimensions(),
        pressureReference.refValue()
      - getRefCellValue(p, pressureReference.refCell())
    );
}

// Limit p_rgh
p_rgh = p - rho*gh;

// Update densities from change in p_rgh
forAll(phases, phasei)
{
    phaseModel& phase = phases[phasei];
    phase.thermoRef().rho() += phase.thermo().psi()*(p_rgh - p_rgh_0);
}

// Correct p_rgh for consistency with p and the updated densities
rho = fluid.rho();
p_rgh = p - rho*gh;
p_rgh.correctBoundaryConditions();
