/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2020-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

Class
    Foam::fv::degassingPhases

Description
    This fvModel mimics a degassing boundary condition, which includes the
    following sources/sinks applied in the cells next to the boundary marked by
    patchName:

    - All moving phases, that are not marked as continuous phases are considered
      as degassing phases and get a sink. The rate is calculated such that the
      whole phase fraction is removed within one time step. Along with the mass
      sink, there are sinks defined for all transported variables accordingly.
    - Exactly one phase has to be marked as continuous phase in phaseProperties
      and gets a sink/source that is calculated, such that the sink of the
      degassing phases (positive rate) and the volume flux of the degassing
      phases across the boundaries is also counteracted. Hence, this phase will
      give way for the degassing phases injected into the domain.

    The boundary patch, to which the degassing option is applied to, has to be
    a wall with slip for the velocity field for the liquid and gas. Pressure
    should be specified by pRefValue.

    Note: This fvModel is developed and tested for incompressible fluid only.

Usage
    \table
        Property | Description | Required | Default value
        excludePhases | Names of phases excluded from degassing BC | no | empty
        alphaName | Name of the phase fraction/number density field | no | alpha
        patchNames | Names of outlet boundary patches | yes
    \endtable

    Example specification:
    \verbatim
    <fvModelName>
    {
        type    degassingPhases;

        libs    ("libHZDRmultiphaseEulerFoamFvModels.so");

        excludePhases   (particles);
        patchNames      (outlet);
    }
    \endverbatim

SourceFiles
    degassingPhases.C

\*---------------------------------------------------------------------------*/

#ifndef degassingPhases_H
#define degassingPhases_H

#include "fvModel.H"
#include "fvCellSet.H"
#include "surfaceFields.H"
#include "volFields.H"
#include "phaseSystem.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{
namespace fv
{

/*---------------------------------------------------------------------------*\
                       Class degassingPhases Declaration
\*---------------------------------------------------------------------------*/

class degassingPhases
:
    public fvModel
{
    // Private Member Data

        //- Names of degassing phases
        wordList degassingPhases_;

        //- Names of excluded phases
        wordList excludePhases_;

        //- Name of the phase that gives way
        word continuousPhase_;

        //- Name of phase fraction/number density fields
        word alphaName_;

        //- Constant access to phaseSystem
        const phaseSystem& fluid_;

        //- Names of alpha fields
        wordList fieldNames_;

        //- Names of outlet boundary patches
        wordList patchNames_;

        //- Set of cells to apply source to
        labelList cells_;

        //- Sum of cell volumes
        scalar V_;


    // Private Member Functions

        //- Check if the phase is a degassing phase
        bool isDegassingPhase(const word& psiName) const;

        //- Calculate volume of cells attached to patches
        scalar patchCellVolume(const wordList& patches) const;

        //- Add source terms to a phase equation with constant density
        template<class Type>
        void addSupType
        (
            const volScalarField& alpha,
            fvMatrix<Type>& eqn,
            const word& fieldName
        ) const;

        //- Add source terms to a phase equation
        template<class Type>
        void addSupType
        (
            const volScalarField& alpha,
            const volScalarField& rho,
            fvMatrix<Type>& eqn,
            const word& fieldName
        ) const;


public:

    //- Runtime type information
    TypeName("degassingPhases");


    // Constructors

        //- Construct from components
        degassingPhases
        (
            const word& name,
            const word& modelType,
            const dictionary& dict,
            const fvMesh& mesh
        );

        //- Disallow default bitwise copy construction
        degassingPhases(const degassingPhases&) = delete;


    //- Destructor
    virtual ~degassingPhases();


    // Member Functions

        // Checks

            //- Return true if the option adds a source term to the given
            //  field's transport equation
            virtual bool addsSupToField(const word& fieldName) const;

            //- Return the list of fields for which the option adds source term
            //  to the transport equation
            virtual wordList addSupFields() const;


        // Sources

            //- Add a source term to a phase equation with constant density
            FOR_ALL_FIELD_TYPES(DEFINE_FV_MODEL_ADD_RHO_SUP);

            //- Add a source term to a phase equation
            FOR_ALL_FIELD_TYPES(DEFINE_FV_MODEL_ADD_ALPHA_RHO_SUP);

        // Mesh motion

            //- Update topology using the given map
            virtual void topoChange(const polyTopoChangeMap&);

            //- Update from another mesh using the given map
            virtual void mapMesh(const polyMeshMap&);

            //- Redistribute or update using the given distribution map
            virtual void distribute(const polyDistributionMap&);

            //- Update for mesh motion
            virtual bool movePoints();


    // Member Operators

        //- Disallow default bitwise assignment
        void operator=(const degassingPhases&) = delete;
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace fv
} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
