/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2021-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "displacementFreeSurfaceMotionSolver.H"
#include "addToRunTimeSelectionTable.H"
#include "surfaceInterpolate.H"
#include "fvModels.H"
#include "cyclicAMIFvPatch.H"
#include "fvMatrix.H"
#include "phaseSystem.H"
#include "pimpleControl.H"
#include "fvcMeshPhi.H"

// * * * * * * * * * * * * * * Static Data Members * * * * * * * * * * * * * //

namespace Foam
{
    defineTypeNameAndDebug(displacementFreeSurfaceMotionSolver, 0);

    addToRunTimeSelectionTable
    (
        motionSolver,
        displacementFreeSurfaceMotionSolver,
        dictionary
    );
}


// * * * * * * * * * * * * * Private Member Functions  * * * * * * * * * * * //

void
Foam::displacementFreeSurfaceMotionSolver::writeDisplacementProperties() const
{
    IOdictionary propsDict
    (
        IOobject
        (
            typeName + "Properties",
            mesh().time().timeName(),
            "uniform",
            mesh(),
            IOobject::NO_READ,
            IOobject::NO_WRITE,
            false
        )
    );

    propsDict.add("displacement", displacement_);
    propsDict.regIOobject::write();
}


void Foam::displacementFreeSurfaceMotionSolver::readDisplacementProperties()
{
    IOobject propsDictHeader
    (
        typeName + "Properties",
        mesh().time().timeName(mesh().time().startTime().value()),
        "uniform",
        mesh(),
        IOobject::MUST_READ_IF_MODIFIED,
        IOobject::NO_WRITE,
        false
    );

    if (!propsDictHeader.headerOk())
    {
        Info<< "    Starting mesh motion at time "
            << mesh().time().timeName() << endl;

        return;
    }

    IOdictionary propsDict(propsDictHeader);

    displacement_ = propsDict.lookup<scalar>("displacement");

    Info<< "    Restarting mesh motion with displacement: "
        << displacement_ << endl;
}


// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

Foam::displacementFreeSurfaceMotionSolver::displacementFreeSurfaceMotionSolver
(
    const word& name,
    const polyMesh& mesh,
    const dictionary& dict
)
:
    points0MotionSolver(name, mesh, dict, typeName),
    axis_(normalised(vector(coeffDict().lookup("axis")))),
    xFixed_(coeffDict().lookup<scalar>("xFixed")),
    xMoving_(coeffDict().lookup<scalar>("xMoving")),
    displacement_(Zero),
    patchName_(dict.lookup("patchName"))
{
    readDisplacementProperties();
}


// * * * * * * * * * * * * * * * * Destructor  * * * * * * * * * * * * * * * //

Foam::displacementFreeSurfaceMotionSolver::
~displacementFreeSurfaceMotionSolver()
{}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

Foam::tmp<Foam::pointField>
Foam::displacementFreeSurfaceMotionSolver::curPoints() const
{
    const pimpleControl& pimple =
        mesh().lookupObject<pimpleControl>("solutionControl");

    if (pimple.firstPimpleIter())
    {
        displacement0_ = displacement_;
    }

    // lookup the fluid model and the phase
    const phaseSystem& fluid =
        mesh().lookupObject<phaseSystem>(phaseSystem::propertiesName);

    tmp<pointField> tcurPoints(new pointField(points0()));
    pointField& curPoints = tcurPoints.ref();
    const scalar deltaT = mesh().time().deltaT0().value();

    scalar volumeRate(0.0);
    scalar boundaryRate(0.0);

    forAll(fluid.movingPhases(), movingPhasei)
    {
        const phaseModel& phase = fluid.movingPhases()[movingPhasei];
        const volScalarField& alpha = phase;
        const surfaceScalarField& phi = phase.phi();
        const volScalarField& rhoAlpha = phase.rho();

        forAll(mesh().boundaryMesh(), patchi)
        {

            if
            (
                mesh().boundaryMesh()[patchi].coupled()
             || isA<cyclicAMIFvPatch>(mesh().boundaryMesh()[patchi])
             || mesh().boundaryMesh()[patchi].name() == "defaultFaces"
            )
            {
                continue;
            }

            const fvsPatchScalarField& phip = phi.boundaryField()[patchi];
            const fvPatchScalarField& alphap = alpha.boundaryField()[patchi];

            boundaryRate -= sum(alphap*phip);
        }

        const fvModels& fvModels = fluid.fvModels();

        if (fvModels.addsSupToField(rhoAlpha.name()))
        {
            const volScalarField source
            (
                fvModels.source(alpha, rhoAlpha) & rhoAlpha
            );

            volumeRate += sum(source*mesh().cellVolumes()/rhoAlpha);
        }
    }

    reduce(boundaryRate, sumOp<scalar>());
    reduce(volumeRate, sumOp<scalar>());

    const scalar patchArea =
        gSum(mesh().boundaryMesh()[patchName_].magFaceAreas());

    displacement_ =
        displacement0_ + deltaT*(boundaryRate + volumeRate)/patchArea;

    if (debug)
    {
        Pout<< "Free surface adjustment enabled at patch "
            << patchName_ << nl
            << "    boundaryRate: " << boundaryRate << nl
            << "    volumeRate  : " << volumeRate << nl
            << "    displacement: " << displacement_ << endl;
    }

    forAll(curPoints, i)
    {
        const scalar lambda
        (
            (xFixed_ - (axis_ & curPoints[i]))/(xFixed_ - xMoving_)
        );

        if (lambda > 1)
        {
            curPoints[i] += axis_*displacement_;
        }
        else if (lambda > 0)
        {
            curPoints[i] += axis_*lambda*displacement_;
        }
    }

    return tcurPoints;
}


bool Foam::displacementFreeSurfaceMotionSolver::write() const
{
    writeDisplacementProperties();

    return true;
}

// ************************************************************************* //
