/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2021-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

Class
    Foam::displacementFreeSurfaceMotionSolver

Description
    Mesh motion solver that mimics a rising/lowering free surface, e.g., in a
    bubble column based on the net volume flux of gass across the domain
    boundaries and via mass sinks/sources.

    Example:
    \verbatim
        dynamicFvMesh   dynamicMotionSolverFvMesh;

        motionSolver    displacementFreeSurfaceMotion;

        axis            (0 1 0);
        xFixed          0.8;
        xMoving         0;
        patchName       "outlet";

    \endverbatim

    This mesh moves between \c xFixed and \c xMoving in the direction
    \c axis. The \c patchName should be the patch considered as free surface
    and is required to determine the patch area. The \c axis should be set
    along with the gravitational axis and has to be normal to the patch
    considered as free surface.

SourceFiles
    displacementFreeSurfaceMotionSolver.C

\*---------------------------------------------------------------------------*/

#ifndef displacementFreeSurfaceMotionSolver_H
#define displacementFreeSurfaceMotionSolver_H

#include "points0MotionSolver.H"
#include "surfaceInterpolate.H"
#include "phaseSystem.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

/*---------------------------------------------------------------------------*\
            Class displacementFreeSurfaceMotionSolver Declaration
\*---------------------------------------------------------------------------*/

class displacementFreeSurfaceMotionSolver
:
    public points0MotionSolver
{
    // Private Data

        //- Axis of the linear motion
        const vector axis_;

        //- The location of the end of motion
        const scalar xFixed_;

        //- The location moving with the specified displacement
        const scalar xMoving_;

        //- The displacement as a function of time
        mutable scalar displacement_;

        //- displacement base value
        mutable scalar displacement0_;

        //- Name of the free surface patch
        const word patchName_;


    // Private Member Functions

        //- Write displacement properties
        void writeDisplacementProperties() const;

        //- Read displacement properties
        void readDisplacementProperties();


public:

    //- Runtime type information
    TypeName("displacementFreeSurfaceMotion");


    // Constructors

        //- Construct from polyMesh and dictionary
        displacementFreeSurfaceMotionSolver
        (
            const word& name,
            const polyMesh&,
            const dictionary&
        );

        //- Disallow default bitwise copy construction
        displacementFreeSurfaceMotionSolver
        (
            const displacementFreeSurfaceMotionSolver&
        );


    //- Destructor
    ~displacementFreeSurfaceMotionSolver();


    // Member Functions

        //- Return point location obtained from the current motion field
        virtual tmp<pointField> curPoints() const;

        //- Solve for motion
        virtual void solve()
        {}

        //- Optionally write motion state information for restart
        virtual bool write() const;


    // Member Operators

        //- Disallow default bitwise assignment
        void operator=(const displacementFreeSurfaceMotionSolver&) = delete;
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
