/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2020-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "oneDimensionalDelta.H"
#include "addToRunTimeSelectionTable.H"

// * * * * * * * * * * * * * * Static Data Members * * * * * * * * * * * * * //

namespace Foam
{
namespace SGSModels
{
    defineTypeNameAndDebug(oneDimensionalDelta, 0);
    addToRunTimeSelectionTable(SGSdelta, oneDimensionalDelta, dictionary);
}
}


// * * * * * * * * * * * * * Private Member Functions  * * * * * * * * * * * //

void Foam::SGSModels::oneDimensionalDelta::calcDelta()
{
    const label nD = mesh_.nGeometricD();

    if (nD != 1)
    {
        FatalErrorInFunction
            << "Case is not 1D, this 'delta' method is not applicable"
            << exit(FatalError);
    }

    const Vector<label>& directions = mesh_.geometricD();

    // Get non-empty direction
    vector basevector(0, 0, 0);
    for (direction dir=0; dir<directions.nComponents; dir++)
    {
        if (directions[dir] == 1)
        {
            basevector[dir] = 1;
            break;
        }
    }

    const cellList& cells = mesh_.cells();
    scalarField hmax(cells.size());

    forAll(cells,celli)
    {
        scalar minCoord = great;
        scalar maxCoord = 0.0;
        const labelList& cFaces = mesh_.cells()[celli];

        forAll(cFaces, cFacei)
        {
            label facei = cFaces[cFacei];
            const point& facevector = mesh_.faceCentres()[facei];
            const scalar tmp = facevector & basevector;
            if (tmp > maxCoord)
            {
                maxCoord = tmp;
            }

            if (tmp < minCoord)
            {
                minCoord = tmp;
            }
        }

        hmax[celli] = maxCoord - minCoord;
    }

    delta_.primitiveFieldRef() = hmax;
}


// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

Foam::SGSModels::oneDimensionalDelta::oneDimensionalDelta
(
    const word& name,
    const fvMesh& mesh,
    const dictionary& dict
)
:
    SGSdelta(name, mesh)
{
    calcDelta();
}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

void Foam::SGSModels::oneDimensionalDelta::correct()
{
    if (mesh_.changing())
    {
        calcDelta();
    }
}


// ************************************************************************* //
