/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2019-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

Class
    Foam::subGridSurfaceTensionModel::ReynoldsDependentWeberNumber

Description
    Sub grid surface tension with effective Weber number approach according to
    Ketterl et al. (2019).

    Reference:
    \verbatim
        Ketterl, S., Reißmann, M., Klein, M. (2019).
        Large eddy simulation of multiphase flows using the volume of fluid
        method: Part 2 - A-posteriori analysis of liquid jet atomization.
        Experimental and Computational Multiphase Flow, 1(3), 201-211.
    \endverbatim

    Instead of prescribing a constant value for the critical breakup Weber
    number, the latter is evaluated depending on the local grid Reynolds number
    according to Rysik and Leal (1984).

    \f[
        \left(\frac{1}{\mathit{We}_\mathrm{crit}}\right)^\frac{10}{9}
      =
        \left(\frac{1}{2.76}\right)^\frac{10}{9}
      + \left(
            \frac{1}{0.247 \mathit{Re}_\mathrm{local}^\frac{3}{4}}
        \right)^\frac{10}{9}
    \f]

    The above dependency can be partially evaluated and transposed:

    \f[
        \mathit{We}_\mathrm{crit}
      = \left(
            0.32366893582752687
          + 4.729129024951329 \mathit{Re}_\mathrm{local}^{-\frac{5}{6}}
        \right)^{-0.9}.
    \f]

    Reference:
    \verbatim
        Ryskin, G., Leal, L. G. (1984).
        Numerical solution of free-boundary problems in fluid mechanics. Part 3.
        Bubble deformation in an axisymmetric straining flow.
        Journal of Fluid Mechanics, 148, 37-43.
    \endverbatim

SourceFiles
    ReynoldsDependentWeberNumber.C

\*---------------------------------------------------------------------------*/

#ifndef ReynoldsDependentWeberNumber_H
#define ReynoldsDependentWeberNumber_H

#include "WeberNumberModel.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

namespace subGridSurfaceTensionModels
{

/*---------------------------------------------------------------------------*\
                Class ReynoldsDependentWeberNumber Declaration
\*---------------------------------------------------------------------------*/

class ReynoldsDependentWeberNumber
:
    public WeberNumberModel
{
    // Private Data

        //- Residual Reynolds number
        const dimensionedScalar residualRe_;


    // Private Member Functions

        //- Kinematic viscosity of an interface
        tmp<volScalarField> nu() const;

        //- Local Reynolds number
        tmp<volScalarField> ReLocal() const;

        //- Critical Weber number
        tmp<volScalarField> WeCrit() const;


public:

    //- Runtime type information
    TypeName("ReynoldsDependentWeberNumber");


    // Constructors

        //- Construct from a dictionary and an interface
        ReynoldsDependentWeberNumber
        (
            const dictionary& dict,
            const phaseInterface& interface
        );


    //- Destructor
    virtual ~ReynoldsDependentWeberNumber();
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace subGridSurfaceTensionModels
} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
