/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2019-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "WeberNumberModel.H"

// * * * * * * * * * * * * * Private Member Functions  * * * * * * * * * * * //

Foam::tmp<Foam::volScalarField>
Foam::subGridSurfaceTensionModels::WeberNumberModel::sigmaEff() const
{
    return interface_.sigma() + sigmat();
}


Foam::tmp<Foam::volVectorField>
Foam::subGridSurfaceTensionModels::WeberNumberModel::subGridSurfaceTensionForce
() const
{
    return sigmat()*interface_.K()*interface_.gradAlpha();
}


// * * * * * * * * * * * * Protected Member Functions  * * * * * * * * * * * //

Foam::tmp<Foam::volScalarField>
Foam::subGridSurfaceTensionModels::WeberNumberModel::sigmat() const
{
    tmp<volScalarField> sigmaCrit
    (
        interface_.rho()*magSqr(U())*sqr(delta())*mag(interface_.K())/WeCrit()
    );

    return
        max
        (
            sigmaCrit - interface_.sigma(),
            dimensionedScalar(interface_.sigma()().dimensions(), Zero)
        );
}


Foam::tmp<Foam::surfaceScalarField>
Foam::subGridSurfaceTensionModels::WeberNumberModel::sigmatf() const
{
    tmp<surfaceScalarField> sigmaCritf
    (
        fvc::interpolate(interface_.rho())*fvc::interpolate(magSqr(U()))
       *magSqr(interface_.mesh().delta())*fvc::interpolate(mag(interface_.K()))
       /fvc::interpolate(WeCrit())
    );

    return
        max
        (
            sigmaCritf - fvc::interpolate(interface_.sigma()),
            dimensionedScalar(interface_.sigma()().dimensions(), Zero)
        );
}


Foam::tmp<Foam::volVectorField>
Foam::subGridSurfaceTensionModels::WeberNumberModel::U() const
{
    const phaseModel& phase1 = interface_.phase1();
    const phaseModel& phase2 = interface_.phase2();

    return
        phase1*phase1.U() + phase2*phase2.U()
       /max
        (
            phase1 + phase2,
            phase1.residualAlpha() + phase2.residualAlpha()
        );
}


bool Foam::subGridSurfaceTensionModels::WeberNumberModel::write() const
{
    if (subGridSurfaceTensionModel::write())
    {
        tmp<volScalarField> sigmaEffective
        (
            volScalarField::New
            (
                IOobject::groupName("sigmaEff", interface_.name()),
                sigmaEff()
            )
        );
        sigmaEffective.ref().write();

        return true;
    }
    else
    {
        return false;
    }
}


// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

Foam::subGridSurfaceTensionModels::WeberNumberModel::WeberNumberModel
(
    const dictionary& dict,
    const phaseInterface& interface
)
:
    subGridSurfaceTensionModel(dict, interface),

    delta_
    (
        SGSdelta::New
        (
            IOobject::groupName("delta", interface.name()),
            interface.mesh(),
            dict
        )
    )
{}


// * * * * * * * * * * * * * * * * Destructor  * * * * * * * * * * * * * * * //

Foam::subGridSurfaceTensionModels::WeberNumberModel::~WeberNumberModel()
{}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

Foam::tmp<Foam::surfaceScalarField>
Foam::subGridSurfaceTensionModels::WeberNumberModel::subGridSurfaceTensionForcef
() const
{
    if (debug && interface_.mesh().time().outputTime()) write();

    return sigmatf()*fvc::interpolate(interface_.K())*interface_.gradAlphaf();
}


// ************************************************************************* //
