/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2018-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

Class
    Foam::clusteringByMomentumEqnModel

Description
    Baseclass for modeling the formation of a resolved structure from a smeared
    phase fraction field by an additional force in the momentum equations, the
    clustering force.

See also
    Foam::MomentumTransferPhaseSystem
    Foam::TransitionPhaseSystem

SourceFiles
    clusteringByMomentumEqnModel.C
    clusteringByMomentumEqnModelNew.C

\*---------------------------------------------------------------------------*/

#ifndef clusteringByMomentumEqnModel_H
#define clusteringByMomentumEqnModel_H

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#include "volFields.H"
#include "dictionary.H"
#include "dispersedPhaseInterface.H"
#include "runTimeSelectionTables.H"
#include "BlendedInterfacialModel.H"

namespace Foam
{

/*---------------------------------------------------------------------------*\
                Class clusteringByMomentumEqnModel Declaration
\*---------------------------------------------------------------------------*/

class clusteringByMomentumEqnModel
{
protected:

    // Protected Data

        //- Phase interface
        const dispersedPhaseInterface interface_;


    // Protected Member Functions

        //- Return clustering coefficient
        virtual tmp<volScalarField> C() const = 0;


public:

    //- Runtime type information
    TypeName("clusteringByMomentumEqnModel");


    // Declare runtime construction

        declareRunTimeSelectionTable
        (
            autoPtr,
            clusteringByMomentumEqnModel,
            dictionary,
            (
                const dictionary& dict,
                const phaseInterface& interface
            ),
            (dict, interface)
        );


    // Static data members

        //- Force dimensions
        static const dimensionSet dimF;

        //- Does this model require correcting on fixed flux boundaries?
        static const bool correctFixedFluxBCs = true;


    // Constructors

        //- Construct from a dictionary and an interface
        clusteringByMomentumEqnModel
        (
            const dictionary& dict,
            const phaseInterface& interface
        );


    //- Destructor
    virtual ~clusteringByMomentumEqnModel();


    // Selectors

        static autoPtr<clusteringByMomentumEqnModel> New
        (
            const dictionary& dict,
            const phaseInterface& interface,
            const bool outer=true
        );


    // Member Functions

        //- Return phase-intensive clustering force
        virtual tmp<volVectorField> Fi() const;

        //- Return clustering force
        virtual tmp<volVectorField> F() const;

        //- Return face clustering force
        virtual tmp<surfaceScalarField> Ff() const;
};


/*---------------------------------------------------------------------------*\
             Class blendedClusteringByMomentumEqnModel Declaration
\*---------------------------------------------------------------------------*/

class blendedClusteringByMomentumEqnModel
:
    public BlendedInterfacialModel<clusteringByMomentumEqnModel>
{
public:

    // Constructors

        //- Inherit base class constructors
        using
            BlendedInterfacialModel<clusteringByMomentumEqnModel>
          ::BlendedInterfacialModel;


    // Selectors

        static autoPtr<blendedClusteringByMomentumEqnModel> New
        (
            const dictionary& dict,
            const phaseInterface& interface
        );


    // Member Functions

        //- Return clustering force
        tmp<volVectorField> F() const;

        //- Return face clustering force
        tmp<surfaceScalarField> Ff() const;
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
