/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2016-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

Class
    Foam::dragModels::IshiiMishima

Description
    Drag model based on a constant drag coefficient and the interfacial area
    density.

    Interfacial momentum transfer due to drag
    \f[
        \vec{M}_\alpha = K\,\left(\vec{U}_1-\vec{U}_2\right)
    \f]

    Dragfunction K
    \f[
        K = c_d\rho_{mix}A_{i}\left|\vec{U}_1-\vec{U}_2\right|
    \f]
    with
    \f[
        \rho_{mix} = \alpha_1\rho_1+\alpha_2\rho_2
    \f]
    Interfacial area density for dispersed fluids
    \f[
        A_{i} = \frac{6\alpha_{disp}}{d_{disp}}
    \f]
    or interfacial area density for free surface flows
    \f[
        A_{i} = \frac{2\left|\nabla\alpha_1\right|
            \left|\nabla\alpha_2\right|}
            {\left|\nabla\alpha_1\right|+\left|\nabla\alpha_2\right|}
    \f]
    The interfacial area density is clipped to be not smaller than
    \f$1/L_{max}\f$.

Usage
    \table
        Parameter   | Name  | Description       | Default value
        \f$c_d\f$    | Cd   | drag coefficient  | -
        \f$U_{r,min}\f$    | residualSlip | minimum slip velocity  | -
        \f$L_{max}\f$| LMax | maximum length scale for area density  | 1 m
    \endtable

SourceFiles
    IshiiMishima.C

\*---------------------------------------------------------------------------*/

#ifndef IshiiMishima_H
#define IshiiMishima_H

#include "dragModel.H"
#include "segregatedPhaseInterface.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{
namespace dragModels
{

/*---------------------------------------------------------------------------*\
                        Class IshiiMishima Declaration
\*---------------------------------------------------------------------------*/

class IshiiMishima
:
    public dragModel
{
    // Private Data

        //- Interface
        const segregatedPhaseInterface interface_;

        //- Swarm correction
        autoPtr<swarmCorrection> swarmCorrection_;

        //- Drag coefficient
        const dimensionedScalar Cd_;

        //- Length scale to calculate minimum interfacial area density
        const dimensionedScalar LMax_;

        //- Residual slip velocity to avoid division by zero
        const dimensionedScalar residualSlip_;

public:

    //- Runtime type information
    TypeName("IshiiMishima");


    // Constructors

        //- Construct from a dictionary and an interface
        IshiiMishima
        (
            const dictionary& dict,
            const phaseInterface& interface,
            const bool registerObject
        );


    //- Destructor
    virtual ~IshiiMishima();


    // Member Functions

        //- Return the drag coefficient K
        //  used in the momentum equations
        //    ddt(alpha1*rho1*U1) + ... = ... K*(U1-U2)
        //    ddt(alpha2*rho2*U2) + ... = ... K*(U2-U1)
        virtual tmp<volScalarField> K() const;

        //- Return the drag coefficient Kf
        //  used in the face-momentum equations
        virtual tmp<surfaceScalarField> Kf() const;
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace dragModels
} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
