/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2018-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "PorombkaHoehne.H"
#include "fvcGrad.H"
#include "addToRunTimeSelectionTable.H"

// * * * * * * * * * * * * * * Static Data Members * * * * * * * * * * * * * //

namespace Foam
{
namespace dragModels
{
    defineTypeNameAndDebug(PorombkaHoehne, 0);
    addToRunTimeSelectionTable(dragModel, PorombkaHoehne, dictionary);
}
}


// * * * * * * * * * * * * * Private Member Functions  * * * * * * * * * * * //

Foam::tmp<Foam::volSymmTensorField> Foam::dragModels::PorombkaHoehne::tauVisc
(
    const phaseModel& phase
) const
{
    return 2*phase.thermo().mu()*dev(symm(fvc::grad(phase.U())));
}


// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

Foam::dragModels::PorombkaHoehne::PorombkaHoehne
(
    const dictionary& dict,
    const phaseInterface& interface,
    const bool registerObject
)
:
    dragModel(dict, interface, registerObject),
    interface_(interface.phase1(), interface.phase2()),
    LMax_(dimensionedScalar::lookupOrDefault("LMax", dict, dimLength, 1.0)),
    residualSlip_("residualSlip", dimVelocity, dict)
{}


// * * * * * * * * * * * * * * * * Destructor  * * * * * * * * * * * * * * * //

Foam::dragModels::PorombkaHoehne::~PorombkaHoehne()
{}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

Foam::tmp<Foam::volScalarField> Foam::dragModels::PorombkaHoehne::K() const
{
    // phase normal vector
    const volVectorField alphaN(-interface_.nHat());

    // phase shear stress vectors
    const volVectorField tAlpha1(tauVisc(interface_.phase1()) & alphaN);
    const volVectorField tAlpha2(tauVisc(interface_.phase2()) & -alphaN);

    // phase shear stress parallel to the free surface
    const volScalarField magTAlpha1Fs(mag(tAlpha1 - (alphaN & tAlpha1)*alphaN));
    const volScalarField magTAlpha2Fs(mag(tAlpha2 - (alphaN & tAlpha2)*alphaN));

    return
        max
        (
            max
            (
                0.01*interface_.rho()*sqr(interface_.magUr()),
                interface_.phase1()*magTAlpha1Fs
              + interface_.phase2()*magTAlpha2Fs
            ),
            dimensionedScalar(dimensionSet(1, -1, -2, 0, 0), small)
        )
       *max(interface_.Ai(), 1/LMax_)
       /max(interface_.magUr(), residualSlip_);
}


Foam::tmp<Foam::surfaceScalarField>
Foam::dragModels::PorombkaHoehne::Kf() const
{
    return fvc::interpolate(K());
}


// ************************************************************************* //
