/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2018-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

Class
    Foam::dragModels::PorombkaHoehne

Description

    Reference:
    \verbatim
        Porombka, P., & Höhne, T. (2015).
        Drag and turbulence modelling for free surface flows within the
        two-fluid Euler–Euler framework.
        Chemical Engineering Science, 134, 348-359.
    \endverbatim

    Interfacial momentum transfer due to drag
    \f[
        \vec{M}_\alpha = K\,\left(\vec{U}_1-\vec{U}_2\right)
    \f]

    Dragfunction K
    \f[
        K = max\left(0.01\rho_{mix}A_{i}\left|\vec{U}_\alpha-\vec{U}_2\right|,
            \frac{\left(\alpha_1\left|\vec{t}_{fs,1}\right|+
            \alpha_2\left|\vec{t}_{fs,2}\right|\right)a_{12}}
            {\left|\vec{U}_1-\vec{U}_2\right|}\right)
    \f]
    with
    \f[
        \vec{t}_{fs,\varphi} = \vec{t}_\varphi-\left(\hat{\vec{n}}\cdot
            \vec{t}_\varphi\right)\hat{\vec{n}}
    \f]
    \f[
        \vec{t}_\varphi = \tau_\varphi\cdot\hat{\vec{n}}
    \f]
    \f[
        \tau_\varphi = \nu_\varphi \rho_\varphi \left(\nabla\vec{U}_\varphi+
            \left(\nabla\vec{U}_\varphi\right)^T\right)-
            \frac{1}{3}\delta_{ij}\nabla\vec{U}_\varphi
    \f]
    \f[
        \hat{\vec{n}} = -\frac{\alpha_2\nabla\alpha_1-\alpha_1\nabla\alpha_2}
            {\left|\alpha_2\nabla\alpha_1-\alpha_1\nabla\alpha_2\right|}
    \f]
    \f[
        A_{i} = \frac{2\left|\nabla\alpha_1\right|
            \left|\nabla\alpha_2\right|}
            {\left|\nabla\alpha_1\right|+\left|\nabla\alpha_2\right|}
    \f]
    \f[
        \rho_{mix} = \alpha_1\rho_1+\alpha_2\rho_2
    \f]
    The interfacial area density is clipped to be not smaller than
    \f$1/L_{max}\f$.

Usage
    \table
        Parameter       | Name  | Description       | Default value
        \f$U_{r,min}\f$ | residualSlip | minimum slip velocity  | -
        \f$L_{max}\f$   | LMax  | maximum length scale for area density  | 1 m
    \endtable

SourceFiles
    PorombkaHoehne.C

\*---------------------------------------------------------------------------*/

#ifndef PorombkaHoehne_H
#define PorombkaHoehne_H

#include "dragModel.H"
#include "segregatedPhaseInterface.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{
namespace dragModels
{

/*---------------------------------------------------------------------------*\
                       Class PorombkaHoehne Declaration
\*---------------------------------------------------------------------------*/

class PorombkaHoehne
:
    public dragModel
{
    // Private Data

        //- Interface
        const segregatedPhaseInterface interface_;

        //- Length scale to calculate minimum interfacial area density
        const dimensionedScalar LMax_;

        //- Residual slip velocity to avoid division by zero
        const dimensionedScalar residualSlip_;


    // Private Member Functions

        //- Return viscous stress tensor
        inline tmp<volSymmTensorField> tauVisc
        (
            const phaseModel& phase
        ) const;


public:

    //- Runtime type information
    TypeName("PorombkaHoehne");


    // Constructors

        //- Construct from a dictionary and an interface
        PorombkaHoehne
        (
            const dictionary& dict,
            const phaseInterface& interface,
            const bool registerObject
        );


    //- Destructor
    virtual ~PorombkaHoehne();


    // Member Functions

        //- Return the drag coefficient K
        //  used in the momentum equations
        //    ddt(alpha1*rho1*U1) + ... = ... K*(U1-U2)
        //    ddt(alpha2*rho2*U2) + ... = ... K*(U2-U1)
        virtual tmp<volScalarField> K() const;

        //- Return the drag coefficient Kf
        //  used in the face-momentum equations
        virtual tmp<surfaceScalarField> Kf() const;
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace dragModels
} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
