/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2021-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "LiaoLucas.H"
#include "heatTransferPhaseSystem.H"
#include "saturationModel.H"
#include "phaseCompressibleMomentumTransportModel.H"
#include "addToRunTimeSelectionTable.H"

// * * * * * * * * * * * * * * Static Data Members * * * * * * * * * * * * * //

namespace Foam
{
namespace heatTransferModels
{
    defineTypeNameAndDebug(LiaoLucas, 0);
    addToRunTimeSelectionTable(heatTransferModel, LiaoLucas, dictionary);
}
}

using Foam::constant::mathematical::pi;


// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

Foam::heatTransferModels::LiaoLucas::LiaoLucas
(
    const dictionary& dict,
    const phaseInterface& interface
)
:
    heatTransferModel(dict, interface),
    interface_
    (
        interface.modelCast<heatTransferModel, dispersedPhaseInterface>()
    )
{}


// * * * * * * * * * * * * * * * * Destructor  * * * * * * * * * * * * * * * //

Foam::heatTransferModels::LiaoLucas::~LiaoLucas()
{}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

Foam::tmp<Foam::volScalarField>
Foam::heatTransferModels::LiaoLucas::K(const scalar residualAlpha) const
{
    const phaseModel& liquid = interface_.continuous();
    const phaseModel& vapor = interface_.dispersed();
    const rhoThermo& lThermo = liquid.thermo();
    const rhoThermo& vThermo = vapor.thermo();
    const phaseSystem& fluid = interface_.continuous().fluid();
    const word volatileSpecie(fluid.lookupOrDefault<word>("volatile", "none"));

    const phaseCompressible::momentumTransportModel& lTurbulence =
        interface_.mesh().lookupObject
        <
            phaseCompressible::momentumTransportModel
        >
        (
            IOobject::groupName
            (
                momentumTransportModel::typeName,
                liquid.name()
            )
        );

    if (!fluid.foundInterfacialModel<saturationModel>(interface_))
    {
        FatalErrorInFunction
            << "Saturation model for phase interface " << interface_.name()
            << " required for " << this->type()
            << " heat transfer model." << exit(FatalError);
    }

    const saturationModel& satModel =
        fluid.lookupInterfacialModel<saturationModel>(interface_);

    const volScalarField L
    (
        volatileSpecie != "none"
      ? refCast<const heatTransferPhaseSystem>(fluid)
        .Li
         (
            interface_,
            volatileSpecie,
            fluid.dmdtf(interface_),
            satModel.Tsat(lThermo.p()),
            heatTransferPhaseSystem::latentHeatScheme::upwind
         )
      : refCast<const heatTransferPhaseSystem>(fluid)
        .L
         (
            interface_,
            fluid.dmdtf(interface_),
            satModel.Tsat(lThermo.p()),
            heatTransferPhaseSystem::latentHeatScheme::upwind
         )
    );

    const volScalarField Ja
    (
        lThermo.Cp()*lThermo.rho()*(lThermo.T() - satModel.Tsat(lThermo.p()))
       /(vThermo.rho()*L)
    );

    const volScalarField Pe
    (
        interface_.magUr()*vapor.d()*lThermo.rho()/lThermo.alphahe()
    );

    const volScalarField tauTurb(sqrt(lThermo.nu()/lTurbulence.epsilon()));

    const volScalarField Nu
    (
        12/pi*Ja
      + 2*sqrt(Pe/pi)
      + 2*vapor.d()/sqrt(pi*lThermo.alphahe()/lThermo.rho()*tauTurb)
    );

    return
        6
       *max(interface_.dispersed(), residualAlpha)
       *interface_.continuous().thermo().kappa()
       *Nu
       /sqr(interface_.dispersed().d());
}


// ************************************************************************* //
