/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2020-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

Class
    Foam::heatTransferModels::homogeneous

Description
    Model applies a very large heat transfer coefficient to equalise the phase
    temperatures and mimic the homogeneous (single temperature field) model.

    Coefficient is based on average heat conductivity of phase interface,
    local grid size and Nusselt number.

Usage
    \verbatim
    heatTransfer
    (
        (phase1 and phase2)
        {
            type        homogeneous;
            delta       cubeRootVol;
            Nu          100;
        }
    );
    \endverbatim

    \table
        Property     | Description                | Required    | Default value
        delta        | Local grid size (SGSdelta) | yes         | n/a
        Nu           | Nusselt number             | no          | 100
    \endtable

    For 1D cases use 'oneDimensional' delta model.

See also
    Foam::SGSdelta

SourceFiles
    homogeneous.C

\*---------------------------------------------------------------------------*/

#ifndef homogeneous_H
#define homogeneous_H

#include "heatTransferModel.H"
#include "SGSdelta.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{
namespace heatTransferModels
{

/*---------------------------------------------------------------------------*\
                        Class homogeneous Declaration
\*---------------------------------------------------------------------------*/

class homogeneous
:
    public heatTransferModel
{
    // Private Data

        //- Interface
        const phaseInterface interface_;

        //- Nusselt number
        dimensionedScalar Nu_;

        // Local grid size approximation model
        autoPtr<SGSdelta> delta_;


public:

    //- Runtime type information
    TypeName("homogeneous");


    // Constructors

        //- Construct from components
        homogeneous
        (
            const dictionary& dict,
            const phaseInterface& interface
        );


    //- Destructor
    virtual ~homogeneous();


    // Member Functions

        //- The heat transfer function K used in the enthalpy equation
        tmp<volScalarField> K(const scalar residualAlpha) const;
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace heatTransferModels
} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
