/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2018-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

Class
    Foam::inducedTurbulenceModel

Description

SourceFiles
    inducedTurbulenceModel.C
    inducedTurbulenceModelNew.C

\*---------------------------------------------------------------------------*/

#ifndef inducedTurbulenceModel_H
#define inducedTurbulenceModel_H

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#include "volFields.H"
#include "dictionary.H"
#include "dispersedPhaseInterface.H"
#include "dragModel.H"
#include "runTimeSelectionTables.H"

namespace Foam
{

/*---------------------------------------------------------------------------*\
                   Class inducedTurbulenceModel Declaration
\*---------------------------------------------------------------------------*/

class inducedTurbulenceModel
{
protected:

    // Protected data

        //- Phase interface
        const phaseInterface interface_;

        //- Reference to phaseInterface drag model
        const dragModel& dragModel_;


    // Protected Member Functions

        //- Induced turbulence kinetic energy coefficient
        virtual tmp<volScalarField> Ck() const = 0;

        //- Induced turbulence dissipation rate coefficient
        virtual tmp<volScalarField> Ceps() const = 0;


public:

    //- Runtime type information
    TypeName("inducedTurbulenceModel");


    // Declare runtime construction

        declareRunTimeSelectionTable
        (
            autoPtr,
            inducedTurbulenceModel,
            dictionary,
            (
                const dictionary& dict,
                const phaseInterface& interface
            ),
            (dict, interface)
        );


    // Static data members

        //- Turbulent kinetic energy coefficient dimensions
        static const dimensionSet dimKk;

        //- Turbulent dissipation coefficient dimensions
        static const dimensionSet dimKeps;

        //- Does this model require correcting on fixed flux boundaries?
        static const bool correctFixedFluxBCs = true;


    // Constructors

        //- Construct from a dictionary and an interface
        inducedTurbulenceModel
        (
            const dictionary& dict,
            const phaseInterface& interface
        );


    //- Destructor
    virtual ~inducedTurbulenceModel();


    // Selectors

        static autoPtr<inducedTurbulenceModel> New
        (
            const dictionary& dict,
            const phaseInterface& interface,
            const bool outer=true
        );


    // Member Functions

        //- Return the turbulent kinetic energy source for phase
        virtual tmp<volScalarField> Kk() const;

        //- Return the turbulent dissipation rate source for phase
        virtual tmp<volScalarField> Keps() const;
};


/*---------------------------------------------------------------------------*\
                Class blendedInducedTurbulenceModel Declaration
\*---------------------------------------------------------------------------*/

class blendedInducedTurbulenceModel
:
    public BlendedInterfacialModel<inducedTurbulenceModel>
{
public:

    // Constructors

        //- Inherit base class constructors
        using
            BlendedInterfacialModel<inducedTurbulenceModel>
          ::BlendedInterfacialModel;


    // Selectors

        static autoPtr<blendedInducedTurbulenceModel> New
        (
            const dictionary& dict,
            const phaseInterface& interface
        );


    // Member Functions

        //- Return the turbulent kinetic energy source for phase
        tmp<volScalarField> Kk() const;

        //- Return the turbulent dissipation rate source for phase
        tmp<volScalarField> Keps() const;
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
