/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

Class
    Foam::phaseTransferModels::dropletEntrainment

Description
    Droplet entrainment model

    The volume-specific mass transfer rate is computed based on a Weber number
    criterion. The Weber number and the time scale is computed by means of the
    interfacial velocity difference of the continuous phases. This implies that
    the entrainment rate goes to zero in fully resolved cases, which is a
    desired behaviour. The model takes into account that the entrainment rate
    saturates, i.e. it is only growing with the Weber number until a certain
    value is reached.

    \f[
        \Dot{m}=I\min(\Dot{m}_I,\Dot{m}_{\max})
    \f]

    with

    \f[
        \Dot{m}_I=C\frac{\rho_1\rho_2 |\mathbf{U}_\mathrm{r}|^3}{\sigma}
    \f]

    References:
    \verbatim
        Internal report by Wang et al.
    \endverbatim

Usage
    \table
        Property    | Description               | Required | Default
        cof         | dimensionless coefficient | no       | 1e-7
        rateMax     | maximum entrainment rate  | no       | 100 kg/m^3/s
    \endtable

SourceFiles
    dropletEntrainment.C

\*---------------------------------------------------------------------------*/

#ifndef dropletEntrainment_H
#define dropletEntrainment_H

#include "entrainmentModel.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

namespace phaseTransferModels
{

/*---------------------------------------------------------------------------*\
                     Class dropletEntrainment Declaration
\*---------------------------------------------------------------------------*/

class dropletEntrainment
:
    public entrainmentModel
{
    // Private Data

        //- Dimensionless coefficient
        const dimensionedScalar cof_;

        //- Maximum entrainment rate
        const dimensionedScalar rateMax_;

protected:

    // Protected Member Functions

        //- Returns the droplet transfer rate
        virtual tmp<volScalarField> rate(const phaseModel&) const;


public:

    //- Runtime type information
    TypeName("dropletEntrainment");


    // Constructors

        //- Construct from components
        dropletEntrainment
        (
            const dictionary& dict,
            const phaseInterface& interface
        );


    //- Destructor
    virtual ~dropletEntrainment();
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace phaseTransferModels
} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
