/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2021-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

Class
    Foam::phaseTransferModels::accumulationToContinuous

Description
    Morphology transfer model representing change from a dispersed phase to a
    continuous phase with identical physical properties. This change is
    triggered by a threshold value for the void fraction sum of the gas phase
    pair. I.e. the transfer from disperse to continuous is active as long as
    alphaGasDisperse + alphaGasContinuous >= alphaThreshold is true.

Usage
    \table
        Property       | Description                | Required | Default
        alphaThreshold | Critical void fraction sum | no   | continuousThreshold
    \endtable

See also
    Foam::phaseTransferModels::morphologyTransferModel

SourceFiles
    accumulationToContinuous.C

\*---------------------------------------------------------------------------*/

#ifndef accumulationToContinuous_H
#define accumulationToContinuous_H

#include "morphologyTransferModel.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

namespace phaseTransferModels
{

/*---------------------------------------------------------------------------*\
                Class accumulationToContinuous Declaration
\*---------------------------------------------------------------------------*/

class accumulationToContinuous
:
    public morphologyTransferModel
{
    // Private Data

        //- Disperse member of phase pair
        const phaseModel& pairDisperse_;

        //- Continuous member of phase pair
        const phaseModel& pairContinuous_;

        //- Threshold for volume fraction
        const dimensionedScalar alphaThreshold_;


public:

    //- Runtime type information
    TypeName("accumulationToContinuous");


    // Constructors

        //- Construct from components
        accumulationToContinuous
        (
            const dictionary& dict,
            const phaseInterface& interface
        );


    //- Destructor
    virtual ~accumulationToContinuous();


    // Member Functions

        //- The mass transfer rate for the mixture
        virtual tmp<volScalarField> dmdtf() const;
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace phaseTransferModels
} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
