/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2020-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

Class
    Foam::turbulentDispersionModels::ShiRzehakTurbulentDispersion

Description
    Turbulent dispersion model of Shi and Rzehak for turbulent dispersed
    solid-liquid flows.

    \verbatim
        Shi, P., & Rzehak, R. (2020).
        Solid-liquid flow in stirred tanks: Euler-Euler / RANS modeling.
        Chemical Engineering Science, 227, 115875.
    \endverbatim

    Note: Model is written in consistency with Burns et al. 2004 model using:

              \f[
                  \sigma_T = 1 + St
              \f]

SourceFiles
    ShiRzehakTurbulentDispersion.C

\*---------------------------------------------------------------------------*/

#ifndef ShiRzehakTurbulentDispersion_H
#define ShiRzehakTurbulentDispersion_H

#include "dispersedTurbulentDispersionModel.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

namespace turbulentDispersionModels
{

/*---------------------------------------------------------------------------*\
                Class ShiRzehakTurbulentDispersion Declaration
\*---------------------------------------------------------------------------*/

class ShiRzehakTurbulentDispersion
:
    public dispersedTurbulentDispersionModel
{
    // Private Data

        //- Constant to describe the Lagrangian integral timescale
        //  following the fluid motion
        const dimensionedScalar CT_;

        //- Constant to describe the Eulerian longitudinal
        //  integral length scale
        const dimensionedScalar CLambda_;


public:

    //- Runtime type information
    TypeName("ShiRzehak");


    // Constructors

        //- Construct from a dictionary and an interface
        ShiRzehakTurbulentDispersion
        (
            const dictionary& dict,
            const phaseInterface& interface
        );


    //- Destructor
    virtual ~ShiRzehakTurbulentDispersion();


    // Member Functions

        //- Turbulent diffusivity
        //  multiplying the gradient of the phase-fraction
        virtual tmp<volScalarField> D() const;
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace turbulentDispersionModels
} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
