/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2018-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

Class
    Foam::turbulentNormalStressPressureFvPatchScalarField

Description
    This boundary condition provides a turbulentNormalStressPressure condition,
    calculated as:

        \f[
            p_{TNS} = p + \frac{2}{3}\sum_\varphi\varphi\rho_\varphi\,k_\varphi
        \f]

    where
    \vartable
        p_{TNS}      | turbulent normal stress pressure [Pa]
        p            | static pressure [units]
        \varphi      | phase fraction [-]
        \rho_\varphi | phase density [kg/m^3]
        k_\varphi    | phase turbulent kinetic energy [m^2/s^2]
    \endtable

Usage
    \table
        Property | Description                      | Req'd? | Default
        pTNS     | turbulent normal stress pressure | yes    |
    \endtable

    Example of the boundary condition specification:
    \verbatim
    <patchName>
    {
        type       turbulentNormalStressPressure;
        pTNS       101329;
        value      uniform 101329; // optional initial value
    }
    \endverbatim

SourceFiles
    turbulentNormalStressPressureFvPatchScalarField.C

\*---------------------------------------------------------------------------*/

#ifndef turbulentNormalStressPressureFvPatchScalarField_H
#define turbulentNormalStressPressureFvPatchScalarField_H

#include "fixedValueFvPatchFields.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

/*---------------------------------------------------------------------------*\
       Class turbulentNormalStressPressureFvPatchScalarField Declaration
\*---------------------------------------------------------------------------*/

class turbulentNormalStressPressureFvPatchScalarField
:
    public fixedValueFvPatchScalarField
{
    // Private Data

        //- Static pressure that includes turbulent normal stress
        scalarField pTNS_;

    // Private Member Functions

        //- Check explicitTurbulentNormalStress switch
        void explicitTurbulentNormalStressRequired
        (
            const dictionary&
        );

public:

    //- Runtime type information
    TypeName("turbulentNormalStressPressure");


    // Constructors

        //- Construct from patch and internal field
        turbulentNormalStressPressureFvPatchScalarField
        (
            const fvPatch&,
            const DimensionedField<scalar, volMesh>&
        );

        //- Construct from patch, internal field and dictionary
        turbulentNormalStressPressureFvPatchScalarField
        (
            const fvPatch&,
            const DimensionedField<scalar, volMesh>&,
            const dictionary&
        );

        //- Construct by mapping given fixedValueTypeFvPatchField
        //  onto a new patch
        turbulentNormalStressPressureFvPatchScalarField
        (
            const turbulentNormalStressPressureFvPatchScalarField&,
            const fvPatch&,
            const DimensionedField<scalar, volMesh>&,
            const fvPatchFieldMapper&
        );

        //- Disallow copy without setting internal field reference
        turbulentNormalStressPressureFvPatchScalarField
        (
            const turbulentNormalStressPressureFvPatchScalarField&
        ) = delete;

        //- Construct as copy setting internal field reference
        turbulentNormalStressPressureFvPatchScalarField
        (
            const turbulentNormalStressPressureFvPatchScalarField&,
            const DimensionedField<scalar, volMesh>&
        );

        //- Construct and return a clone setting internal field reference
        virtual tmp<fvPatchScalarField> clone
        (
            const DimensionedField<scalar, volMesh>& iF
        ) const
        {
            return tmp<fvPatchScalarField>
            (
                new turbulentNormalStressPressureFvPatchScalarField
                (
                    *this,
                    iF
                )
            );
        }


    // Member Functions

        // Mapping functions

            //- Map (and resize as needed) from self given a mapping object
            virtual void autoMap
            (
                const fvPatchFieldMapper&
            );

            //- Reverse map the given fvPatchField onto this fvPatchField
            virtual void rmap
            (
                const fvPatchScalarField&,
                const labelList&
            );


        // Evaluation functions

            //- Update the coefficients associated with the patch field
            virtual void updateCoeffs();


        //- Write
        virtual void write(Ostream&) const;
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //


#endif

// ************************************************************************* //
