/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2021-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "ShinJonesDepartureDiameter.H"
#include "addToRunTimeSelectionTable.H"
#include "uniformDimensionedFields.H"
#include "compressibleMomentumTransportModels.H"
#include "phaseCompressibleMomentumTransportModel.H"
#include "phaseSystem.H"

// * * * * * * * * * * * * * * Static Data Members * * * * * * * * * * * * * //

namespace Foam
{
namespace wallBoilingModels
{
namespace departureDiameterModels
{
    defineTypeNameAndDebug(ShinJonesDepartureDiameter, 0);
    addToRunTimeSelectionTable
    (
        departureDiameterModel,
        ShinJonesDepartureDiameter,
        dictionary
    );
}
}
}


// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

Foam::wallBoilingModels::departureDiameterModels::
ShinJonesDepartureDiameter::
ShinJonesDepartureDiameter
(
    const dictionary& dict
)
:
    departureDiameterModel(),
    phaseName_(dict.lookupOrDefault<word>("phase", "liquid")),
    K_(dict.lookupOrDefault<scalar>("K", scalar(1))),
    dMin_(dict.lookupOrDefault<scalar>("dMin", scalar(1e-6)))
{}


Foam::wallBoilingModels::departureDiameterModels::
ShinJonesDepartureDiameter::
ShinJonesDepartureDiameter
(
    const ShinJonesDepartureDiameter& model
)
:
    departureDiameterModel()
{}


// * * * * * * * * * * * * * * * * Destructor  * * * * * * * * * * * * * * * //

Foam::wallBoilingModels::departureDiameterModels::
ShinJonesDepartureDiameter::
~ShinJonesDepartureDiameter()
{}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

Foam::tmp<Foam::scalarField>
Foam::wallBoilingModels::departureDiameterModels::
ShinJonesDepartureDiameter::dDeparture
(
    const phaseModel& liquid,
    const phaseModel& vapor,
    const label patchi,
    const scalarField& Tl,
    const scalarField& Tsatw,
    const scalarField& L
) const
{
    const scalarField sigmaw
    (
        liquid.fluid().sigma(phaseInterfaceKey(liquid, vapor), patchi)
    );

    const scalarField rhoLiquid(liquid.thermo().rho(patchi));
    const scalarField rhoVapor(vapor.thermo().rho(patchi));
    const scalarField muLiquid(liquid.thermo().mu(patchi));
    const scalarField nuLiquid(liquid.thermo().nu(patchi));
    const scalarField nuVapor(vapor.thermo().nu(patchi));

    const word modelName
    (
        IOobject::groupName
        (
            momentumTransportModel::typeName,
            phaseName_
        )
    );

    const compressibleMomentumTransportModel& model =
        liquid.mesh().lookupObject<compressibleMomentumTransportModel>
        (
            modelName
        );

    const volSymmTensorField tau(model.devTau());
    const symmTensorField taup(tau.boundaryField()[patchi]);
    const vectorField Sfp(liquid.mesh().Sf().boundaryField()[patchi]);
    const scalarField magSfp(liquid.mesh().magSf().boundaryField()[patchi]);
    const scalarField magTauw(mag((-Sfp/magSfp) & taup));

    const scalarField dCav
    (
        pos(Tl - Tsatw)*4*sigmaw*Tsatw
       /(rhoVapor*max((Tl - Tsatw), scalar(1e-8))*L)
    );

    const scalarField dDepReLow
    (
        pos(dCav)*2/sqrt(3*sigmaw*max(dCav, scalar(1e-16))/magTauw)
    );

    const scalarField dDepReMed
    (
        pos(dCav)*1.1574*pow(sigmaw*max(dCav, scalar(1e-16))/2, 0.3571)
       *pow(rhoLiquid, 0.1429)*pow(nuLiquid, 0.2857)/sqrt(magTauw)
    );

    const scalarField dDepReHigh
    (
        pos(dCav)*3.4728*pow025(sigmaw*rhoLiquid*max(dCav, scalar(1e-16))/2)
       *sqrt(nuLiquid/magTauw)
    );

    const scalarField ReLow
    (
        magTauw*sqr(dDepReLow/2)/(rhoLiquid*sqr(nuLiquid))
    );

    const scalarField ReMed
    (
        magTauw*sqr(dDepReMed/2)/(rhoLiquid*sqr(nuLiquid))
    );

    const scalarField ReHigh
    (
        magTauw*sqr(dDepReHigh/2)/(rhoLiquid*sqr(nuLiquid))
    );

    return
       max
       (
           K_
          *(
               neg(ReLow - 1)*dDepReLow
             + pos0(ReMed - 1)*neg(ReMed - 500)*dDepReMed
             + pos0(ReHigh - 500)*dDepReHigh
           ),
           dMin_
       );
}


void Foam::wallBoilingModels::departureDiameterModels::
ShinJonesDepartureDiameter::write(Ostream& os) const
{
    departureDiameterModel::write(os);
    writeKeyword(os, "liquid") << phaseName_ << token::END_STATEMENT << nl;
    writeKeyword(os, "K") << K_ << token::END_STATEMENT << nl;
    writeKeyword(os, "dMin") << dMin_ << token::END_STATEMENT << nl;
}


// ************************************************************************* //
