/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2018-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "InducedTurbulencePhaseSystem.H"

#include "inducedTurbulenceModel.H"

#include "HashPtrTable.H"


// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

template<class BasePhaseSystem>
Foam::InducedTurbulencePhaseSystem<BasePhaseSystem>::
InducedTurbulencePhaseSystem
(
    const fvMesh& mesh
)
:
    BasePhaseSystem(mesh)
{
    this->generateInterfacialModels(inducedTurbulenceModels_);

    forAllConstIter
    (
        inducedTurbulenceModelTable,
        inducedTurbulenceModels_,
        inducedTurbulenceModelIter
    )
    {
        const phaseInterface& interface =
            inducedTurbulenceModelIter()->interface();

        Kks_.insert
        (
            inducedTurbulenceModelIter.key(),
            new volScalarField
            (
                IOobject::groupName("Kk", interface.name()),
                inducedTurbulenceModelIter()->Kk()
            )
        );

        Kepss_.insert
        (
            inducedTurbulenceModelIter.key(),
            new volScalarField
            (
                IOobject::groupName("Keps", interface.name()),
                inducedTurbulenceModelIter()->Keps()
            )
        );
    }
}


// * * * * * * * * * * * * * * * * Destructor  * * * * * * * * * * * * * * * //

template<class BasePhaseSystem>
Foam::InducedTurbulencePhaseSystem<BasePhaseSystem>::
~InducedTurbulencePhaseSystem()
{}


// * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * * //

template<class BasePhaseSystem>
Foam::autoPtr<Foam::phaseSystem::momentumTransferTable>
Foam::InducedTurbulencePhaseSystem<BasePhaseSystem>::momentumTransfer()
{
    // Create a momentum transfer matrix for each phase and fill it from
    // BasePhaseSystem (momentumTransferPhaseSystem)
    autoPtr<phaseSystem::momentumTransferTable> eqnsPtr
    (
        BasePhaseSystem::momentumTransfer()
    );

    // Update the induced turbulence source terms
    forAllConstIter
    (
        inducedTurbulenceModelTable,
        inducedTurbulenceModels_,
        inducedTurbulenceModelIter
    )
    {
        *Kks_[inducedTurbulenceModelIter.key()] =
            inducedTurbulenceModelIter()->Kk();

        *Kepss_[inducedTurbulenceModelIter.key()] =
            inducedTurbulenceModelIter()->Keps();
    }

    return eqnsPtr;
}


template<class BasePhaseSystem>
Foam::tmp<Foam::volScalarField>
Foam::InducedTurbulencePhaseSystem<BasePhaseSystem>::Kk
(
    const phaseModel& phase
) const
{
    tmp<volScalarField> tKk
    (
        volScalarField::New
        (
            inducedTurbulenceModel::typeName + ":Kk",
            this->mesh_,
            dimensionedScalar(inducedTurbulenceModel::dimKk, 0)
        )
    );

    forAllConstIter
    (
        KkTable,
        Kks_,
        KkIter
    )
    {
        const volScalarField& K(*KkIter());

        const phaseInterface interface(*this, KkIter.key());

        if (interface.contains(phase)) tKk.ref() += K;
    }

    return tKk;
}


template<class BasePhaseSystem>
Foam::tmp<Foam::volSymmTensorField>
Foam::InducedTurbulencePhaseSystem<BasePhaseSystem>::tensorKk
(
    const phaseModel& phase
) const
{
    tmp<volSymmTensorField> tKkUU
    (
        volSymmTensorField::New
        (
            inducedTurbulenceModel::typeName + ":KkUU",
            this->mesh_,
            dimensionedSymmTensor(inducedTurbulenceModel::dimKk, Zero)
        )
    );

    forAllConstIter
    (
        KkTable,
        Kks_,
        KkIter
    )
    {
        const volScalarField& K(*KkIter());

        const phaseInterface interface(*this, KkIter.key());

        if (interface.contains(phase))
        {
            const dimensionedScalar smallU("smallU", dimVelocity, 1e-16);

            // Sign doesn't matter for Ur as we calculate outer product later
            const volVectorField Ur
            (
                interface.phase1().U() - interface.phase2().U()
            );
            const volVectorField Ur_hat(Ur/(mag(Ur) + smallU));
            const volSymmTensorField uu(symm(Ur_hat*Ur_hat));

            tKkUU.ref() += 0.5*K*(symmTensor::I + uu);
        }
    }

    return tKkUU;
}


template<class BasePhaseSystem>
Foam::tmp<Foam::volScalarField>
Foam::InducedTurbulencePhaseSystem<BasePhaseSystem>::Keps
(
    const phaseModel& phase
) const
{
    tmp<volScalarField> tKeps
    (
        volScalarField::New
        (
            inducedTurbulenceModel::typeName + ":Keps",
            this->mesh_,
            dimensionedScalar(inducedTurbulenceModel::dimKeps, 0)
        )
    );

    forAllConstIter
    (
        KepsTable,
        Kepss_,
        KepsIter
    )
    {
        const volScalarField& K(*KepsIter());

        const phaseInterface interface(*this, KepsIter.key());

        if (interface.contains(phase)) tKeps.ref() += K;
    }

    return tKeps;
}


template<class BasePhaseSystem>
bool Foam::InducedTurbulencePhaseSystem<BasePhaseSystem>::read()
{
    if (BasePhaseSystem::read())
    {
        bool readOK = true;

        // Read models ...

        return readOK;
    }
    else
    {
        return false;
    }
}


// ************************************************************************* //
