/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2018-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "ResolvedPhaseSystem.H"

#include "interfaceCompressionModel.H"
#include "surfaceTensionForceModel.H"
#include "subGridSurfaceTensionModel.H"
#include "interfaceIndicatorMethod.H"
#include "underResolutionModel.H"

#include "segregatedPhaseInterface.H"

// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

template<class BasePhaseSystem>
Foam::ResolvedPhaseSystem<BasePhaseSystem>::ResolvedPhaseSystem
(
    const fvMesh& mesh
)
:
    BasePhaseSystem(mesh)
{
    this->generateInterfacialModels(interfaceCompressionModels_);

    surfaceTensionForceModel_ = surfaceTensionForceModel::New(*this);

    if (this->found("subGridSurfaceTension"))
    {
        this->generateInterfacialModels(subGridSurfaceTensionModels_);
    }

    interfaceIndicatorMethod_ = interfaceIndicatorMethod::New(*this);

    if (this->found("underResolution"))
    {
        this->generateInterfacialModels(underResolutionModels_);
    }
}


// * * * * * * * * * * * * * * * * Destructor  * * * * * * * * * * * * * * * //

template<class BasePhaseSystem>
Foam::ResolvedPhaseSystem<BasePhaseSystem>::~ResolvedPhaseSystem()
{}


// * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * * //

template<class BasePhaseSystem>
Foam::tmp<Foam::surfaceScalarField>
Foam::ResolvedPhaseSystem<BasePhaseSystem>::phir
(
    const phaseModel& phase1,
    const phaseModel& phase2
) const
{
    tmp<surfaceScalarField> phir(BasePhaseSystem::phir(phase1, phase2));

    const phaseInterfaceKey key(phase1, phase2);

    if (interfaceCompressionModels_.found(key))
    {
        if (interfaceCompressionModels_[key]().type() != "none")
        {
            phir.ref() += interfaceCompressionModels_[key]().phirc(phase1);
        }
    }

    return phir;
}


template<class BasePhaseSystem>
void Foam::ResolvedPhaseSystem<BasePhaseSystem>::addCompressionFlux
(
    const phaseModel& phase,
    surfaceScalarField& alphaPhi
) const
{
    BasePhaseSystem::addCompressionFlux(phase, alphaPhi);

    if (!interfaceCompressionModels_.empty())
    {
        const volScalarField& alpha = phase;

        forAll(this->phases(), phasei)
        {
            const phaseModel& phase2 = this->phases()[phasei];
            const volScalarField& alpha2 = phase2;
            const phaseInterfaceKey key(phase, phase2);

            if (&phase2 == &phase) continue;

            interfaceCompressionModelTable::const_iterator cAlpha
            (
                interfaceCompressionModels_.find(key)
            );

            if (cAlpha != interfaceCompressionModels_.end())
            {
                const surfaceScalarField phir
                (
                    interfaceCompressionModels_[key]().phirc(phase)
                );

                const word phirScheme
                (
                    "div(phir,"
                  + alpha2.name() + ',' + alpha.name()
                  + ')'
                );

                alphaPhi += fvc::flux
                (
                    -fvc::flux(-phir, alpha2, phirScheme),
                    alpha,
                    phirScheme
                );
            }
        }
    }
}


template<class BasePhaseSystem>
Foam::tmp<Foam::surfaceScalarField>
Foam::ResolvedPhaseSystem<BasePhaseSystem>::surfaceTension
(
    const phaseModel& phase1
) const
{
    tmp<surfaceScalarField> tSurfaceTension
    (
        new surfaceScalarField
        (
            IOobject
            (
                "surfaceTension",
                this->mesh().time().timeName(),
                this->mesh()
            ),
            this->mesh(),
            dimensionedScalar
            (
                "surfaceTension",
                dimensionSet(1, -2, -2, 0, 0),
                0
            )
        )
    );

    forAll(this->phases(), phasej)
    {
        const phaseModel& phase2 = this->phases()[phasej];

        if (&phase2 == &phase1) continue;

        const phaseInterfaceKey key(phase1, phase2);

        if (interfaceCompressionModels_.found(key))
        {
            if (interfaceCompressionModels_[key]().type() != "none")
            {
                const segregatedPhaseInterface interface(phase1, phase2);

                tSurfaceTension.ref() +=
                    surfaceTensionForceModel_->surfaceTensionForcef(interface);

                if (subGridSurfaceTensionModels_.found(key))
                {
                    if
                    (
                        subGridSurfaceTensionModels_[key]().type()
                        != "none"
                    )
                    {
                        tSurfaceTension.ref() +=
                            subGridSurfaceTensionModels_[key]().
                            subGridSurfaceTensionForcef();
                    }
                }
            }
        }
    }

    return tSurfaceTension;
}


template<class BasePhaseSystem>
Foam::tmp<Foam::volScalarField>
Foam::ResolvedPhaseSystem<BasePhaseSystem>::interfaceIndicator
(
    const phaseModel& phase
) const
{
   return interfaceIndicatorMethod_->interfaceIndicator(phase);
}


template<class BasePhaseSystem>
Foam::tmp<Foam::volVectorField> Foam::ResolvedPhaseSystem<BasePhaseSystem>::UrI
(
    const phaseInterfaceKey& key
) const
{
    if (underResolutionModels_.found(key))
    {
        return underResolutionModels_[key]().UrI();
    }
    else
    {
        FatalErrorInFunction
            << "Requesting UrI from phase interface "
            << phaseInterface(*this, key).name()
            << ", underResolutionModel not found." << endl;

        return tmp<volVectorField>(nullptr);
    }
}


template<class BasePhaseSystem>
Foam::tmp<Foam::volScalarField>
Foam::ResolvedPhaseSystem<BasePhaseSystem>::underResolution
(
    const phaseInterfaceKey& key
) const
{
    if (underResolutionModels_.found(key))
    {
        return underResolutionModels_[key]().underResolution();
    }
    else
    {
        FatalErrorInFunction
            << "Requesting underResolution from phase interface "
            << phaseInterface(*this, key).name()
            << ", underResolutionModel not found." << endl;

        return tmp<volScalarField>(nullptr);
    }
}


template<class BasePhaseSystem>
Foam::tmp<Foam::volScalarField>
Foam::ResolvedPhaseSystem<BasePhaseSystem>::K
(
    const phaseInterfaceKey& key
) const
{
    const phaseInterface interface(*this, key);

    return surfaceTensionForceModel_->K
    (
        segregatedPhaseInterface(interface.phase1(), interface.phase2())
    );
}


template<class BasePhaseSystem>
void Foam::ResolvedPhaseSystem<BasePhaseSystem>::correct()
{
    BasePhaseSystem::correct();

    forAllIter
    (
        underResolutionModelTable,
        underResolutionModels_,
        underResolutionModelIter
    )
    {
        underResolutionModelIter()->correct();
    }
}


template<class BasePhaseSystem>
bool Foam::ResolvedPhaseSystem<BasePhaseSystem>::read()
{
    if (BasePhaseSystem::read())
    {
        bool readOK = true;

        // Read models ...

        return readOK;
    }
    else
    {
        return false;
    }
}


// ************************************************************************* //
