/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2018-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

Class
    Foam::ResolvedPhaseSystem

Description
    Class which adds functionality for modeling resolved interfaces such as
    interface compression, surface tension and contact angle correction.

SourceFiles
    ResolvedPhaseSystem.C

\*---------------------------------------------------------------------------*/

#ifndef ResolvedPhaseSystem_H
#define ResolvedPhaseSystem_H

#include "phaseSystem.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

class interfaceCompressionModel;
class surfaceTensionForceModel;
class subGridSurfaceTensionModel;
class interfaceIndicatorMethod;
class underResolutionModel;

/*---------------------------------------------------------------------------*\
                     Class ResolvedPhaseSystem Declaration
\*---------------------------------------------------------------------------*/

template<class BasePhaseSystem>
class ResolvedPhaseSystem
:
    public BasePhaseSystem
{
private:

    // Private Typedefs

        typedef HashTable
        <
            autoPtr<interfaceCompressionModel>,
            phaseInterfaceKey,
            phaseInterfaceKey::hash
        > interfaceCompressionModelTable;

        typedef HashTable
        <
            autoPtr<subGridSurfaceTensionModel>,
            phaseInterfaceKey,
            phaseInterfaceKey::hash
        > subGridSurfaceTensionModelTable;

        typedef HashTable
        <
            autoPtr<underResolutionModel>,
            phaseInterfaceKey,
            phaseInterfaceKey::hash
        > underResolutionModelTable;


    // Private Data

        //- Interface compression models
        interfaceCompressionModelTable interfaceCompressionModels_;

        //- Surface tension force models
        autoPtr<surfaceTensionForceModel> surfaceTensionForceModel_;

        //- Sub grid surface tension models
        subGridSurfaceTensionModelTable subGridSurfaceTensionModels_;

        //- Resolved interface detection method
        autoPtr<interfaceIndicatorMethod> interfaceIndicatorMethod_;

        //- Under resolution models
        underResolutionModelTable underResolutionModels_;


public:

    // Constructors

        //- Construct from fvMesh
        ResolvedPhaseSystem(const fvMesh&);


    //- Destructor
    virtual ~ResolvedPhaseSystem();


    // Member Functions

        //- Return relative flux for interface compression
        virtual tmp<surfaceScalarField> phir
        (
            const phaseModel& phase1,
            const phaseModel& phase2
        ) const;

        //- Add interface compression flux
        virtual void addCompressionFlux
        (
            const phaseModel& phase,
            surfaceScalarField& alphaPhi
        ) const;

        //- Return the surface tension force
        tmp<surfaceScalarField> surfaceTension(const phaseModel& phase) const;

        //- Indicator field of the proximity of a resolved phase interface
        //  Values are distributed from 0 (far) to 1 (near).
        virtual tmp<volScalarField> interfaceIndicator
        (
            const phaseModel& phase
        ) const;

        //- Relative velocity across the interface
        virtual tmp<volVectorField> UrI(const phaseInterfaceKey& key) const;

        //- Degree of under resolution due to shear and curvature for phase pair
        virtual tmp<volScalarField> underResolution
        (
            const phaseInterfaceKey& key
        ) const;

        //- Interface curvature for phase pair
        tmp<volScalarField> K(const phaseInterfaceKey& key) const;

        //- Correct derived properties
        virtual void correct();

        //- Read base phaseProperties dictionary
        virtual bool read();
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#ifdef NoRepository
    #include "ResolvedPhaseSystem.C"
#endif

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
