/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2020-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

Class
    Foam::interfaceIndicatorMethods::gradAlpha

Description
    Determine the interface indicator field for a pair of continuous phases,
    i and j:

    \f[
        \mathrm{RI}_{ij} =
            \mathrm{max}(\mathrm{mag}((\nabla\alpha)_{ij}\Delta x), 1),
    \f]

    where

    \f[
        (\nabla\alpha)_{ij} = \alpha_j\nabla\alpha_i - \alpha_i\nabla\alpha_j,
    \f]

    and \f$\Delta x\f$ is the typical grid size calculated using
    a SGSdelta model.

Usage
    \verbatim
    interfaceIndicatorMethod
    {
        type    gradAlpha;
        delta   cubeRootVol;
    }
    \endverbatim

See also
    Foam::SGSdelta

SourceFiles
    gradAlpha.C

\*---------------------------------------------------------------------------*/

#ifndef gradAlpha_H
#define gradAlpha_H

#include "interfaceIndicatorMethod.H"
#include "SGSdelta.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{
namespace interfaceIndicatorMethods
{

/*---------------------------------------------------------------------------*\
                          Class gradAlpha Declaration
\*---------------------------------------------------------------------------*/

class gradAlpha
:
    public interfaceIndicatorMethod
{
    // Private Data

        // Local grid size approximation model
        autoPtr<SGSdelta> delta_;


protected:

    // Protected Member Functions

        //- Indicator field between a phase pair
        virtual tmp<volScalarField> interfaceIndicator
        (
            const phaseModel& phase1,
            const phaseModel& phase2
        ) const;


public:

    //- Runtime type information
    TypeName("gradAlpha");


    // Constructors

        //- Construct from phaseSystem
        gradAlpha(const phaseSystem& fluid);


    //- Destructor
    virtual ~gradAlpha();
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace interfaceIndicatorMethods
} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
