/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2021-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

Class
    Foam::surfaceTensionForceModels::smoothedContinuumSurfaceForce

Description
    Smoothing alpha fields prior to calculating the gradient vector of a phase
    pair to get interface curvature helps to dampen spurious currents. A number
    of two smoothing iterations show the least spurious currents.

    Reference:
    \verbatim
        Ubbink, O. (1997).
        Numerical prediction of two fluid systems with sharp interfaces.
        PhD Thesis, 193 pp. Imperial College of Science, Technology & Medicine
        London.
    \endverbatim

Usage
    \table
        Name         | Description                        | Default
        nSmoothIters | Number of iterations for smoothing | 2
    \endtable

SourceFiles
    smoothedContinuumSurfaceForce.C

\*---------------------------------------------------------------------------*/

#ifndef smoothedContinuumSurfaceForce_H
#define smoothedContinuumSurfaceForce_H

#include "continuumSurfaceForce.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{
namespace surfaceTensionForceModels
{

/*---------------------------------------------------------------------------*\
                    Class smoothedContinuumSurfaceForce Declaration
\*---------------------------------------------------------------------------*/

class smoothedContinuumSurfaceForce
:
    public continuumSurfaceForce
{
    // Private Data

        //- Number of times to smoothen field
        const label nSmoothIters_;


    // Private Member Functions

        //- Smoothen a volScalarField for defined number of times
        tmp<volScalarField> smoothen(const volScalarField& vsf) const;


public:

    //- Runtime type information
    TypeName("smoothedContinuumSurfaceForce");


    // Constructors

        //- Construct from phaseSystem
        smoothedContinuumSurfaceForce(const phaseSystem& fluid);


    //- Destructor
    virtual ~smoothedContinuumSurfaceForce();


    // Member Functions

        //- Return the curvature of this interface
        virtual tmp<volScalarField> K
        (
            const segregatedPhaseInterface& interface
        ) const;
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace surfaceTensionForceModels
} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
