/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2021-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

Class
    Foam::underResolutionModel

Description
    Base class for under-resolution models.

See also
    Foam::ResolvedPhaseSystem

SourceFiles
    underResolutionModel.C
    underResolutionModelNew.C

\*---------------------------------------------------------------------------*/

#ifndef underResolutionModel_H
#define underResolutionModel_H

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#include "volFields.H"
#include "dictionary.H"
#include "segregatedPhaseInterface.H"
#include "runTimeSelectionTables.H"

namespace Foam
{

/*---------------------------------------------------------------------------*\
                    Class underResolutionModel Declaration
\*---------------------------------------------------------------------------*/

class underResolutionModel
{
protected:

    // Protected Data

        //- Phase pair
        const segregatedPhaseInterface interface_;

        //- Reference to the mesh
        const fvMesh& mesh_;

        //- Degree of under resolution
        volScalarField underResolution_;


    // Protected Member Functions

        //- Primitive smoothing procedure for arbitrary field
        template<class fieldType>
        void smoothen
        (
            fieldType& field,
            const label nLayers = 2
        ) const;


public:

    //- Runtime type information
    TypeName("underResolutionModel");


    // Declare runtime construction

        declareRunTimeSelectionTable
        (
            autoPtr,
            underResolutionModel,
            dictionary,
            (
                const dictionary& dict,
                const phaseInterface& interface
            ),
            (dict, interface)
        );


    // Static Data Members

        //- Does this model require correcting on fixed flux boundaries?
        static const bool correctFixedFluxBCs = true;


    // Constructors

        //- Construct from a dictionary and an interface
        underResolutionModel
        (
            const dictionary& dict,
            const phaseInterface& interface
        );


    //- Destructor
    virtual ~underResolutionModel();


    // Selectors

        static autoPtr<underResolutionModel> New
        (
            const dictionary& dict,
            const phaseInterface& interface,
            const bool outer=true
        );


    // Member Functions

        //- Relative velocity across the interface
        virtual tmp<volVectorField> UrI() const;

        //- Degree of under resolution
        tmp<volScalarField> underResolution() const;

        //- Correct under resolution factor
        virtual void correct() {}
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#ifdef NoRepository
    #include "underResolutionModelTemplates.C"
#endif

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
