/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2018-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "TransitionPhaseSystem.H"

#include "BlendedInterfacialModel.H"
#include "clusteringByMomentumEqnModel.H"
#include "clusteringByContinuityEqnModel.H"
#include "populationBalanceModel.H"
#include "fvMeshSubset.H"
#include "cellSet.H"

#include "HashPtrTable.H"


// * * * * * * * * * * * * Private Member Functions * * * * * * * * * * * * //

template<class BasePhaseSystem>
template<class Type>
Foam::Map<Type>
Foam::TransitionPhaseSystem<BasePhaseSystem>::regionSum
(
    const volScalarField& regionField,
    const label nStructures,
    const Field<Type>& fld
) const
{
    Map<Type> regionToSum(nStructures);

    forAll(regionField, celli)
    {
        label regionI = regionField[celli];

        // Omit cells with negative index
        if (regionI != -1)
        {
            // Find and return an const_iterator set at the hashedEntry
            // If not found iterator = end()
            typename Map<Type>::iterator fnd = regionToSum.find(regionI);
            if (fnd == regionToSum.end())
            {
                regionToSum.insert(regionI, fld[celli]);
            }
            else
            {
                fnd() += fld[celli];
            }
        }
    }

    // Gather and scatter information from and to all processors
    Pstream::mapCombineGather(regionToSum, plusEqOp<Type>());
    Pstream::mapCombineScatter(regionToSum);

    return regionToSum;
}


template<class BasePhaseSystem>
template<class Type>
void Foam::TransitionPhaseSystem<BasePhaseSystem>::combineFields
(
    Field<Type>& field
)
{
    List<Field<Type>> allValues(Pstream::nProcs());

    allValues[Pstream::myProcNo()] = field;

    Pstream::gatherList(allValues);

    if (Pstream::master())
    {
        field =
            ListListOps::combine<Field<Type>>
            (
                allValues,
                accessOp<Field<Type>>()
            );
    }
}

// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

template<class BasePhaseSystem>
Foam::TransitionPhaseSystem<BasePhaseSystem>::
TransitionPhaseSystem
(
    const fvMesh& mesh
)
:
    BasePhaseSystem(mesh),

    transitioningPhase_
    (
        mesh.lookupObject<phaseModel>
        (
            IOobject::groupName("alpha", this->lookup("transitioningPhase"))
        )
    ),

    velGroupPtr_(nullptr),

    scotch_(),

    distanceToCgNormalised_
    (
        IOobject
        (
            "distanceToCgNormalised",
            mesh.time().timeName(),
            mesh,
            IOobject::NO_READ,
            IOobject::AUTO_WRITE
        ),
        mesh,
        dimensionedScalar("zero", dimless, Zero)
    )
{
    read();

    if
    (
        !isA<diameterModels::velocityGroup>(transitioningPhase_.dPtr()())
    )
    {
        FatalErrorInFunction
            << "Diameter model of the transitioning phase "
            << transitioningPhase_.name()
            << " must be of type "
            << diameterModels::velocityGroup::typeName
            << endl
            << exit(FatalError);
    }
    else
    {
        velGroupPtr_ =
            &refCast<const diameterModels::velocityGroup>
            (
                transitioningPhase_.dPtr()()
            );
    }

    this->generateInterfacialModels(clusteringByMomentumEqnModels_);
    this->generateInterfacialModels(clusteringByContinuityEqnModels_);
}


// * * * * * * * * * * * * * * * * Destructor  * * * * * * * * * * * * * * * //

template<class BasePhaseSystem>
Foam::TransitionPhaseSystem<BasePhaseSystem>::
~TransitionPhaseSystem()
{}


// * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * * //

template<class BasePhaseSystem>
Foam::PtrList<Foam::surfaceScalarField>
Foam::TransitionPhaseSystem<BasePhaseSystem>::phiFs
(
    const PtrList<volScalarField>& rAUs
)
{
    PtrList<surfaceScalarField> phiFs(BasePhaseSystem::phiFs(rAUs));

    // Add the clustering force
    forAllConstIter
    (
        clusteringByMomentumEqnModelTable,
        clusteringByMomentumEqnModels_,
        clusteringByMomentumEqnModelIter
    )
    {
        const volVectorField F(clusteringByMomentumEqnModelIter()->F());
        const phaseInterface interface
        (
            *this,
            clusteringByMomentumEqnModelIter.key()
        );

        addField
        (
            interface.phase1(),
            "phiF",
            fvc::flux(rAUs[interface.phase1().index()]*F),
            phiFs
        );
        addField
        (
            interface.phase2(),
            "phiF",
          - fvc::flux(rAUs[interface.phase2().index()]*F),
            phiFs
        );
    }

    return phiFs;
}


template<class BasePhaseSystem>
Foam::PtrList<Foam::surfaceScalarField>
Foam::TransitionPhaseSystem<BasePhaseSystem>::phiFfs
(
    const PtrList<surfaceScalarField>& rAUfs
)
{
    PtrList<surfaceScalarField> phiFfs(BasePhaseSystem::phiFfs(rAUfs));

    // Add the clustering force
    forAllConstIter
    (
        clusteringByMomentumEqnModelTable,
        clusteringByMomentumEqnModels_,
        clusteringByMomentumEqnModelIter
    )
    {
        const surfaceScalarField Ff(clusteringByMomentumEqnModelIter()->Ff());
        const phaseInterface interface
        (
            *this,
            clusteringByMomentumEqnModelIter.key()
        );

        addField
        (
            interface.phase1(),
            "phiFs",
            rAUfs[interface.phase1().index()]*Ff,
            phiFfs
        );
        addField
        (
            interface.phase2(),
            "phiFf",
          - rAUfs[interface.phase2().index()]*Ff,
            phiFfs
        );
    }

    return phiFfs;
}


template<class BasePhaseSystem>
Foam::tmp<Foam::surfaceScalarField>
Foam::TransitionPhaseSystem<BasePhaseSystem>::phir
(
    const phaseModel& phase1,
    const phaseModel& phase2
) const
{
    tmp<surfaceScalarField> phir(BasePhaseSystem::phir(phase1, phase2));

    const phaseInterfaceKey key(phase1, phase2);

    if (clusteringByContinuityEqnModels_.found(key))
    {
        if (clusteringByContinuityEqnModels_[key]().type() != "none")
        {
            phir.ref() += clusteringByContinuityEqnModels_[key]().phirc(phase1);
        }
    }

    return phir;
}


template<class BasePhaseSystem>
void Foam::TransitionPhaseSystem<BasePhaseSystem>::addCompressionFlux
(
    const phaseModel& phase,
    surfaceScalarField& alphaPhi
) const
{
    BasePhaseSystem::addCompressionFlux(phase, alphaPhi);

    if (!clusteringByContinuityEqnModels_.empty())
    {
        const volScalarField& alpha = phase;

        forAll(this->phases(), phasei)
        {
            const phaseModel& phase2 = this->phases()[phasei];
            const volScalarField& alpha2 = phase2;
            const phaseInterfaceKey key(phase, phase2);

            if (&phase2 == &phase) continue;

            clusteringByContinuityEqnModelTable::const_iterator cAlpha
            (
                clusteringByContinuityEqnModels_.find(key)
            );

            if (cAlpha != clusteringByContinuityEqnModels_.end())
            {
                const surfaceScalarField phir
                (
                    clusteringByContinuityEqnModels_[key]().phirc(phase)
                );

                const word phirScheme
                (
                    "div(phir,"
                  + alpha2.name() + ',' + alpha.name()
                  + ')'
                );

                alphaPhi += fvc::flux
                (
                    -fvc::flux(-phir, alpha2, phirScheme),
                    alpha,
                    phirScheme
                );
            }
        }
    }
}


template<class BasePhaseSystem>
bool Foam::TransitionPhaseSystem<BasePhaseSystem>::read()
{
    if (BasePhaseSystem::read())
    {
        bool readOK = true;

        threshold_ = this->template lookupOrDefault<scalar>("threshold", 0.0);

        return readOK;
    }
    else
    {
        return false;
    }
}


template<class BasePhaseSystem>
void Foam::TransitionPhaseSystem<BasePhaseSystem>::solve
(
    const PtrList<volScalarField>& rAUs,
    const PtrList<surfaceScalarField>& rAUfs
)
{
    // Calculate number of resolvable structures from number concentration
    // of the transitioning phase (round off)
    label nStructures
    (
        floor
        (
            fvc::domainIntegrate
            (
                velGroupPtr_->numberConcentration()
            ).value()
        )
    );

    // Set of cells in which the transition occurs
    cellSet transitionCells
    (
        this->mesh(),
        "transitionCells",
        this->mesh().nCells()
    );

    // Sets the the transitionCells based on threshold
    forAll(transitioningPhase_, celli)
    {
        if (transitioningPhase_[celli] >= threshold_)
        {
            transitionCells.insert(celli);
        }
    }

    // Construct subset of the mesh
    fvMeshSubset transitionRegion(this->mesh());

    // Fill mesh subset with transitionCells
    transitionRegion.setLargeCellSubset(transitionCells);

    if (nStructures != 0)
    {
        // Decompose mesh subset into number of structures
        labelList regions
        (
            scotch_.decompose
            (
                transitionRegion.subMesh(),
                transitionRegion.subMesh().cellCentres(),
                nStructures
            )
        );

        // Instantiate index field
        // (negative index for cells outside transitionRegion)
        volScalarField regionField
        (
            IOobject
            (
                "regionField",
                this->mesh().time().timeName(),
                this->mesh()
            ),
            this->mesh(),
            dimensionedScalar("regionField", dimless, -1.0)
        );

        // Fill regionField using the cellMap of the transitionRegion subset
        forAll(regions, i)
        {
            regionField[transitionRegion.cellMap()[i]] = regions[i];
        }

        regionField.correctBoundaryConditions();

        // Actual volume occupied by the phase in a cell
        const scalarField alphaVol
        (
            transitioningPhase_.primitiveField()*this->mesh().V()
        );

        // Total (mesh) volume of every region
        // Map is a HashTable to objects of type T with a label key
        Map<scalar> allRegionVolume
        (
            regionSum(regionField, nStructures, this->mesh().V())
        );

        // Total (alpha-weighted) volume for every region
        Map<scalar> allRegionAlphaVolume
        (
            regionSum(regionField, nStructures, alphaVol)
        );

        // Total number of cells for every region
        Map<label> allRegionNumCells
        (
            regionSum
            (
                regionField,
                nStructures,
                labelField(this->mesh().nCells(), 1.0)
            )
        );

        // Centre of gravity for every region,
        Map<vector> allRegionCentreOfGravity(allRegionVolume.size());

        forAllIter(Map<scalar>, allRegionVolume, vIter)
        {
            // Currently uses uniform cell weights
            volScalarField weightField
            (
                IOobject
                (
                    "regionIFld",
                    this->mesh().time().timeName(),
                    this->mesh(),
                    IOobject::NO_READ,
                    IOobject::NO_WRITE
                ),
                this->mesh(),
                dimensionedScalar("zero", dimless, 0)
            );

            forAll(weightField, celli)
            {
                if (allRegionVolume.find(regionField[celli]) == vIter)
                {
                    weightField[celli] = 1.0;
                }
            }

            vectorField cellCentres(this->mesh().C());
            scalarField cellVolumes(this->mesh().V());

            // Combine onto master
            combineFields(weightField);
            combineFields(cellCentres);
            combineFields(cellVolumes);

            if (Pstream::master())
            {
                vector phaseByCSum
                (
                    sum(vectorField(weightField*cellCentres*cellVolumes))
                );

                scalar phaseSum(sum(scalarField(weightField*cellVolumes)));
                allRegionCentreOfGravity.insert
                (
                    vIter.key(),
                    phaseByCSum/phaseSum
                );
            }
        }

        Pstream::mapCombineScatter(allRegionCentreOfGravity);

        distanceToCgNormalised_ = 0;

        forAll(distanceToCgNormalised_, celli)
        {
            if (allRegionCentreOfGravity.found(regionField[celli]))
            {
                distanceToCgNormalised_[celli] =
                    mag
                    (
                        allRegionCentreOfGravity[regionField[celli]]
                      - this->mesh().C()[celli]
                    );
            }
        }

        distanceToCgNormalised_ /= gMax(distanceToCgNormalised_);

        distanceToCgNormalised_.correctBoundaryConditions();
    }

    BasePhaseSystem::solve(rAUs, rAUfs);
}


// ************************************************************************* //
