/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2018-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

Class
    Foam::TransitionPhaseSystem

Description
    Class which provides basic functionality for the transition of a disperse
    phase between its smeared (statistical) and resolved representation and vice
    versa, based on a suitable diameter threshold that is determined by the
    mesh.

    An important requirement for this bidirectional transition is consistency
    between the modeled, i.e. subgrid size distribution (class
    populationBalanceModel) and the detected size distribution in the resolved
    representation with respect to number and size of the disperse phase
    elements.

    The transition from a resolved to a smeared representation is
    straightforward. The size of the resolved structures needs to be detected
    before they can be assigned to the corresponding sizeGroup. The phase
    fraction field then smears by numerical diffusion. Size, position and number
    of the structures are known.

    In the opposite direction, only size and number of the potentially
    resolvable structures are known, but not their position for which only
    statistical information (the phase fraction field) is available. To form
    the resolved gas structures, this phase fraction field must be split
    such that the expected number and size constraint is fulfilled, perhaps by
    means of a water-shedding algorithm. The subsequent redistribution requires
    a force/flux, calculated by the clusteringBy~EqnModel class.

    A large portion of the detection functionality stems from the pre-existing
    regionSizeDistribution functionObject.

    The algorithm for calculating the centre of gravity of each structure stems
    from the pre-existing phaseCG functionObject.

Usage
    \table
        Property           | Description           | Required    | Default value
        transitioningPhase | Phase subject to transition | yes   | none
        patches            | Patches to walk from  | yes         | none
        threshold          | Clip value            | yes         | none
    \endtable

See also
    Foam::clusteringByMomentumEqnModel
    Foam::PopulationBalancePhaseSystem

SourceFiles
    TransitionPhaseSystem.C

\*---------------------------------------------------------------------------*/

#ifndef TransitionPhaseSystem_H
#define TransitionPhaseSystem_H

#include "phaseSystem.H"
#include "HashPtrTable.H"
#include "regionSplit.H"
#include "HZDRScotchDecomp.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

class blendingMethod;
class blendedClusteringByMomentumEqnModel;
class clusteringByContinuityEqnModel;

namespace diameterModels
{
    class velocityGroup;
}

/*---------------------------------------------------------------------------*\
                    Class TransitionPhaseSystem Declaration
\*---------------------------------------------------------------------------*/

template<class BasePhaseSystem>
class TransitionPhaseSystem
:
    public BasePhaseSystem
{
protected:

    // Protected Typedefs

        typedef HashTable
        <
            autoPtr<blendedClusteringByMomentumEqnModel>,
            phaseInterfaceKey,
            phaseInterfaceKey::hash
        > clusteringByMomentumEqnModelTable;


        typedef HashTable
        <
            autoPtr<clusteringByContinuityEqnModel>,
            phaseInterfaceKey,
            phaseInterfaceKey::hash
        > clusteringByContinuityEqnModelTable;


private:

    // Private Data

        //- Phase which is subject to transition
        const phaseModel& transitioningPhase_;

        const diameterModels::velocityGroup* velGroupPtr_;

        //- Decomposition method
        HZDRScotchDecomp scotch_;

        //- Threshold value to determine transitionRegion
        scalar threshold_;

        //- Distance of a cell(-centre) to the centre of gravity of the region
        //  it belongs to
        volScalarField distanceToCgNormalised_;

        //- Models for clustering by momentum equation
        clusteringByMomentumEqnModelTable clusteringByMomentumEqnModels_;

        //- Models for clustering by continuity equation
        clusteringByContinuityEqnModelTable clusteringByContinuityEqnModels_;


    // Private Member Functions

        template<class Type>
        Map<Type> regionSum
        (
            const volScalarField& regionField,
            const label nStructures,
            const Field<Type>& fld
        ) const;

        //- Combine fields from all processor domains into single field
        template<class Type>
        void combineFields(Field<Type>& field);


public:

    // Constructors

        //- Construct from fvMesh
        TransitionPhaseSystem(const fvMesh&);


    //- Destructor
    virtual ~TransitionPhaseSystem();


    // Member Functions

        //- Add force for formation of resolved structures
        virtual PtrList<surfaceScalarField> phiFs
        (
            const PtrList<volScalarField>& rAUs
        );

        //- Add force for formation of resolved structures for face-based
        //  algorithm
        virtual PtrList<surfaceScalarField> phiFfs
        (
            const PtrList<surfaceScalarField>& rAUfs
        );

        //- Return relative flux for formation of resolved structures by
        //  continuity equation
        virtual tmp<surfaceScalarField> phir
        (
            const phaseModel& phase1,
            const phaseModel& phase2
        ) const;

        //- Add clustering flux
        virtual void addCompressionFlux
        (
            const phaseModel& phase,
            surfaceScalarField& alphaPhi
        ) const;

        //- Read base phaseProperties dictionary
        virtual bool read();

        //- Calculate isolated regions and their centres of gravity
        virtual void solve
        (
            const PtrList<volScalarField>& rAUs,
            const PtrList<surfaceScalarField>& rAUfs
        );
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#ifdef NoRepository
    #include "TransitionPhaseSystem.C"
#endif

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
