/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2021-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

Class
    Foam::diameterModels::cudaUtilities::CUDAModelRunner

Description
    Object runs CUDA kernel models.

SourceFiles
    CUDAModelRunner.C

\*---------------------------------------------------------------------------*/

#ifndef CUDAModelRunner_H
#define CUDAModelRunner_H

#include <vector>
#include "CUDAPopulationBalance.H"
#include <iostream>

using namespace std;

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{
namespace diameterModels
{
namespace cudaUtilities
{

/*---------------------------------------------------------------------------*\
                       Class CUDAModelRunner Declaration
\*---------------------------------------------------------------------------*/

template <typename T>
__global__ void Run(
    const T kernel,
    CUDAPopulationBalance popBal,
    double* field,
    int i, int j,
    int domainSize)
{
    // Specific cell in the domain
    int index = threadIdx.x;

    // Loop over all the cells in the domain
    while(index<domainSize)
    {
        kernel.computeRate(popBal, field, i, j, index, domainSize);
        // Move to the next cell
        index += blockDim.x;
    }
}

template<class Model>
class CUDAModelRunner
{
    public:
        CUDAPopulationBalance popBal_;
        Model model_;
    // Member Functions
    // Protected Data
    protected:

        //- Results field
        double* resultsField_;
        std::size_t resultsFieldSize_;

        //- Stream array
        std::vector<cudaStream_t> streams_;

        //- Size of the domain
        std::size_t domainSize_;

        //- Counter for computer pairs
        std::size_t readPos_;

        //- List of pair need to be computed
        std::vector<std::pair<int, int>> pairs_;

        //- Calculation of CUDA block size
        std::size_t getBlockSize(std::size_t domainSize)
        {
            // The function will return round up to the nearest power 2 number.
            // The number is limited to 1024 for CUDA architecture purposes.
            return min((int)pow(2, ceil(log(domainSize)/log(2))), 1024);
        }

        cudaStream_t getStream(int position)
        {
            return streams_[position%streams_.size()];
        }

        cudaStream_t getNextStream()
        {
            return streams_[readPos_%streams_.size()];
        }

        std::size_t getBufferIndex(std::size_t position)
        {
            return position%(streams_.size()+1);
        }

        double* getMemory(std::size_t position)
        {
            std::size_t index = getBufferIndex(position);
            return resultsField_ + index*domainSize_;
        }

            void queueNextPair()
        {
            queuePair(readPos_+streams_.size());
        }

        void queuePair(int position)
        {
            if(position<pairs_.size())
            {
                int j = pairs_[position].second;
                int i = pairs_[position].first;
                //Calling the kernels
                int blockSize = getBlockSize(domainSize_);
                // Call to global CUDA function to start the computation.
                Run<<<1, blockSize, 0, getStream(position)>>>(
                    model_,
                    popBal_,
                    getMemory(position),
                    i,
                    j,
                    domainSize_
                );
            }
        }

    public:

        void calculateRate
        (
            CUDAPopulationBalance popBal,
            int domainSize,
            const std::vector<std::pair<int, int>> &pairs
        )
        {
            popBal_ = popBal;
            domainSize_ = domainSize;
            pairs_ = pairs;

            //Memory allocation
            int coalescenceFieldSize = (streams_.size()+1)*domainSize_;
            if(this->resultsFieldSize_!=coalescenceFieldSize)
            {
                cudaMallocManaged
                (
                    &this->resultsField_,
                    coalescenceFieldSize*sizeof(double)
                );
                this->resultsFieldSize_ = coalescenceFieldSize;
            }

            // Reset position
            readPos_ = 0;

            // Queue pairs
            for(int i = 0; i < streams_.size(); i++)
            {
                queuePair(i);
            }
        }

        OFMemoryManagedField getRate
        (
            int i,
            int j
        )
        {
            // Wait for the event that signals the end of computation
            // for specific pair
            cudaError_t cudaerr = cudaStreamSynchronize(getNextStream());
            if (cudaerr != cudaSuccess)
            {
                printf("Kernel launch failed with error \"%s\".\n",
                    cudaGetErrorString(cudaerr));
            }

            // Start the computation if a new pair
            queueNextPair();
            return OFMemoryManagedField(getMemory(readPos_++), domainSize_);
        }

    // Constructor
        CUDAModelRunner(std::size_t numStreams, Model model)
          : streams_(numStreams), resultsFieldSize_(0), model_(model)
        {
            // Creation of streams
            for (std::size_t i = 0; i < streams_.size(); i++)
            {
                cudaStreamCreate(&streams_[i]);
            }

        }

    //- Destructor
        ~CUDAModelRunner()
        {
            // Destruction of streams
            for (std::size_t i = 0; i < streams_.size(); i++)
            {
                cudaStreamDestroy(streams_[i]);
            }
        }
};

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace cudaUtilities
} // End namespace diameterModels
} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
