/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2021-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf
    and some other institution

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "CUDALiaoSupportModel.H"
#include "OFMemoryManagedField.H"
#include <iostream>
#include "CUDAModelRunner.H"

using namespace std;
using namespace Foam::diameterModels::cudaUtilities;


// * * * * * * * * * * * * * * * * * * CUDA  * * * * * * * * * * * * * * * * //


__device__ double sqr(double x)
{
    return pow(x, 2);
}


__device__ double mag(double x)
{
    return sqrt(pow(x, 2));
}


__device__ int pos(double x)
{
    return (x > 0) ? 1: 0;
}


__device__ int pos0(double x)
{
    return (x >= 0) ? 1: 0;
}


struct Foam::diameterModels::binaryBreakupModels::CUDALiaoSupportModel::
CUDALiaoBinaryBreakupModelPimpl
{
    double BTurb_;
    double BShear_;
    double BEddy_;
    double BFric_;
    bool turbulence_;
    bool laminarShear_;
    bool turbulentShear_;
    bool interfacialFriction_;

    OFMemoryManagedField kolmogorovLengthScale_;
    OFMemoryManagedField shearStrainRate_;
    OFMemoryManagedField eddyStrainRate_;
    OFMemoryManagedField UTerminal_;
    OFMemoryManagedField Cd_;

    __device__ void computeRate(CUDAPopulationBalance popBal, double* resultField, int i, int j, int index, int domainSize) const
    {
        resultField[index] = 0.0;

        double dk = cbrt(pow(popBal.sizeGroupsdSph()[j], 3) - pow(popBal.sizeGroupsdSph()[i], 3));

        // Eq. (34)
        double tauCrit1 =
        (
            6*popBal.sigmaWithContinuousPhaseI()[j*domainSize+index]/popBal.sizeGroupsdSph()[j]
           *(pow(popBal.sizeGroupsdSph()[i]/popBal.sizeGroupsdSph()[j], 2) + pow(dk/popBal.sizeGroupsdSph()[j],2 ) - 1)
        );

        // Eq. (35)
        double tauCrit2 =
        (
            popBal.sigmaWithContinuousPhaseI()[j*domainSize+index]/min(dk, popBal.sizeGroupsdSph()[i])
        );

        // Eq. (33)
        double tauCrit = max(tauCrit1, tauCrit2);

        // Eqs. (32) and (36)
        if(turbulence_)
        {
            double tauTurb =
            (
                pos(popBal.sizeGroupsdSph()[j] - kolmogorovLengthScale_[index])*BTurb_*popBal.continuousPhaseRho()[index]
               *sqr(cbrt(popBal.continuousTurbulenceEpsilon()[index]*popBal.sizeGroupsdSph()[j]))
            );

            resultField[index] +=
                pos(tauTurb - tauCrit)*1/popBal.sizeGroupsdSph()[j]
               *sqrt(mag(tauTurb - tauCrit)/popBal.continuousPhaseRho()[index])/popBal.sizeGroupsX()[j];
        }

        if (laminarShear_)
        {
            double tauShear
            (
                BShear_*popBal.continuousPhaseThermoMu()[index]*shearStrainRate_[index]
            );

            resultField[index] +=
                pos(tauShear - tauCrit)*1/popBal.sizeGroupsdSph()[j]
               *sqrt(mag(tauShear - tauCrit)/popBal.continuousPhaseRho()[index])/popBal.sizeGroupsX()[j];
        }

        if(turbulentShear_)
        {
            double tauEddy =
            (
                pos0(kolmogorovLengthScale_[index] - popBal.sizeGroupsdSph()[j])
               *BEddy_*popBal.continuousPhaseThermoMu()[index]*eddyStrainRate_[index]
            );

            resultField[index] +=
                pos(tauEddy - tauCrit)*1/popBal.sizeGroupsdSph()[j]
               *sqrt(mag(tauEddy - tauCrit)/popBal.continuousPhaseRho()[index])/popBal.sizeGroupsX()[j];
        }

        if (interfacialFriction_)
        {
            double tauFric =
            (
                BFric_*0.5*popBal.continuousPhaseRho()[index]*sqr(UTerminal_[j])*Cd_[j]
            );

            resultField[index] +=
                pos(tauFric - tauCrit)*1/popBal.sizeGroupsdSph()[j]
               *sqrt(mag(tauFric - tauCrit)/popBal.continuousPhaseRho()[index])/popBal.sizeGroupsX()[j];
        }
    }
};


class
Foam::diameterModels::binaryBreakupModels::CUDALiaoSupportModel::
CUDALiaoBinaryBreakupRunnerPimpl
: public CUDAModelRunner<CUDALiaoBinaryBreakupModelPimpl>
{
public:
    CUDALiaoBinaryBreakupRunnerPimpl(int numStreams, CUDALiaoBinaryBreakupModelPimpl model)
     : CUDAModelRunner<CUDALiaoBinaryBreakupModelPimpl>(numStreams, model)
    {}
};


// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

Foam::diameterModels::binaryBreakupModels::CUDALiaoSupportModel::
CUDALiaoSupportModel(int numStreams)
: pimpl(new CUDALiaoBinaryBreakupRunnerPimpl(numStreams, CUDALiaoBinaryBreakupModelPimpl()))
{}

// * * * * * * * * * * * * * * * * Destructor  * * * * * * * * * * * * * * * //

Foam::diameterModels::binaryBreakupModels::CUDALiaoSupportModel::
~CUDALiaoSupportModel() = default;

// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //


OFMemoryManagedField
Foam::diameterModels::binaryBreakupModels::CUDALiaoSupportModel::
getKolmogorovLengthScale(int domainSize)
{
    pimpl->model_.kolmogorovLengthScale_.resize(domainSize);
    return pimpl->model_.kolmogorovLengthScale_;
}


OFMemoryManagedField
Foam::diameterModels::binaryBreakupModels::CUDALiaoSupportModel::
getShearStrainRate(int domainSize)
{
    pimpl->model_.shearStrainRate_.resize(domainSize);
    return pimpl->model_.shearStrainRate_;
}


OFMemoryManagedField
Foam::diameterModels::binaryBreakupModels::CUDALiaoSupportModel::
getEddyStrainRate(int domainSize)
{
    pimpl->model_.eddyStrainRate_.resize(domainSize);
    return pimpl->model_.eddyStrainRate_;
}


OFMemoryManagedField
Foam::diameterModels::binaryBreakupModels::CUDALiaoSupportModel::
getUTerminal(int numberOfGroups)
{
    pimpl->model_.UTerminal_.resize(numberOfGroups);
    return pimpl->model_.UTerminal_;
}


OFMemoryManagedField
Foam::diameterModels::binaryBreakupModels::CUDALiaoSupportModel::
getCd(int numberOfGroups)
{
    pimpl->model_.Cd_.resize(numberOfGroups);
    return pimpl->model_.Cd_;
}


void
Foam::diameterModels::binaryBreakupModels::CUDALiaoSupportModel::
calculateBinaryBreakupRate
(
    CUDAPopulationBalance popBal,
    int currentDomainSize,
    const std::vector<std::pair<int, int>> &pairs
)
{
    pimpl->calculateRate
    (
        popBal,
        currentDomainSize,
        pairs
    );
}


void
Foam::diameterModels::binaryBreakupModels::CUDALiaoSupportModel::
setModelVariables
(
    bool turbulence_,
    bool laminarShear_,
    bool turbulentShear_,
    bool interfacialFriction_,
    double BTurb_,
    double BShear_,
    double BEddy_,
    double BFric_
)
{
    pimpl->model_.turbulence_ = turbulence_;
    pimpl->model_.laminarShear_ = laminarShear_;
    pimpl->model_.turbulentShear_ = turbulentShear_;
    pimpl->model_.interfacialFriction_ = interfacialFriction_;
    pimpl->model_.BTurb_ = BTurb_;
    pimpl->model_.BShear_ = BShear_;
    pimpl->model_.BEddy_ = BEddy_;
    pimpl->model_.BFric_ = BFric_;
}


OFMemoryManagedField
Foam::diameterModels::binaryBreakupModels::CUDALiaoSupportModel::
getBinaryBreakupRate
(
    int i,
    int j
)
{
    return pimpl->getRate
    (
        i,
        j
    );
}


// ************************************************************************* //
