/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2021-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "CUDACoulaloglouTavlarides.H"
#include "addToRunTimeSelectionTable.H"
#include "CUDACoulaloglouTavlaridesSupportModel.H"
#include <vector>

using namespace Foam::diameterModels::cudaUtilities;

// * * * * * * * * * * * * * * Static Data Members * * * * * * * * * * * * * //

namespace Foam
{
namespace diameterModels
{
namespace coalescenceModels
{
    defineTypeNameAndDebug(CUDACoulaloglouTavlarides, 0);
    addToRunTimeSelectionTable
    (
        coalescenceModel,
        CUDACoulaloglouTavlarides,
        dictionary
    );
}
}
}

using Foam::constant::mathematical::pi;

// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

Foam::diameterModels::coalescenceModels::CUDACoulaloglouTavlarides::
CUDACoulaloglouTavlarides
(
    const populationBalanceModel& popBal,
    const dictionary& dict
)
:
    CoulaloglouTavlarides(popBal, dict),
    numberOfStreams_(dict.lookupOrDefault("numberOfStreams", 64)),
    cudaModelObject(numberOfStreams_)
{}

// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

void
Foam::diameterModels::coalescenceModels::CUDACoulaloglouTavlarides::
precompute()
{
    int domainSize = popBal_.alphas().size();
    int numberOfGroups = popBal_.sizeGroups().size();

    OFMemoryManagedField alphaPtr =
        cudaPopBal_.setAlphas(domainSize);

    OFMemoryManagedField popBalContinuousTurbulenceEpsilonPtr =
        cudaPopBal_.setContinuousTurbulenceEpsilon(domainSize);

    OFMemoryManagedField sizeGroupsdSphPtr =
        cudaPopBal_.setSizeGroupsdSph(numberOfGroups);

    OFMemoryManagedField sizeGroupsXPtr =
        cudaPopBal_.setSizeGroupsX(numberOfGroups);

    OFMemoryManagedField popBalContinuousPhaseRhoPtr =
        cudaPopBal_.setContinuousPhaseRho(domainSize);

    OFMemoryManagedField popBalContinuousPhaseThermoMuPtr =
        cudaPopBal_.setContinuousPhaseThermoMu(domainSize);

    OFMemoryManagedField popBalSigmaWithContinuousPhaseIPtr =
        cudaPopBal_.setSigmaWithContinuousPhaseI
        (
            numberOfGroups,
            domainSize
        );


    // Queue pairs to be executed
    std::vector<std::pair<int, int>> pairs;
    List<Pair<label>> cPairs(popBal_.coalescencePairs());
    pairs.reserve(cPairs.size());
    forAll(cPairs, hh)
    {
        label i(cPairs[hh].first());
        label j(cPairs[hh].second());
        pairs.push_back(std::make_pair(j, i));
    }

    forAll(popBal_.alphas(), i)
    {
        alphaPtr[i] = popBal_.alphas()[i];
    }

    volScalarField epsilon(popBal_.continuousTurbulence().epsilon());
    forAll(epsilon, i)
    {
        popBalContinuousTurbulenceEpsilonPtr[i] = epsilon[i];
    }

    forAll(popBal_.sizeGroups(), i)
    {
        const sizeGroup& fi = popBal_.sizeGroups()[i];
        sizeGroupsdSphPtr[i] = fi.dSph().value();
        sizeGroupsXPtr[i] = fi.x().value();
    }

    forAll(popBal_.continuousPhase().rho()(), i)
    {
        popBalContinuousPhaseRhoPtr[i] =
            popBal_.continuousPhase().rho()()[i];
    }

    forAll(popBal_.continuousPhase().thermo().mu()(), i)
    {
        popBalContinuousPhaseThermoMuPtr[i] =
            popBal_.continuousPhase().thermo().mu()()[i];
    }

    forAll(popBal_.sizeGroups(), i)
    {
        const sizeGroup& fi = popBal_.sizeGroups()[i];
        scalarField phaseI = popBal_.sigmaWithContinuousPhase(fi.phase())();
        forAll(phaseI, j)
        {
            popBalSigmaWithContinuousPhaseIPtr[i*domainSize + j] =
                phaseI[j];
        }
    }


    cudaModelObject.setModelVariables
    (
        C1_.value(),
        C2_.value()
    );

    cudaModelObject.calculateCoalescenceRate
    (
        cudaPopBal_,
        domainSize,
        pairs
    );
}


void
Foam::diameterModels::coalescenceModels::CUDACoulaloglouTavlarides::
addToCoalescenceRate
(
    volScalarField& coalescenceRate,
    const label i,
    const label j
)
{
    int domainSize = popBal_.alphas().size();
    OFMemoryManagedField cudaCoalescenceRate =
        cudaModelObject.getCoalescenceRate(i, j);
    for (int h = 0 ; h < domainSize; h++)
    {
        coalescenceRate[h] += cudaCoalescenceRate[h];
    }
}

// ************************************************************************* //
