/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2021-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "CUDALiaoCoalescence.H"
#include "addToRunTimeSelectionTable.H"
#include "fvcGrad.H"
#include "dragModel.H"
#include "phaseSystem.H"
#include <vector>

using namespace Foam::diameterModels::cudaUtilities;

// * * * * * * * * * * * * * * Static Data Members * * * * * * * * * * * * * //

namespace Foam
{
namespace diameterModels
{
namespace coalescenceModels
{
    defineTypeNameAndDebug(CUDALiaoCoalescence, 0);
    addToRunTimeSelectionTable
    (
        coalescenceModel,
        CUDALiaoCoalescence,
        dictionary
    );
}
}
}

using Foam::constant::mathematical::pi;


// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

Foam::diameterModels::coalescenceModels::CUDALiaoCoalescence::
CUDALiaoCoalescence
(
    const populationBalanceModel& popBal,
    const dictionary& dict
)
:
    LiaoCoalescence(popBal, dict),
    numberOfStreams_(dict.lookupOrDefault("numberOfStreams", 64)),
    cudaModelObject(numberOfStreams_)

{}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

void Foam::diameterModels::coalescenceModels::CUDALiaoCoalescence::precompute()
{
    LiaoCoalescence::precompute();

    int domainSize = popBal_.alphas().size();
    int numberOfGroups = popBal_.sizeGroups().size();

    // Memory allocation and management
    OFMemoryManagedField popBalContinuousTurbulenceEpsilonPtr =
        cudaPopBal_.setContinuousTurbulenceEpsilon(domainSize);

    OFMemoryManagedField sizeGroupsdSphPtr =
        cudaPopBal_.setSizeGroupsdSph(numberOfGroups);

    OFMemoryManagedField popBalContinuousPhaseRhoPtr =
        cudaPopBal_.setContinuousPhaseRho(domainSize);

    OFMemoryManagedField popBalContinuousPhaseThermoMuPtr =
        cudaPopBal_.setContinuousPhaseThermoMu(domainSize);

    OFMemoryManagedField popBalSigmaWithContinuousPhaseIPtr =
        cudaPopBal_.setSigmaWithContinuousPhaseI
        (
            numberOfGroups,
            domainSize
        );

    OFMemoryManagedField kolmogorovLengthScale =
        cudaModelObject.getKolmogorovLengthScale(domainSize);

    OFMemoryManagedField shearStrainRate =
        cudaModelObject.getShearStrainRate(domainSize);

    OFMemoryManagedField eddyStrainRate =
        cudaModelObject.getEddyStrainRate(domainSize);

    OFMemoryManagedField UTerminal =
        cudaModelObject.getUTerminal(numberOfGroups);

    OFMemoryManagedField Cd =
        cudaModelObject.getCd(numberOfGroups);

    OFMemoryManagedField dCrit =
        cudaModelObject.getdCrit(domainSize);

    OFMemoryManagedField CPack =
        cudaModelObject.getCPack(domainSize);

    // Queue pairs to be executed
    std::vector<std::pair<int, int>> pairs;
    List<Pair<label>> cPairs(popBal_.coalescencePairs());
    pairs.reserve(popBal_.nCorr()*cPairs.size());
    for (int corr = 0; corr < popBal_.nCorr(); corr++)
    {
        forAll(cPairs, hh)
        {
            label i(cPairs[hh].first());
            label j(cPairs[hh].second());
            pairs.push_back(std::make_pair(j, i));
        }
    }

    // Memory copy
    forAll(uTerminal_, i)
    {
        UTerminal[i] = uTerminal_[i].value();
    }

    forAll(Cd_, i)
    {
        Cd[i] = Cd_[i].value();
    }

    forAll(shearStrainRate_, i)
    {
        shearStrainRate[i] = shearStrainRate_[i];
    }

    forAll(eddyStrainRate_, i)
    {
        eddyStrainRate[i] = eddyStrainRate_[i];
    }

    forAll(popBal_.continuousTurbulence().epsilon()(), i)
    {
        popBalContinuousTurbulenceEpsilonPtr[i] =
            popBal_.continuousTurbulence().epsilon()()[i];
    }

    forAll(popBal_.sizeGroups(), i)
    {
        const sizeGroup& fi = popBal_.sizeGroups()[i];
        sizeGroupsdSphPtr[i] = fi.dSph().value();
    }

    forAll(popBal_.continuousPhase().rho()(), i)
    {
        popBalContinuousPhaseRhoPtr[i] =
            popBal_.continuousPhase().rho()()[i];
    }

    forAll(popBal_.continuousPhase().thermo().mu()(), i)
    {
        popBalContinuousPhaseThermoMuPtr[i] =
            popBal_.continuousPhase().thermo().mu()()[i];
    }

    forAll(popBal_.sizeGroups(), i)
    {
        const sizeGroup& fi = popBal_.sizeGroups()[i];
        scalarField phaseI = popBal_.sigmaWithContinuousPhase(fi.phase())();
        forAll(phaseI, j)
        {
            popBalSigmaWithContinuousPhaseIPtr[i*domainSize + j] =
                phaseI[j];
        }
    }

    forAll(kolmogorovLengthScale_, i)
    {
        kolmogorovLengthScale[i] = kolmogorovLengthScale_[i];
    }

    forAll(dCrit_, i)
    {
        dCrit[i] = dCrit_[i];
    }

    forAll(CPack_, i)
    {
        CPack[i] = CPack_[i];
    }

    cudaModelObject.setModelVariables
    (
        PMax_.value(),
        AH_.value(),
        CEff_.value(),
        CTurb_.value(),
        CBuoy_.value(),
        CShear_.value(),
        CEddy_.value(),
        CWake_.value(),
        turbulence_,
        buoyancy_,
        laminarShear_,
        eddyCapture_,
        wakeEntrainment_,
        CPackMax_.value()
    );

    // Kernel execution
    cudaModelObject.calculateCoalescenceRate
    (
        cudaPopBal_,
        domainSize,
        pairs
    );
}


void Foam::diameterModels::coalescenceModels::CUDALiaoCoalescence::
addToCoalescenceRate
(
    volScalarField& coalescenceRate,
    const label i,
    const label j
)
{
    int domainSize = popBal_.alphas().size();
    OFMemoryManagedField cudaCoalescenceRate =
        cudaModelObject.getCoalescenceRate(i, j);
    for (int h = 0 ; h < domainSize; h++)
    {
        coalescenceRate[h] += cudaCoalescenceRate[h];
    }
}

// ************************************************************************* //
