/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2021-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf
    and some other institution

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "CUDALiaoCoalescenceSupportModel.H"
#include "OFMemoryManagedField.H"
#include <iostream>
#include "CUDAModelRunner.H"

using namespace std;
using namespace Foam::diameterModels::cudaUtilities;

// * * * * * * * * * * * * * * * * * * PIMPL * * * * * * * * * * * * * * * * //


__device__ double sqr_coal(double x)
{
    return pow(x, 2);
}


__device__ double mag_coal(double x)
{
    return sqrt(pow(x, 2));
}


__device__ int neg_coal(double x)
{
    return (x < 0) ? 1: 0;
}

__device__ double max_coal(double a, double b)
{
    return (a > b) ? a: b;
}


__device__ int pos_coal(double x)
{
    return (x > 0) ? 1: 0;
}


__device__ int pos0_coal(double x)
{
    return (x >= 0) ? 1: 0;
}


__device__ double pow6_coal(double x)
{
    return pow(x, 6);
}


struct Foam::diameterModels::coalescenceModels::CUDALiaoCoalescenceSupportModel::
CUDALiaoCoalescenceModelPimpl
{
    double PMax_;
    double AH_;
    double CEff_;
    double CTurb_;
    double CBuoy_;
    double CShear_;
    double CEddy_;
    double CWake_;
    double CPackMax_;
    bool turbulence_;
    bool buoyancy_;
    bool laminarShear_;
    bool eddyCapture_;
    bool wakeEntrainment_;

    OFMemoryManagedField kolmogorovLengthScale_;
    OFMemoryManagedField shearStrainRate_;
    OFMemoryManagedField eddyStrainRate_;
    OFMemoryManagedField uTerminal_;
    OFMemoryManagedField Cd_;
    OFMemoryManagedField dCrit_;
    OFMemoryManagedField CPack_;

    __device__ void computeRate(CUDAPopulationBalance popBal, double* resultField, int i, int j, int index, int domainSize) const
    {
        resultField[index] = 0.0;

        double pi = 3.141592654f;

        double dEq = (2*popBal.sizeGroupsdSph()[i]*popBal.sizeGroupsdSph()[j]/(popBal.sizeGroupsdSph()[i] + popBal.sizeGroupsdSph()[j]));
        double Aij = (pi*0.25*sqr_coal(popBal.sizeGroupsdSph()[i] + popBal.sizeGroupsdSph()[j]));

        double uRelTurb_ = 0.0;
        if (turbulence_)
        {
            uRelTurb_ =
                CTurb_*sqrt(2.0)
               *sqrt(sqr_coal(cbrt(popBal.sizeGroupsdSph()[i])) + sqr_coal(cbrt(popBal.sizeGroupsdSph()[j])))
               *cbrt(popBal.continuousTurbulenceEpsilon()[index]);
        }

        double uRelBuoy_ = 0;
        if (buoyancy_)
        {
            uRelBuoy_ = CBuoy_*mag_coal(uTerminal_[i] - uTerminal_[j]);
        }

        double uRelShear_ = 0;
        if (laminarShear_)
        {
            uRelShear_ = CShear_*0.5/pi*(popBal.sizeGroupsdSph()[i] + popBal.sizeGroupsdSph()[j])*shearStrainRate_[index];
        }

        // Eq. (18')
        double collisionEfficiency =
        (
            neg_coal(kolmogorovLengthScale_[index] - (popBal.sizeGroupsdSph()[i] + popBal.sizeGroupsdSph()[j]))
           *exp
            (
              - CEff_*sqrt
                (
                    popBal.continuousPhaseRho()[index]*dEq
                   /popBal.sigmaWithContinuousPhaseI()[i*domainSize+index]
                   *sqr_coal(max_coal(uRelTurb_, max_coal(uRelBuoy_, uRelShear_)))
                )
            )
          + pos0_coal(kolmogorovLengthScale_[index] - (popBal.sizeGroupsdSph()[i] + popBal.sizeGroupsdSph()[j]))
           *exp
            (
              - 3*popBal.continuousPhaseThermoMu()[index]*dEq*eddyStrainRate_[index]
               /(4*popBal.sigmaWithContinuousPhaseI()[i*domainSize+index])
               *log
                (
                    cbrt
                    (
                        pi*popBal.sigmaWithContinuousPhaseI()[i*domainSize+index]*sqr_coal(dEq)
                       /(32*AH_)
                    )
                )
            )
        );

        // Eq. (17)
        if (turbulence_)
        {
            resultField[index] +=
                neg_coal(kolmogorovLengthScale_[index] - (popBal.sizeGroupsdSph()[i] + popBal.sizeGroupsdSph()[j]))
               *CPack_[index]*Aij*uRelTurb_*collisionEfficiency;
        }

        // Eq. (18)
        if (buoyancy_)
        {
            resultField[index] += CPack_[index]*0.5*Aij*uRelBuoy_*collisionEfficiency;
        }

        // Eq. (21)
        if (laminarShear_)
        {
            resultField[index] += CPack_[index]*0.5*Aij*uRelShear_*collisionEfficiency;
        }

        // Eq. (22)
        if (eddyCapture_)
        {
            double uRelEddy =
            (
                CEddy_*0.5/pi*(popBal.sizeGroupsdSph()[i] + popBal.sizeGroupsdSph()[j])*eddyStrainRate_[index]
            );

            resultField[index] +=
                pos0_coal(kolmogorovLengthScale_[index] - (popBal.sizeGroupsdSph()[i] + popBal.sizeGroupsdSph()[j]))
               *CPack_[index]*0.5*Aij*uRelEddy*collisionEfficiency;
        }

        // Eq. (25)
        if (wakeEntrainment_)
        {
            // Eq. (24) for bubble i
            double uRelWakeI =
            (
                CWake_*uTerminal_[i]*cbrt(Cd_[i])
            );

            // Eq. (24) for bubble j
            double uRelWakeJ =
            (
                CWake_*uTerminal_[j]*cbrt(Cd_[j])
            );

            resultField[index]+=
                CPack_[index]*0.125*pi
               *(
                    sqr_coal(popBal.sizeGroupsdSph()[i])*uRelWakeI
                   *pos0_coal(popBal.sizeGroupsdSph()[i] - 0.5*dCrit_[index])
                   *(
                        pow6_coal(popBal.sizeGroupsdSph()[i] - 0.5*dCrit_[index])
                       /(pow6_coal(popBal.sizeGroupsdSph()[i] - 0.5*dCrit_[index]) + pow6_coal(0.5*dCrit_[index]))
                    )
                  + sqr_coal(popBal.sizeGroupsdSph()[j])*uRelWakeJ
                   *pos0_coal(popBal.sizeGroupsdSph()[j] - 0.5*dCrit_[index])
                   *(
                        pow6_coal(popBal.sizeGroupsdSph()[j] - 0.5*dCrit_[index])
                      /(pow6_coal(popBal.sizeGroupsdSph()[j] - 0.5*dCrit_[index]) + pow6_coal(0.5*dCrit_[index]))
                    )
                );
        }
    }
};

class
Foam::diameterModels::coalescenceModels::CUDALiaoCoalescenceSupportModel::
CUDALiaoCoalescenceRunnerPimpl
: public CUDAModelRunner<CUDALiaoCoalescenceModelPimpl>
{
public:
    CUDALiaoCoalescenceRunnerPimpl(int numStreams, CUDALiaoCoalescenceModelPimpl model)
     : CUDAModelRunner<CUDALiaoCoalescenceModelPimpl>(numStreams, model)
    {}
};

// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

Foam::diameterModels::coalescenceModels::CUDALiaoCoalescenceSupportModel::
CUDALiaoCoalescenceSupportModel(int numStreams)
  : pimpl(new CUDALiaoCoalescenceRunnerPimpl(numStreams, CUDALiaoCoalescenceModelPimpl()))
{}

// * * * * * * * * * * * * * * * * Destructor  * * * * * * * * * * * * * * * //

Foam::diameterModels::coalescenceModels::CUDALiaoCoalescenceSupportModel::
~CUDALiaoCoalescenceSupportModel() = default;

// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //


OFMemoryManagedField
Foam::diameterModels::coalescenceModels::CUDALiaoCoalescenceSupportModel::
getKolmogorovLengthScale(int domainSize)
{
    pimpl->model_.kolmogorovLengthScale_.resize(domainSize);
    return pimpl->model_.kolmogorovLengthScale_;
}


OFMemoryManagedField
Foam::diameterModels::coalescenceModels::CUDALiaoCoalescenceSupportModel::
getShearStrainRate(int domainSize)
{
    pimpl->model_.shearStrainRate_.resize(domainSize);
    return pimpl->model_.shearStrainRate_;
}


OFMemoryManagedField
Foam::diameterModels::coalescenceModels::CUDALiaoCoalescenceSupportModel::
getEddyStrainRate(int domainSize)
{
    pimpl->model_.eddyStrainRate_.resize(domainSize);
    return pimpl->model_.eddyStrainRate_;
}


OFMemoryManagedField
Foam::diameterModels::coalescenceModels::CUDALiaoCoalescenceSupportModel::
getUTerminal(int numberOfGroups)
{
    pimpl->model_.uTerminal_.resize(numberOfGroups);
    return pimpl->model_.uTerminal_;
}


OFMemoryManagedField
Foam::diameterModels::coalescenceModels::CUDALiaoCoalescenceSupportModel::
getCd(int numberOfGroups)
{
    pimpl->model_.Cd_.resize(numberOfGroups);
    return pimpl->model_.Cd_;
}


OFMemoryManagedField
Foam::diameterModels::coalescenceModels::CUDALiaoCoalescenceSupportModel::
getdCrit(int domainSize)
{
    pimpl->model_.dCrit_.resize(domainSize);
    return pimpl->model_.dCrit_;
}

OFMemoryManagedField
Foam::diameterModels::coalescenceModels::CUDALiaoCoalescenceSupportModel::
getCPack(int domainSize)
{
    pimpl->model_.CPack_.resize(domainSize);
    return pimpl->model_.CPack_;
}


void
Foam::diameterModels::coalescenceModels::CUDALiaoCoalescenceSupportModel::
setModelVariables
(
    double PMax_,
    double AH_,
    double CEff_,
    double CTurb_,
    double CBuoy_,
    double CShear_,
    double CEddy_,
    double CWake_,
    bool turbulence_,
    bool buoyancy_,
    bool laminarShear_,
    bool eddyCapture_,
    bool wakeEntrainment_,
    double CPackMax_
)
{
    pimpl->model_.PMax_ = PMax_;
    pimpl->model_.AH_ = AH_;
    pimpl->model_.CEff_ = CEff_;
    pimpl->model_.CTurb_ = CTurb_;
    pimpl->model_.CBuoy_ = CBuoy_;
    pimpl->model_.CShear_ = CShear_;
    pimpl->model_.CEddy_ = CEddy_;
    pimpl->model_.CWake_ = CWake_;
    pimpl->model_.turbulence_ = turbulence_;
    pimpl->model_.buoyancy_ = buoyancy_;
    pimpl->model_.laminarShear_ = laminarShear_;
    pimpl->model_.eddyCapture_ = eddyCapture_;
    pimpl->model_.wakeEntrainment_ = wakeEntrainment_;
    pimpl->model_.CPackMax_ = CPackMax_;
}


void
Foam::diameterModels::coalescenceModels::CUDALiaoCoalescenceSupportModel::
calculateCoalescenceRate
(
    CUDAPopulationBalance popBal,
    int currentDomainSize,
    const std::vector<std::pair<int, int>> &pairs
)
{
    pimpl->calculateRate
    (
        popBal,
        currentDomainSize,
        pairs
    );
}

OFMemoryManagedField
Foam::diameterModels::coalescenceModels::CUDALiaoCoalescenceSupportModel::
getCoalescenceRate
(
    int i,
    int j
)
{
    return pimpl->getRate
    (
        i,
        j
    );
}


// ************************************************************************* //
