/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

Class
    Foam::diameterModels::nucleationModels::DeaneStokes

Description
    The distribution from Deane and Stokes has been modified to be continuous,
    such that
    \f[
        f_{g}(D_{g}) = C*(D_{g})^{-3/2} if D_{g} < D_{crit}
    \f]
    \f[
        f_{g}(D_{g}) = (D_{g})^{-10/3} if D_{g} > D_{crit}
    \f]
    where
    \f[
        C = (D_{crit})^{-11/6}
    \f]

    Example of dictionary:

    \verbatim
    nucleationModels
    (
        DeaneStokes
        {
            velocityGroup   airD;
            interfaces      ((airD and airC));

            dCrit           0.001;
        }
    );
    \endverbatim

    References:
    \verbatim
        Deane, G. B, & Stokes, M. D. (2002).
        Scale dependence of bubble creation mechanisms in breaking waves.
        Nature, 418, 839-844.
    \endverbatim

Usage
    \table
        Property      | Description                     | Required | Default
        velocityGroup | Velocity group entrained        | yes      | None
        interfaces    | List of phaseInterface names    | yes      | None
        dCrit         | Critical diameter               | no       | 0.001
    \endtable

SourceFiles
    DeaneStokes.C

\*---------------------------------------------------------------------------*/

#ifndef DeaneStokes_H
#define DeaneStokes_H

#include "phaseTransferEntrainmentModel.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{
namespace diameterModels
{
namespace nucleationModels
{

/*---------------------------------------------------------------------------*\
                  Class DeaneStokes Declaration
\*---------------------------------------------------------------------------*/

class DeaneStokes
:
    public phaseTransferEntrainmentModel
{
    // Private Data

        //- Scale factor for Deane distribution
        scalar scaleFactor_;

        //- Critical diameter for Deane distribution
        const scalar dCrit_;


    // Private Member Functions

        //- Add to nucleationRate
        virtual dimensionedScalar distribution(const label i);


public:

    //- Runtime type information
    TypeName("DeaneStokes");

    // Constructor

        DeaneStokes
        (
            const populationBalanceModel& popBal,
            const dictionary& dict
        );


    //- Destructor
    virtual ~DeaneStokes()
    {}


    // Member Functions

        //- Precompute diameter independent expressions
        virtual void precompute();
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace nucleationModels
} // End namespace diameterModels
} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
