#!/bin/bash
#---------------------------------*- sh -*-------------------------------------
# ==  == ====== ====   ====    |
#                   \\     ||  | HZDR Multiphase Addon for OpenFOAM
# ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
# ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
# ==  == ====== ====   ==  ==  |
#------------------------------------------------------------------------------
# License
#     This file is part of the HZDR Multiphase Addon for OpenFOAM.
#
#     Copyright (C) 2021-2022 Helmholtz-Zentrum Dresden-Rossendorf
#
#     HZDR Multiphase Addon for OpenFOAM is free software:
#     you can redistribute it and/or modify it under the terms of the
#     GNU General Public License as published by the Free Software Foundation,
#     either version 3 of the License, or (at your option) any later version.
#
#     HZDR Multiphase Addon for OpenFOAM is distributed in the
#     hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#     implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#     See the GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with the HZDR Multiphase Addon for OpenFOAM.
#     If not, see <http://www.gnu.org/licenses/>.
#
# Description
#     Create VS Code setting files for OpenFOAM applications
#
#------------------------------------------------------------------------------

PROGNAME=$(basename "$0")
FORCE=no

usage() {
   while [ "$#" -ge 1 ]; do echo "$1"; shift; done
   cat<<USAGE

usage: $PROGNAME [OPTION]

options:
  -force        remove existing project files
  -help         this usage

* Create VS Code setting files for OpenFOAM applications.

USAGE
   exit 1
}

# parse options
while [ "$#" -gt 0 ]
do
   case "$1" in
   -h | -help)
      usage
      ;;
   -f | -force)
      FORCE=yes
      shift
      ;;
   -*)
      usage "Error: Invalid option '$1'"
      ;;
   *)
      break
      ;;
   esac
done

# Some user information
echo -e "\nCreating VS Code setting files\n"

# Remove old files if requested
if [ "$FORCE" == "yes" ]; then
    echo "  Removing old setting files"
    rm -f .vscode/tasks.json .vscode/settings.json .vscode/c_cpp_properties.json
fi

# Create user configuration file to set default extensions for docker container
folder=~/.config/Code/User
file=~/.config/Code/User/settings.json
# spell-checker: disable
content='{
    "remote.containers.defaultExtensions": [
        "ms-vscode.cpptools",        // C++
        "ms-vscode.cpptools-themes", // C++ highlighthing
        "usernamehw.errorlens",      // error handling assistent
        "zhikui.vscode-openfoam",    // OpenFOAM syntax highlighting
        "mammothb.gnuplot",          // Gnuplot syntax highlighting
        "gitlab.gitlab-workflow"     // GitLab extension
    ]
}'
# spell-checker: enable
if [ ! -d $folder ] || [ -f $file ]; then
    echo "  Skipping ~/.config/Code/User/settings.json"
    echo "   Folder not found or file already exists."
    echo "   Docker users should make sure that such a file exists containing:"
    printf '%s\n' "$content"
else
    echo "  Creating ~/.config/Code/User/settings.json"

cat<<EOF>>"$file"
$content
EOF

fi

# Create local VS Code folder
mkdir -p .vscode

# Write task definitions
file=.vscode/tasks.json
if [ -f $file ]; then
    echo "  Skipping $file"
else
    echo "  Creating $file"

cat<<EOF>>"$file"
{
  "version": "2.0.0",
  "tasks": [
    {
      "type": "shell",
      "label": "Allwmake",
      "command": "./Allwmake -j",
      "group": {
        "kind": "build",
        "isDefault": true
      },
      "problemMatcher": "\$gcc",
      "options": {
        "env": {
          // fix VS Code problem matching: use absolute path for compiling
          // overwrite Ctoo defined in $WM_PROJECT_DIR/wmake/rules/*/c++
          "Ctoo_VSCode":"\$(WM_SCHEDULER) \$(CC) \$(c++FLAGS) -c \$(CURDIR)/\$< -o \$@"
        }
      }
    },
    {
      "type": "shell",
      "label": "Allwclean",
      "command": "./Allwclean"
    }
  ]
}
EOF

fi

# Write folder settings
file=.vscode/settings.json
if [ -f $file ]; then
    echo "  Skipping $file"
else
    echo "  Creating $file"

cat<<EOF>>"$file"
{
    "extensions.ignoreRecommendations": true,
    "C_Cpp.errorSquiggles": "Disabled", // disable buggy red zigzag underlines
    "editor.rulers": [80], // editor settings
    "editor.insertSpaces": true,
    "editor.tabSize": 4,
    "files.eol": "\n",
    "files.trimTrailingWhitespace": true,
    "files.insertFinalNewline": false,
    "editor.detectIndentation": false,
    "gitlab.instanceUrl": "https://gitlab.hzdr.de", // set gitlab domain
    "files.associations": {
        "**/0/*": "OpenFOAM",
        "**/system/*": "OpenFOAM",
        "**/constant/*": "OpenFOAM",
        "*.C": "cpp",
        "*.H": "cpp",
        "log.*": "log",
        "createGraphs": "gnuplot"
    },
    "files.exclude": {
        "**/.vscode": true, // excluded to avoid problem matching there
        "**/platforms": true, // build directories
        "**/linux*Clang*": true, // build directories
        "**/linux*Gcc*": true, // build directories
        "**/linux*Icc*": true, // build directories
        "**/solaris*Gcc*": true, // build directories
        "**/SunOS*Gcc*": true, // build directories
        "**/lnInclude": true // lnInclude (symlink) directories
    },
    "files.watcherExclude": {
        "**/node_modules/*/**": true, // VS Code default
        "**/.hg/store/**": true, // VS Code default
        "**/.git/**": true,
        "**/.vscode": true, // same as in files.exclude above
        "**/platforms": true,
        "**/linux*Clang*": true,
        "**/linux*Gcc*": true,
        "**/linux*Icc*": true,
        "**/solaris*Gcc*": true,
        "**/SunOS*Gcc*": true,
        "**/lnInclude": true
    },
    "search.exclude": {
        "**/node_modules": true, // VS Code default
        "**/*.code-search": true, // VS Code default
        "**/.git/**": true,
        "**/.vscode": true, // same as in files.exclude above
        "**/platforms": true,
        "**/linux*Clang*": true,
        "**/linux*Gcc*": true,
        "**/linux*Icc*": true,
        "**/solaris*Gcc*": true,
        "**/SunOS*Gcc*": true,
        "**/lnInclude": true
    },
}
EOF

fi

# Write C++ settings
file=.vscode/c_cpp_properties.json
if [ -f $file ]; then
    echo "  Skipping $file"
else
    echo "  Creating $file"

cat<<EOF>>"$file"
{
    "configurations": [
        {
            "name": "OpenFOAM",
            "includePath": [
                "\${WM_PROJECT_DIR}/applications/**",
                "\${WM_PROJECT_DIR}/src/**",
                "\${WM_PROJECT_SITE}/\${WM_PROJECT_VERSION}/applications/**",
                "\${WM_PROJECT_SITE}/\${WM_PROJECT_VERSION}/src/**",
                "\${WM_PROJECT_USER_DIR}/src/**",
                "\${WM_PROJECT_USER_DIR}/applications/**"
            ]
        }
    ]
}
EOF

fi

echo -e "\nDone"

exit 0

#------------------------------------------------------------------------------
