#! /usr/bin/env python3
"""----------------------------------------------------------------------------
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is part of the HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2018-2022 Helmholtz-Zentrum Dresden-Rossendorf

    HZDR Multiphase Addon for OpenFOAM is free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    HZDR Multiphase Addon for OpenFOAM is distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

Description
    Creates a fourier transform of arbitrary data.

    Following column format is necessary:

    coordinate v1 v2 v3 v4 v5 v6 v7 v8 v9 v10 ...

    Abscissa data, distance (first column) or wave number, are always stored
    in 1D numpy arrays. The ordinate data, variable data, spectral energy or
    spectral energy density, are stored in an own ND numpy array, respectively,
    where N is the number of data sets to process.

    The individual processing steps are organised in functions. The functions
    detrend(), taper(), spectralEnergy() and showData() may take 1D or ND numpy
    arrays. Result data has identical structure.

----------------------------------------------------------------------------"""

import sys
import argparse
import numpy as np
import scipy
from matplotlib import pyplot as plt


#==============================================================================
#                                    MAIN
#==============================================================================

def main():
    """Control, which data to process and in what order the individual steps are
    done.
    """
    variables = globalvars

    for filename in args.files:
        print(f'Processing file: {filename}')

        # Read data
        data = read_data(filename)

        data = make_num_even(data)

        coordinate = data[:, 0]

        # Select and extract data
        if not varsavail:
            variables = create_vars_list(data)

        # Reorganise data
        velocity = extract_columns(data, variables)

        # Prepare data
        if args.no_window:
            # Subtract linear fit
            detrend(velocity)
            # Multiply window function to make signal periodic
            taper(velocity)

        # Fourier transform
        energy = spectral_energy(velocity)
        wavenumber = wave_numbers(coordinate)

        if args.no_smoothing:
            # Smooth, filter width is ~0.5% the length of the data set
            energy = smooth(energy, 5e-3*energy.shape[0])

        if args.normalise:
            # Normalise spectrum
            energy = normalise(energy)

        # Write output file
        outfile = filename + '_spectrum'
        varsheader = create_header(variables)
        write_file(outfile, wavenumber, energy, header=varsheader)

        if args.plot:
            show_data(wavenumber, energy, xlabel='k', ylabel='E',
                      title='Spectrum', log=True, id_vars=variables)

    return 0


#==============================================================================
#                                  NORMALISE
#==============================================================================

def normalise(ydataarg):
    """Normalise spectrum with spectral energy at smallest wave number greater
    zero.
    """
    if args.verbose:
        print('Normalising')

    if len(ydataarg.shape) == 1:
        ydataarg /= ydataarg[1]
    else:
        ycols = ydataarg.shape[1]
        for ycol in range(ycols):
            ydataarg[:, ycol] /= ydataarg[1, ycol]

    return ydataarg


#==============================================================================
#                                  SMOOTHING
#==============================================================================

def smooth(ydataarg, filterwidth):
    """Smooth data via convolution with box filter kernel. Takes filterwidth
    (may be a float, conversion to number of points in ensured in this function)
    """
    if args.verbose:
        print('Filtering')

    # Ensure integer number for filter width
    filterwidth = int(filterwidth)

    # Ensure odd number for filter width
    if filterwidth%2 == 0:
        filterwidth += 1

    # Filter kernel, box filter in this case. The integral is one.
    kernel = np.ones(filterwidth)/filterwidth

    # Filter operation - convolution
    if len(ydataarg.shape) == 1:
        ydataarg = np.convolve(ydataarg, kernel, mode='same')
    else:
        ycols = ydataarg.shape[1]
        for ycol in range(ycols):
            ydataarg[:, ycol] = np.convolve(ydataarg[:, ycol], kernel,
                                            mode='same')

    if args.verbose:
        print(f'Filter width: {filterwidth}')

    return ydataarg


#==============================================================================
#                                WAVE NUMBERS
#==============================================================================

def wave_numbers(tdata):
    """Calculate frequencies. The number matches half of the number of the time
    samples, ranging from 1/Period to 0.5*num/Period. The latter is equal to the
    Nyquist frequency. Only for equally spaced data.
    """
    if args.verbose:
        print('Calculate frequencies')

    # Round off time data to correct rounding errors from single graphs
    rtdata = np.round(tdata, 8)

    # Overall length
    period = rtdata[-1] - rtdata[0]

    if args.verbose:
        print(f'  Over all sample line length: {period}')

    # Wave number range
    return np.arange(rtdata.shape[0]/2)/period


#==============================================================================
#                               SPECTRAL ENERGY
#==============================================================================

def spectral_energy(ydataarg):
    """Do the fast Fourier transform. Its results are arrays of complex numbers.
    The spectral energy is calculated via the magnitude of those complex arrays.
    As the fft results are symmetric to the Nyquist frequency, the results are
    collapsed to the frequencies lower than the Nyquist frequency
    (f_Nyquist = 0.5*f_max).
    """
    if args.verbose:
        print('Fourier transform')

    num = ydataarg.shape[0]

    # Fourier transform
    if len(ydataarg.shape) == 1:
        ydata = scipy.fft.fft(ydataarg)/num
    else:
        if args.verbose:
            print('  Data set 1')
        ydata = scipy.fft.fft(ydataarg[:, 0])/num

        ycols = ydataarg.shape[1]
        for ycol in range(1, ycols):
            if args.verbose:
                print('  Data set', ycol+1)
            tydata = scipy.fft.fft(ydataarg[:, ycol])/num

            # Save data to common array
            ydata = np.vstack((ydata.transpose(), tydata)).transpose()

    # Square norm
    ydata = np.real(ydata*ydata.conjugate())

    # Spectral energy, f < f_Nyquist
    if len(ydata.shape) == 1:
        edata = ydata[:num//2]
        edata[:-1] = 2.*edata[:-1]

    else:
        edata = 2*ydata[:num//2, :]
        edata[-1, :] = edata[-1, :]

    return edata


#==============================================================================
#                                  TAPERING
#==============================================================================

def taper(ydataarg):
    """Weighting via bell window. This smoothly alters the data at begin and
    end of the data series. In this way, both end points match. This smooth
    periodic transition is crucial to prevent parasitic contributions in phase
    space.
    """
    if args.verbose:
        print('Tapering via bell window')

    xdata = np.linspace(0, 1, ydataarg.shape[0])

    # Define weight function, bell window in this case
    weight = np.piecewise(
        xdata,
        [(xdata >= 0)&(xdata <= 0.1),
         (xdata > 0.1)&(xdata < 0.9),
         (xdata >= 0.9)&(xdata <= 1)],
        [lambda xdata: np.sin(5*np.pi*xdata)**2.,
         1,
         lambda xdata: np.sin(5*np.pi*xdata)**2.,
         0]
        )

    if len(ydataarg.shape) == 1:
        ydataarg *= weight

    else:
        ycols = ydataarg.shape[1]
        for ycol in range(ycols):
            ydataarg[:, ycol] *= weight

    return ydataarg


#==============================================================================
#                                  DE-TREND
#==============================================================================

def detrend(ydataarg):
    """Do a linear fit of the ydataarg(xdata) and subtract this linear part.
    This function is intended to prevent saw tooth modes in FFT.
    """
    if args.verbose:
        print('De-trending via linear fit')

    xdata = np.arange(ydataarg.shape[0])

    if len(ydataarg.shape) == 1:
        # Fit polynomial of degree 1 and build a easy-to-use function
        fit = np.polyfit(xdata, ydataarg, 1)
        fit_fn = np.poly1d(fit)
        if args.verbose:
            print(fit_fn)

        ydataarg -= fit_fn(xdata)

    else:
        ycols = ydataarg.shape[1]
        for ycol in range(ycols):
            # Fit polynomial of degree 1 and build a easy-to-use function
            fit = np.polyfit(xdata, ydataarg[:, ycol], 1)
            fit_fn = np.poly1d(fit)
            if args.verbose:
                print(fit_fn)

            ydataarg[:, ycol] -= fit_fn(xdata)

    return ydataarg


#==============================================================================
#                         MAKE NUMBER OF SAMPLES EVEN
#==============================================================================

def make_num_even(data):
    """Ensure that number of data points is even. Cut last one if necessary."""

    if data.shape[0]%2 == 0:
        if args.verbose:
            print('Number of samples is already even')
        return data

    if args.verbose:
        print('Number of samples is odd, deleting last line')
    return data[:-1, :]


#==============================================================================
#                            CREATE VARIABLES LIST
#==============================================================================

def create_vars_list(data):
    """Create a list of all available variables. Intended in case, user does not
    give a variable list argument.
    """
    # Get number of columns
    ncols = data.shape[1] - 1

    # Create list of variables and of columns
    return (np.arange(ncols) + 1).tolist()


#==============================================================================
#                               EXTRACT COLUMNS
#==============================================================================

def extract_columns(data, cids):
    """Extract the selected columns (list) from data.
    """
    if args.verbose:
        print('Extracting columns')

    try:
        # Extract the selected data
        data = data[:, cids]
    except IndexError:
        sys.exit(f'Error: Only {data.shape[1]} columns available.' +
                 'Please, adjust VARIABLE.')

    return data


#==============================================================================
#                                CREATE HEADER
#==============================================================================

def create_header(id_vars_list):
    """Return the header for the given variables list.
    """
    if args.verbose:
        print('Creating file header')

    # Save to string
    headerstr = '  Wave number'
    for i in id_vars_list:
        headerstr += f'         Var {i:3}'

    return headerstr


#==============================================================================
#                                  READ DATA
#==============================================================================

def read_data(fname):
    """Try to read the file fName and, if valid, return content as numpy array.
    Special care is taken, if parentheses occur in the data, which is the case
    for vector probes from the OpenFOAM 'probes' functionObject.
    """
    if args.verbose:
        print(f'Reading file: {fname}')

    # Get number of columns
    ncols = np.loadtxt(fname, dtype='str').shape[1]

    # Remove parentheses, apply to all available columns
    convs = {k: lambda s: np.float64(s.strip(b'()'))
            for k in range(0, ncols)}

    # Save data to numpy array
    data = np.loadtxt(fname, converters=convs)

    return data


#==============================================================================
#                                  WRITE FILE
#==============================================================================

def write_file(fname, xdata, ydataarg, header=''):
    """Save header and all data to file of given name.
    """
    if args.verbose:
        print(f'Writing file: {fname}')

    if len(ydataarg.shape) == 1:
        alldata = np.vstack((xdata, ydataarg)).transpose()
    else:
        ycols = ydataarg.shape[1]
        for dummy in range(ycols):
            alldata = np.vstack((xdata, ydataarg.transpose())).transpose()

    np.savetxt(fname, alldata, header=header, fmt='%.9e')


#==============================================================================
#                                  SHOW DATA
#==============================================================================

def show_data(xdata, ydataarg, id_vars, xlabel='', ylabel='', title='',
              show=True, log=False):
    """Directly shows a plot of given data. Intended for development purposes.
    If you want to plot multiple data sets, call the function with show=False
    every time except for the last call. In the latter also give the labels if
    required.
    """
    if args.verbose:
        print('Plotting data')

    # Only one single column for ydataarg
    if len(ydataarg.shape) == 1:
        labeli = f'Variable {id_vars[0] + 1}'
        if log:
            plt.loglog(xdata[1:], ydataarg[1:], label=labeli)

        else:
            plt.plot(xdata, ydataarg, label=labeli)

    # Multiple columns for ydataarg
    else:
        ycols = ydataarg.shape[1]

        for ycol in range(ycols):
            labeli = f'Variable {id_vars[ycol]}'

            if log:
                plt.loglog(xdata[1:], ydataarg[1:, ycol], label=labeli)

            else:
                plt.plot(xdata, ydataarg[:, ycol], label=labeli)

    if show:
        if title:
            plt.title(title)
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.legend(loc=3)
        plt.grid()
        plt.show()


#==============================================================================
#                                  CALL MAIN
#==============================================================================

if __name__ == '__main__':
    # Parse arguments
    parser = argparse.ArgumentParser(description='Creates a spectrum of the \
    resolved turbulence kinetic energy, based on velocity singleGraph data.')

    # Add positional arguments
    parser.add_argument('-files', nargs='+', type=str,
                        help='File(s), containing velocity singleGraph data')
    parser.add_argument('-variables', type=int, nargs='*',
                        default=argparse.SUPPRESS,
                        help='Index of variable to process; default: all')
    parser.add_argument('-no-smoothing', action='store_false',
                        help='Do not smooth results')
    parser.add_argument('-no-window', action='store_false', help='Do not use \
                        window weighting before fft(for periodic domains only)')
    parser.add_argument('-normalise', action='store_true',
                        help='Normalise spectral energy')
    parser.add_argument('-plot', action='store_true',
                        help='Directly plot result')
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='Increase output verbosity')

    # Save arguments
    args = parser.parse_args()

    # Check variables
    globalvars = []
    try:
        # Argument variable is provided by the user
        globalvars = args.variables
        varsavail = True

        for ivar in globalvars:
            if ivar < 0:
                parser.print_help()
                sys.exit(f'Error: VARIABLE index: {ivar} must be >= 0')
    except AttributeError:
        # No variable argument is provided by the user
        varsavail = False

    # Run main
    main()

#------------------------------------------------------------------------------
