#! /usr/bin/env python3
"""----------------------------------------------------------------------------
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is part of the HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2018-2022 Helmholtz-Zentrum Dresden-Rossendorf

    HZDR Multiphase Addon for OpenFOAM is free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    HZDR Multiphase Addon for OpenFOAM is distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

Description
    Creates an energy spectrum of the resolved kinetic energy, based on velocity
    probe data.

    This script is written for probes (functionObject) of velocity vector
    fields. The following column format is necessary:

    time (u1 v1 w1) (u2 v2 w2) (u3 v3 w3) ...

    Abscissa data, time (first column) or frequency, are always stored in 1D
    numpy arrays. The ordinate data, velocity components, magnitude, spectral
    energy or spectral energy density, are stored in an own ND numpy array,
    respectively, where N is the number of data sets to process.

    The individual processing steps are organised in functions. The functions
    detrend(), taper(), spectral_energy() and show_data() may take 1D or ND
    numpy arrays. Result data has identical structure.

----------------------------------------------------------------------------"""

import sys
import argparse
import numpy as np
import scipy
from matplotlib import pyplot as plt

#==============================================================================
#                                    MAIN
#==============================================================================

def main():
    """Control, which data to process and in what order the individual steps are
    done.
    """
    probes = globalprobes

    for filename in args.files:
        print(f'Processing file: {filename}')

        # Read data
        data = read_data(filename)

        # Select and extract data
        if not probesavail:
            probes = create_probes_list(data)

        cols = id_columns(probes, args.component)
        data = extract_columns(data, cols)
        probesheader = read_header(filename, probes)

        # Reorganise data
        data = make_num_even(data)
        time = data[:, 0]
        velocity = data[:, 1:]
        if args.component == 'mag':
            velocity = magnitude(velocity)

        # Prepare data
        detrend(velocity)
        taper(velocity)

        # Fourier transform
        energy = spectral_energy(velocity)
        frequency = frequencies(time)

        if args.no_smoothing:
            # Smooth, filter width is ~0.5% the length of the data set
            energy = smooth(energy, 5e-3*energy.shape[0])

        if args.normalise:
            # Normalise spectrum
            energy = normalise(energy)

        # Write output file
        outfile = f'{filename}_{args.component}_spectrum'
        write_file(outfile, frequency, energy, header=probesheader)

        if args.plot:
            show_data(frequency, energy, xlabel='f', ylabel='E',
                      title=f'Spectrum {args.component}', log=True,
                      id_probes=probes)

    return 0


#==============================================================================
#                                  NORMALISE
#==============================================================================

def normalise(ydataarg):
    """Normalise spectrum with spectral energy at smallest frequency greater
    zero.
    """
    if args.verbose:
        print('Normalising')

    if len(ydataarg.shape) == 1:
        ydataarg /= ydataarg[1]
    else:
        ycols = ydataarg.shape[1]
        for ycol in range(ycols):
            ydataarg[:, ycol] /= ydataarg[1, ycol]

    return ydataarg


#==============================================================================
#                                  SMOOTHING
#==============================================================================

def smooth(ydataarg, filterwidth):
    """Smooth data via convolution with box filter kernel. Takes filterwidth
    (may be a float, conversion to number of points in ensured in this function)
    """
    if args.verbose:
        print('Filtering')

    # Ensure integer number for filter width
    filterwidth = int(filterwidth)

    # Ensure odd number for filter width
    if filterwidth%2 == 0:
        filterwidth += 1

    # Filter kernel, box filter in this case. The integral is one.
    kernel = np.ones(filterwidth)/filterwidth

    # Filter operation - convolution
    if len(ydataarg.shape) == 1:
        ydataarg = np.convolve(ydataarg, kernel, mode='same')
    else:
        ycols = ydataarg.shape[1]
        for ycol in range(ycols):
            ydataarg[:, ycol] = np.convolve(ydataarg[:, ycol], kernel,
                                            mode='same')

    if args.verbose:
        print(f'  Filter width: {filterwidth}')

    return ydataarg


#==============================================================================
#                               FREQUENCIES
#==============================================================================

def frequencies(tdata):
    """Calculate frequencies. The number matches half of the number of the time
    samples, ranging from 1/Period to 0.5*num/Period. The latter is equal to the
    Nyquist frequency. Only for equally spaced data.
    """
    if args.verbose:
        print('Calculate frequencies')

    # Total time period
    period = tdata[-1] - tdata[0]
    if args.verbose:
        print(f'  Over all period of time: {period}')

    # Frequency range
    fdata = np.arange(tdata.shape[0]/2)/period

    return fdata


#==============================================================================
#                               SPECTRAL ENERGY
#==============================================================================

def spectral_energy(ydataarg):
    """Do the fast Fourier transform. Its results are arrays of complex numbers.
    The spectral energy is calculated via the magnitude of those complex arrays.
    As the fft results are symmetric to the Nyquist frequency, the results are
    collapsed to the frequencies lower than the Nyquist frequency
    (f_Nyquist = 0.5*f_max).
    """
    if args.verbose:
        print('Fourier transform')

    num = ydataarg.shape[0]

    # Fourier transform
    if len(ydataarg.shape) == 1:
        ydata = scipy.fft.fft(ydataarg)/num
    else:
        if args.verbose:
            print('  Data set 1')
        ydata = scipy.fft.fft(ydataarg[:, 0])/num

        ycols = ydataarg.shape[1]
        for ycol in range(1, ycols):
            if args.verbose:
                print('  Data set', ycol+1)
            tydata = scipy.fft.fft(ydataarg[:, ycol])/num

            # Save data to common array
            ydata = np.vstack((ydata.transpose(), tydata)).transpose()

    # Square norm
    ydata = np.real(ydata*ydata.conjugate())

    # Spectral energy, f < f_Nyquist
    if len(ydata.shape) == 1:
        edata = ydata[:num//2]
        edata[:-1] = 2.*edata[:-1]

    else:
        edata = 2*ydata[:num//2, :]
        edata[-1, :] = edata[-1, :]

    return edata


#==============================================================================
#                                  TAPERING
#==============================================================================

def taper(ydataarg):
    """Weighting via bell window. This smoothly alters the data at begin and
    end of the time series. In this way, both end points match. This smooth
    periodic transition is crucial to prevent parasitic contributions in phase
    space.
    """
    if args.verbose:
        print('Tapering via bell window')

    xdata = np.linspace(0, 1, ydataarg.shape[0])

    # Define weight function, bell window in this case
    weight = np.piecewise(
        xdata,
        [(xdata >= 0)&(xdata <= 0.1),
         (xdata > 0.1)&(xdata < 0.9),
         (xdata >= 0.9)&(xdata <= 1)],
        [lambda xdata: np.sin(5*np.pi*xdata)**2.,
         1,
         lambda xdata: np.sin(5*np.pi*xdata)**2.,
         0]
        )

    if len(ydataarg.shape) == 1:
        ydataarg *= weight

    else:
        ycols = ydataarg.shape[1]
        for ycol in range(ycols):
            ydataarg[:, ycol] *= weight

    return ydataarg


#==============================================================================
#                                  DE-TREND
#==============================================================================

def detrend(ydataarg):
    """Do a linear fit of the ydataarg(xdata) and subtract this linear part.
    This function is intended to prevent saw tooth modes in FFT.
    """
    if args.verbose:
        print('De-trending via linear fit')

    xdata = np.arange(ydataarg.shape[0])

    if len(ydataarg.shape) == 1:
        # Fit polynomial of degree 1 and build a easy-to-use function
        fit = np.polyfit(xdata, ydataarg, 1)
        fit_fn = np.poly1d(fit)
        if args.verbose:
            print(fit_fn)

        ydataarg -= fit_fn(xdata)

    else:
        ycols = ydataarg.shape[1]
        for ycol in range(ycols):
            # Fit polynomial of degree 1 and build a easy-to-use function
            fit = np.polyfit(xdata, ydataarg[:, ycol], 1)
            fit_fn = np.poly1d(fit)
            if args.verbose:
                print(fit_fn)

            ydataarg[:, ycol] -= fit_fn(xdata)

    return ydataarg


#==============================================================================
#                         MAKE NUMBER OF SAMPLES EVEN
#==============================================================================

def make_num_even(data):
    """Ensure that number of data points is even. Cut last one if necessary."""

    if data.shape[0]%2 == 0:
        if args.verbose:
            print('Number of samples is already even')
        return data

    if args.verbose:
        print('Number of samples is odd, deleting last line')
    return data[:-1, :]


#==============================================================================
#                                  MAGNITUDE
#==============================================================================

def magnitude(ydataarg):
    """Calculate geometric average for all three components (columns) of given
    probes.
    """
    if args.verbose:
        print('Calculate vector magnitudes from components')

    # Prevent wrong data format
    if len(ydataarg.shape) == 1:
        sys.exit('Error: got vector, need Nx(3*m) matrix')

    ycols = ydataarg.shape[1]
    if ycols%3 != 0:
        sys.exit(f'Error calculating magnitude:\
 got Nx{ydataarg.shape[1]} matrix, need Nx(3*m) matrix')

    nprobes = int(ycols/3)

    # Loop over probes
    magdata = np.array(()) # save time column
    for id_probe in range(nprobes):
        tmagdata = np.zeros(ydataarg.shape[0])

        # Loop over components
        for id_comp in range(3):
            ycol = 3*(id_probe - 1) + id_comp
            tmagdata += np.square(ydataarg[:, ycol])

        # Square root
        tmagdata = np.sqrt(tmagdata)

        # Save data to array
        if magdata.size == 0:
            # New array - first probe
            magdata = tmagdata.transpose()
        else:
            # Append to existing array - all other probes
            magdata = np.vstack((magdata.transpose(), tmagdata)).transpose()

    return magdata


#==============================================================================
#                                  COLUMN IDs
#==============================================================================

def id_columns(id_probes, comp):
    """Calculate the list of column indices according to vector component and
    list of probes.
    """
    if args.verbose:
        print(f'Probes to process: {id_probes}')

    # List of columns according to parser arguments
    if   comp == 'x':
        id_comp = 1
    elif comp == 'y':
        id_comp = 2
    elif comp == 'z':
        id_comp = 3
    else:
        id_cols = [0] # time column
        for id_probe in id_probes:
            for id_comp in range(3):
                # Append all velocity components of one probe data set
                id_cols.append(3*id_probe + id_comp + 1)

        #~ sys.exit('FOAM FATAL ERROR magnitude not implemented yet')
        if args.verbose:
            print(f'  Column indices: {id_cols}')
        return id_cols

    id_cols = [0]  # time column
    for id_probe in id_probes:
        id_cols.append(3*id_probe + id_comp)

    if args.verbose:
        print(f'  Column indices: {id_cols}')
    return id_cols


#==============================================================================
#                               CREATE PROBES LIST
#==============================================================================

def create_probes_list(data):
    """Create a list of all available probes. Intended in case, user does not
    give a probe list argument.
    """
    # Get number of columns
    ncols = data.shape[1]

    # Check number of columns
    if (ncols - 1)%3 != 0:
        sys.exit(f'Error: {ncols} columns found. Need 3m+1.')
    elif ncols <= 0:
        sys.exit('Error: Not enough columns found.')

    # Create list of probes and of columns
    return (np.arange(ncols//3)).tolist()


#==============================================================================
#                               EXTRACT COLUMNS
#==============================================================================

def extract_columns(data, cids):
    """Extract the selected columns (list) from data.
    """
    try:
        # Extract the selected data
        data = data[:, cids]
    except IndexError:
        sys.exit('Error: Only '+ str(data.shape[1])
                 + ' columns available (equals '
                 + str(int((data.shape[1] - 1)/3.))
                 + ' probes). Please, adjust PROBE and/or COMPONENT.')

    if args.verbose:
        print('  Extracting columns')

    return data


#==============================================================================
#                                 READ HEADER
#==============================================================================

def read_header(fname, id_probes_list):
    """Return the probe definition of the given probes list from the given file.
    """
    if args.verbose:
        print('Reading probe header info')

    # Get maximum index value of all probes
    id_probes = np.asarray(id_probes_list, dtype=int)
    max_id_probe = id_probes.max() + 1

    # Save all header lines
    allprobes = []
    with open(fname, 'r', encoding="utf8") as file:
        for i in range(max_id_probe):
            allprobes.append(file.readline())

    # Sort order and save to string
    headerstr = ''
    for i in id_probes_list:
        headerstr += allprobes[i].strip('# ')

    # Probes and frequency columns
    headerstr += '                 Probe'
    for idprobe in id_probes_list:
        headerstr += f'{idprobe:25}'
    headerstr += '\n             Frequency'

    return headerstr


#==============================================================================
#                                  READ DATA
#==============================================================================

def read_data(fname):
    """Try to read the file fName and, if valid, return content as numpy array.
    Special care is taken, if parentheses occur in the data, which is the case
    for vector probes from the OpenFOAM 'probes' functionObject.
    """
    if args.verbose:
        print(f'Reading file: {fname}')

    # Get number of columns
    ncols = np.loadtxt(fname, dtype='str').shape[1]

    # Remove parentheses, apply to all available columns
    convs = {k: lambda s: np.float64(s.strip(b'()'))
             for k in range(0, ncols)}

    # Save data to numpy array
    data = np.loadtxt(fname, converters=convs)

    return data


#==============================================================================
#                                  WRITE FILE
#==============================================================================

def write_file(fname, xdata, ydataarg, header=''):
    """Save header and all data to file of given name.
    """
    if args.verbose:
        print(f'Writing file: {fname}\n')

    if len(ydataarg.shape) == 1:
        alldata = np.vstack((xdata, ydataarg)).transpose()

    else:
        ycols = ydataarg.shape[1]
        for dummy in range(ycols):
            alldata = np.vstack((xdata, ydataarg.transpose())).transpose()

    np.savetxt(fname, alldata, header=header, fmt='%.10e')


#==============================================================================
#                                  SHOW DATA
#==============================================================================

def show_data(xdata, ydataarg, id_probes, xlabel='', ylabel='', title='',
              show=True, log=False):
    """Directly shows a plot of given data. Intended for development purposes.
    If you want to plot multiple data sets, call the function with show=False
    every time except for the last call. In the latter also give the labels if
    required.
    """
    if args.verbose:
        print('Plotting data')

    # Only one single column for ydataarg
    if len(ydataarg.shape) == 1:
        labeli = f'Probe {id_probes[0]}'
        if log:
            plt.loglog(xdata[1:], ydataarg[1:], label=labeli)

        else:
            plt.plot(xdata, ydataarg, label=labeli)

    # Multiple columns for ydataarg
    else:
        ycols = ydataarg.shape[1]

        for ycol in range(ycols):
            labeli = f'Probe {id_probes[ycol]}'

            if log:
                plt.loglog(xdata[1:], ydataarg[1:, ycol], label=labeli)

            else:
                plt.plot(xdata, ydataarg[:, ycol], label=labeli)

    if show:
        if title:
            plt.title(title)
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.legend(loc=3)
        plt.grid()
        plt.show()


#==============================================================================
#                                  CALL MAIN
#==============================================================================

if __name__ == '__main__':
    # Parse arguments
    parser = argparse.ArgumentParser(description='Creates a spectrum of the \
    resolved kinetic energy, based on velocity probe data.')

    # Add positional arguments
    parser.add_argument('-files', nargs='+', type=str,
                        help='File(s), containing velocity probe data')
    parser.add_argument('-probes', type=int, nargs='*',
                        default=argparse.SUPPRESS,
                        help='Index of probe(s) to process; default: all')
    parser.add_argument('-component', type=str, default='x', help='Velocity \
                        component to process: x, y, z or mag; default: x')
    parser.add_argument('-no-smoothing', action='store_false',
                        help='Do not smooth results')
    parser.add_argument('-normalise', action='store_true',
                        help='Normalise spectral energy')
    parser.add_argument('-plot', action='store_true',
                        help='Directly plot result')
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='Increase output verbosity')

    # Save arguments
    args = parser.parse_args()

    # Check spatial components
    if args.component not in ['x', 'y', 'z', 'mag']:
        parser.print_help()
        sys.exit(f'Error: COMPONENT argument: {args.component} '
                + 'must be x, y, z or mag.')

    # Check probes
    globalprobes = []
    try:
        # Argument probe is provided by the user
        globalprobes = args.probes
        probesavail = True

        for iprobe in globalprobes:
            if iprobe < 0:
                parser.print_help()
                sys.exit(f'Error: PROBE index: {iprobe} must be >= 0')
    except AttributeError:
        # No probe argument is provided by the user
        probesavail = False

    # Run main
    main()

#------------------------------------------------------------------------------
