#! /usr/bin/env python3
"""----------------------------------------------------------------------------
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is part of the HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2018-2022 Helmholtz-Zentrum Dresden-Rossendorf

    HZDR Multiphase Addon for OpenFOAM is free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    HZDR Multiphase Addon for OpenFOAM is distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

Description
    Reads data from a file and averages starting from a specified point of time
    until the end of the data set. Multiple files may be processed. Probes and
    vector components may be selected for processing.

    Following column format is necessary:

        time v1 v2 v3 v4 v5 v6 v7 v8 v9 v10 ...

----------------------------------------------------------------------------"""

import sys
import os
import copy
import argparse
import numpy as np

#==============================================================================
#                                    MAIN
#==============================================================================

def main():
    """Control, which data to process and in what order the individual steps are
    done.
    """
    variables = globalvariables

    for filename in args.file:
        print(f'Processing file {filename}')

        # Read data
        data = read_data(filename)

        # Select and extract data
        if not varsavail:
            variables = create_vars_list(data)

        time = data[:, 0]
        jstart = find_index(time, args.startTime)

        data = extract_columns(data, variables)
        var = data[jstart:, :]

        for j, item in enumerate(variables):
            print(f'  Average value of variable {item} for '
                  + f't > {time[jstart]} s')

            print(f'  {np.sum(var[:, j])/np.shape(var[:, j])[0]:0.9f}')

    return 0


#==============================================================================
#                                  FIND INDEX
#==============================================================================

def find_index(array, value):
    """Return index location of element, which is closest to given value.
    """
    return np.argmin(np.abs(array - float(value)))


#==============================================================================
#                            CREATE VARIABLES LIST
#==============================================================================

def create_vars_list(data):
    """Create a list of all available variables. Intended in case, user does not
    give a variable list argument.
    """
    # Get number of columns
    ncols = data.shape[1] - 1

    # Create list of variables and of columns
    return (np.arange(ncols) + 1).tolist()


#==============================================================================
#                               EXTRACT COLUMNS
#==============================================================================

def extract_columns(data, cids):
    """Extract the selected columns (list) from data.
    """
    if args.verbose:
        print('Extracting columns')

    try:
        # Extract the selected data
        data = data[:, cids]
    except IndexError:
        sys.exit(f'Error: Only {data.shape[1] - 1}'
                 + ' variables available. Please, adjust VARIABLE.')

    return data


#==============================================================================
#                                  READ DATA
#==============================================================================

def read_data(fname):
    """Try to read the file fName and, if valid, return content as numpy array.
    Special care is taken, if parentheses occur in the data, which is the case
    for vector probes from the OpenFOAM 'probes' functionObject.
    """
    if args.verbose:
        print(f'Reading file: {fname}')

    # Get number of columns
    ncols = np.loadtxt(fname, dtype='str', skiprows=args.skipRows).shape[1]

    # Remove parentheses, apply to all available columns
    convs = {k: lambda s: np.float64(s.strip(b'()'))
            for k in range(0, ncols)}

    # Save data to numpy array
    data = np.loadtxt(fname, converters=convs, skiprows=args.skipRows)

    return data


#==============================================================================
#                                  CALL MAIN
#==============================================================================

if __name__ == '__main__':
    # Parse arguments
    parser = argparse.ArgumentParser(description='Average over recent time\
    steps.')

    # Add positional arguments
    requirednamed = parser.add_argument_group('required arguments')
    requirednamed.add_argument('-file', nargs='+', type=str, required=True,
                               help='File(s) containing data, e.g. from \
                               phaseStructureProperties')
    parser.add_argument('-variable', type=int, nargs='*',
                        default=argparse.SUPPRESS,
                        help='Index of variable(s) to process; default: all')
    parser.add_argument('-startTime', type=float, default=0,
                        help='Time to start averaging from; default: 0')
    parser.add_argument('-skipRows', type=int, default=0,
                        help='Number of rows to skip at head of file; \
                        default: 0')
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='Increase output verbosity')

    # Save arguments
    args = parser.parse_args()

    # Check given file list
    if args.file is None:
        parser.print_help()
        sys.exit('\nError: empty file list.')

    filenames = copy.copy(args.file)
    for globalfilename in args.file:
        if not os.path.isfile(globalfilename):
            print(f'Not a file: {globalfilename}')
            filenames.remove(globalfilename)

    if not args.file:
        parser.print_help()
        sys.exit('\nError: no valid file')

    args.file = filenames

    # Check variables
    globalvariables = []
    try:
        # Argument variable is provided by the user
        globalvariables = args.variable
        varsavail = True

        for i in globalvariables:
            if i < 0:
                parser.print_help()
                sys.exit(f'\nError: VARIABLE index: {i} must be >= 0.')
    except AttributeError:
        # No variable argument is provided by the user
        varsavail = False

    # Run main
    main()

#------------------------------------------------------------------------------
