#! /usr/bin/env python3
"""----------------------------------------------------------------------------
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is part of the HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2021-2022 Helmholtz-Zentrum Dresden-Rossendorf

    HZDR Multiphase Addon for OpenFOAM is free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    HZDR Multiphase Addon for OpenFOAM is distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

Description
    Reads data from a file and calculates time derivative for time series data,
    e.g. probe data.

    The following column format is required:

        time v1 v2 v3 v4 v5 v6 v7 v8 v9 v10 ...

----------------------------------------------------------------------------"""

import sys
import os
import copy
import argparse
import numpy as np

#==============================================================================
#                                    MAIN
#==============================================================================

def main():
    """Control, which data to process and in what order the individual steps are
    done.
    """
    variables = globalvariables

    for filename in args.file:
        print(f'Processing file {filename}')

        # Read data
        data = read_data(filename)

        # Select and extract data
        if not varsavail:
            variables = create_vars_list(data)

        t = data[:, 0]

        istart = find_index(t, args.startTime)

        data = extract_columns(data, variables)
        var = data[istart:, :]
        t = t[istart:]

        # Calculate derivative
        time = 0.5*(t[1:] + t[:-1])
        derivative = \
            (var[1:] - var[:-1])/(t[1:, np.newaxis] - t[:-1, np.newaxis])

        # Write output file
        outfile = f'{filename}_derivative'
        varsheader = create_header(variables)
        write_file(outfile, time, derivative, header=varsheader)

    return 0


#==============================================================================
#                                  FIND INDEX
#==============================================================================

def find_index(array, value):
    """Return index location of element, which is closest to given value.
    """
    return np.argmin(np.abs(array - float(value)))


#==============================================================================
#                            CREATE VARIABLES LIST
#==============================================================================

def create_vars_list(data):
    """Create a list of all available variables. Intended in case, user does not
    give a variable list argument.
    """
    # Get number of columns
    ncols = data.shape[1] - 1

    # Create list of variables and of columns
    return (np.arange(ncols) + 1).tolist()


#==============================================================================
#                               EXTRACT COLUMNS
#==============================================================================

def extract_columns(data, cids):
    """Extract the selected columns (list) from data.
    """
    if args.verbose:
        print('  Extracting columns')

    try:
        # Extract the selected data
        data = data[:, cids]
    except IndexError:
        sys.exit(f'Error: Only {data.shape[1] - 1} variables available.\
 Please, adjust COLUMN.')

    return data


#==============================================================================
#                                CREATE HEADER
#==============================================================================

def create_header(id_vars_list):
    """Return the header for the given variables list.
    """
    if args.verbose:
        print('  Creating file header')

    # Save to string
    headerstr = '                  Time'
    for j in id_vars_list:
        headerstr += f'         Var {j:3}'

    return headerstr


#==============================================================================
#                                  READ DATA
#==============================================================================

def read_data(fname):
    """Try to read the file fName and, if valid, return content as numpy array.
    Special care is taken, if parentheses occur in the data, which is the case
    for vector probes from the OpenFOAM 'probes' functionObject.
    """
    if args.verbose:
        print(f'  Reading file: {fname}')

    # Get number of columns
    ncols = np.loadtxt(fname, dtype='str', skiprows=args.skipRows).shape[1]

    # Remove parentheses, apply to all available columns
    convs = {k: lambda s: np.float64(s.strip(b'()'))
            for k in range(0, ncols)}

    # Save data to numpy array
    data = np.loadtxt(fname, converters=convs, skiprows=args.skipRows)

    return data


#==============================================================================
#                                  WRITE FILE
#==============================================================================

def write_file(fname, xdata, ydataarg, header=''):
    """Save header and all data to file of given name.
    """
    if args.verbose:
        print(f'  Writing file {fname}')

    if len(ydataarg.shape) == 1:
        alldata = np.vstack((xdata, ydataarg)).transpose()
    else:
        ycols = ydataarg.shape[1]
        for dummy in range(ycols):
            alldata = np.vstack((xdata, ydataarg.transpose())).transpose()

    np.savetxt(fname, alldata, header=header, fmt='%.9e')


#==============================================================================
#                                  CALL MAIN
#==============================================================================

if __name__ == '__main__':
    # Parse arguments
    parser = argparse.ArgumentParser(description='Calculate time derivative\
        for time series data, e.g. probe data.')

    # Add positional arguments
    requirednamed = parser.add_argument_group('required arguments')
    parser.add_argument('-column', type=int, nargs='*',
                        default=argparse.SUPPRESS,
                        help='Index of column(s) to process; default: all')
    parser.add_argument('-startTime', type=float, default=0,
                        help='Time to start averaging from; default: 0')
    parser.add_argument('-skipRows', type=int, default=0,
                        help='Number of rows to skip at head of file; \
                        default: 0')
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='Increase output verbosity')
    requirednamed.add_argument('-file', nargs='+', type=str, required=True,
                               help='File(s) containing data, e.g. from \
                               phaseStructureProperties')

    # Save arguments
    args = parser.parse_args()

    # Check given file list
    if args.file is None:
        parser.print_help()
        sys.exit('\nError: empty file list.')

    filenames = copy.copy(args.file)
    for globalfilename in args.file:
        if not os.path.isfile(globalfilename):
            print(f'Not a file: {globalfilename}')
            filenames.remove(globalfilename)

    if not args.file:
        parser.print_help()
        sys.exit('\nError: no valid file.')

    args.file = filenames

    # Check variables
    globalvariables = []
    try:
        # Argument variable is provided by the user
        globalvariables = args.column
        varsavail = True

        for i in globalvariables:
            if i < 0:
                parser.print_help()
                sys.exit(f'\nError: COLUMN index: {i} must be >= 0.')

    except AttributeError:
        # No variable argument is provided by the user
        varsavail = False

    # Run main
    main()

#------------------------------------------------------------------------------
