#!/bin/bash
#---------------------------------*- sh -*-------------------------------------
# ==  == ====== ====   ====    |
#                   \\     ||  | HZDR Multiphase Addon for OpenFOAM
# ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
# ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
# ==  == ====== ====   ==  ==  |
#------------------------------------------------------------------------------
# License
#     This file is part of the HZDR Multiphase Addon for OpenFOAM.
#
#     Copyright (C) 2018-2022 Helmholtz-Zentrum Dresden-Rossendorf
#
#     HZDR Multiphase Addon for OpenFOAM is free software:
#     you can redistribute it and/or modify it under the terms of the
#     GNU General Public License as published by the Free Software Foundation,
#     either version 3 of the License, or (at your option) any later version.
#
#     HZDR Multiphase Addon for OpenFOAM is distributed in the
#     hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#     implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#     See the GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with the HZDR Multiphase Addon for OpenFOAM.
#     If not, see <http://www.gnu.org/licenses/>.
#
# Description
#    Script launches a OpenFOAM docker container from the FWDC docker registry.
#
#------------------------------------------------------------------------------

PROGNAME=$(basename "$0")

usage() {
   while [ "$#" -ge 1 ]; do echo "$1"; shift; done
   cat<<USAGE

usage: $PROGNAME [OPTIONS] IMAGE

options:
  -h, --help                        help
  -v, --volume SOURCE:TARGET        host directory to mount
      --no-pull                     disables automatic pull for the container
      --no-rm                       disables automatic remove of container on exit

* Launches the specified OpenFOAM Docker image. Run \`docker run --help\` for more options.

USAGE
   exit 1
}

#Check if we can use podman
if [ -x "$(command -v podman)" ]; then
    DOCKER_EXEC="podman"
elif [ -x "$(command -v docker)" ]; then
    DOCKER_EXEC="docker"
else
    usage "Found neither Podman nor Docker"
fi

#Check for at least one argument
[ "$#" -eq 0 ] && usage "No docker image specified."

#Check for ssh agent configuration
if [ -z "$SSH_AUTH_SOCK" ]; then
    eval "$(ssh-agent -s)"
fi

# X11 Authentication
# Create a Xauthority file for docker
XAUTH=/tmp/.docker.xauth
touch $XAUTH

# Generate XAUTHORITY file for docker user
xauth nlist "$DISPLAY" | sed -e 's/^..../ffff/' | xauth -f $XAUTH nmerge -

# Default options for docker run command
DOCKER_OPTIONS=(
    "--interactive"
    "--tty"
    "--rm"
    "--pull=always"
    "--device=/dev/dri:/dev/dri"
    "--volume=$XAUTH:$XAUTH:rw"
    "--env XAUTHORITY=$XAUTH"
    "--env DISPLAY=$DISPLAY"
    "--volume /tmp/.X11-unix:/tmp/.X11-unix"
    "--env LOCAL_USER_NAME=$USER"
    "--volume config:/root/.config"
    "--volume $(dirname "$SSH_AUTH_SOCK"):$(dirname "$SSH_AUTH_SOCK")"
    "--volume /etc/localtime:/etc/localtime:ro"
    "--env SSH_AUTH_SOCK=$SSH_AUTH_SOCK"
    )

# Options for docker run command from command line
CMD_OPTIONS=()

# Parse options
while [ "$#" -gt 0 ]
do
    case "$1" in
    -v | --volume)
        [ "$#" -ge 2 ] || usage "'$1' option requires an argument"
        SOURCE=${2%:*}
        TARGET=${2##*:}
        if [ -d "$SOURCE" ]; then
            DOCKER_OPTIONS+=("--volume $SOURCE:$TARGET")
        else
            usage "No directory exists: $SOURCE"
        fi
        shift 2
        ;;
    --no-pull)
        # Position of --pull='always' in array
        DOCKER_OPTIONS[3]="--pull=never"
        shift
        ;;
    --no-rm)
        # Position of --rm in array
        unset 'DOCKER_OPTIONS[2]'
        shift
        ;;
    --gpus)
        # Position of --device=/dev/dri:/dev/dri in array
        unset 'DOCKER_OPTIONS[4]'
        CMD_OPTIONS+=("$1" "$2")
        shift 2
        ;;
    -h | --help)
        usage
        ;;
    *)
        CMD_OPTIONS+=("$1")
        shift
        ;;
    esac
done

# Extract docker image
DOCKER_IMAGE=${CMD_OPTIONS[-1]}
unset 'CMD_OPTIONS[-1]'

# Create docker container
# shellcheck disable=SC2086,SC2048
CONTAINERID=$($DOCKER_EXEC create ${DOCKER_OPTIONS[*]} ${CMD_OPTIONS[*]} "$DOCKER_IMAGE")

# Get git configuration file
if [ -n "$CONTAINERID" ]
then
    if [ -f "$HOME/.gitconfig" ]
    then
        $DOCKER_EXEC cp "$HOME/.gitconfig" "$CONTAINERID:/root/.gitconfig"
    else
        echo "Warning: Git config not found in $HOME/.gitconfig. Is git installed?"
    fi

    $DOCKER_EXEC start -i "$CONTAINERID"
else
    usage "Unable to launch $DOCKER_IMAGE."
fi
