/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2018-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

Class
    Foam::LESModels::mixedStructuralModel

Description
    Abstract base class for linear combination of any structural LES model with
    another LES model of eddy-viscosity type. Sub-grid stress tensor and derived
    variables summed up.

SourceFiles
    mixedStructuralModel.C

\*---------------------------------------------------------------------------*/

#ifndef mixedStructuralModel_H
#define mixedStructuralModel_H

#include "typeInfo.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{
namespace LESModels
{

/*---------------------------------------------------------------------------*\
                    Class mixedStructuralModel Declaration
\*---------------------------------------------------------------------------*/

template
<
    class BasicMomentumTransportModel,
    template<class> class EddyViscosityMomentumTransportModel,
    template<class> class StructuralMomentumTransportModel
>
class mixedStructuralModel
:
    public StructuralMomentumTransportModel
    <
        EddyViscosityMomentumTransportModel<BasicMomentumTransportModel>
    >
{
    // Private Member Functions

        // Disallow default bitwise copy construct and assignment
        mixedStructuralModel(const mixedStructuralModel&);
        void operator=(const mixedStructuralModel&);

        //- Getter for model coefficient
        virtual dimensionedScalar Ck() const
        {
            return BasicEddyViscosity::Ck_;
        }


public:

    typedef EddyViscosityMomentumTransportModel<BasicMomentumTransportModel>
        BasicEddyViscosity;
    typedef StructuralMomentumTransportModel<BasicEddyViscosity>
        BasicStructural;

    typedef typename BasicMomentumTransportModel::alphaField alphaField;
    typedef typename BasicMomentumTransportModel::rhoField rhoField;

    //- Runtime type information
    TypeName("MixedStructuralModel");


    // Constructors

        //- Construct from components
        mixedStructuralModel
        (
            const alphaField& alpha,
            const rhoField& rho,
            const volVectorField& U,
            const surfaceScalarField& alphaRhoRhi,
            const surfaceScalarField& phi,
            const viscosity& viscosity,
            const word& type = typeName
        );


    //- Destructor
    virtual ~mixedStructuralModel()
    {}


    // Member Functions

        //- Return SGS kinetic energy of the eddy viscosity model only,
        //  necessary to get turbulence quantities of eddy viscosity model right
        virtual tmp<volScalarField> k() const
        {
            return BasicEddyViscosity::k();
        }

        //- Return the turbulence viscosity of the eddy viscosity model only,
        //  necessary to get turbulence quantities of eddy viscosity model right
        virtual tmp<volScalarField> nut() const
        {
            return BasicEddyViscosity::nut();
        }

        //- Return the turbulence viscosity on patch of the eddy viscosity model
        //  only, necessary to get turbulence quantities of eddy viscosity model
        //  right
        virtual tmp<scalarField> nut(const label patchi) const
        {
            return BasicEddyViscosity::nut(patchi);
        }

        //- Return the SGS turbulent dissipation
        virtual tmp<volScalarField> epsilon() const;

        //- Return the turbulence specific dissipation rate
        virtual tmp<volScalarField> omega() const;

        //- Return the Reynolds stress tensor
        virtual tmp<volSymmTensorField> sigma() const;

        //- Return the effective stress tensor
        virtual tmp<volSymmTensorField> devTau() const;

        //- Return the source term for the momentum equation
        virtual tmp<fvVectorMatrix> divDevTau(volVectorField& U) const;

        //- Return the source term for the momentum equation
        virtual tmp<fvVectorMatrix> divDevTau
        (
            const volScalarField& rho,
            volVectorField& U
        ) const;

        //- Correct Eddy-Viscosity and related properties
        virtual void correct();
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace LESModels
} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //
#ifdef NoRepository
    #include "mixedStructuralModel.C"
#endif

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
