/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2018-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "pureStructuralModel.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{
namespace LESModels
{

// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

template
<
    class  BasicMomentumTransportModel,
    template<class> class StructuralMomentumTransportModel
>
pureStructuralModel
<
    BasicMomentumTransportModel,
    StructuralMomentumTransportModel
>::pureStructuralModel
(
    const alphaField& alpha,
    const rhoField& rho,
    const volVectorField& U,
    const surfaceScalarField& alphaRhoPhi,
    const surfaceScalarField& phi,
    const viscosity& viscosity,
    const word& type
)
:
    BasicStructural
    (
        type,
        alpha,
        rho,
        U,
        alphaRhoPhi,
        phi,
        viscosity
    ),

    Ck_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "Ck",
            this->coeffDict_,
            0.094
        )
    )
{
    if (type == typeName)
    {
        this->printCoeffs(type);
    }
}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

template
<
    class  BasicMomentumTransportModel,
    template<class> class StructuralMomentumTransportModel
>
tmp<volSymmTensorField>
pureStructuralModel
<
    BasicMomentumTransportModel,
    StructuralMomentumTransportModel
>::devTau() const
{
    return
        BasicStructural::devTau()
      + BasicLinearViscousStress::devTau();
}


template
<
    class  BasicMomentumTransportModel,
    template<class> class StructuralMomentumTransportModel
>
tmp<fvVectorMatrix>
pureStructuralModel
<
    BasicMomentumTransportModel,
    StructuralMomentumTransportModel
>::divDevTau
(
    volVectorField& U
) const
{
    return
        BasicStructural::divDevTau(U)
      + BasicLinearViscousStress::divDevTau(U);
}


template
<
    class  BasicMomentumTransportModel,
    template<class> class StructuralMomentumTransportModel
>
tmp<fvVectorMatrix>
pureStructuralModel
<
    BasicMomentumTransportModel,
    StructuralMomentumTransportModel
>::divDevTau
(
    const volScalarField& rho,
    volVectorField& U
) const
{
    return
        BasicStructural::divDevTau(rho, U)
      + BasicLinearViscousStress::divDevTau(rho, U);
}


template
<
    class  BasicMomentumTransportModel,
    template<class> class StructuralMomentumTransportModel
>
void
pureStructuralModel
<
    BasicMomentumTransportModel,
    StructuralMomentumTransportModel
>::correct()
{
    BasicStructural::correct();

    if (momentumTransportModel::debug && this->runTime_.outputTime())
    {
        tmp<volSymmTensorField> tauSgs(BasicStructural::devTau());
        tauSgs.ref().rename
        (
            IOobject::groupName("tauSgs", this->alphaRhoPhi_.group())
        );
        Info<< "Writing " << tauSgs->name() << endl;
        tauSgs().write();

        tmp<volSymmTensorField> tauEff(this->devTau());
        tauEff.ref().rename
        (
            IOobject::groupName("tauEff", this->alphaRhoPhi_.group())
        );
        Info<< "Writing " << tauEff->name() << endl;
        tauEff().write();

        Info<< "Writing " << this->nuSgs_.name() << endl;
        this->nuSgs_.write();
    }
}


template
<
    class  BasicMomentumTransportModel,
    template<class> class StructuralMomentumTransportModel
>
bool
pureStructuralModel
<
    BasicMomentumTransportModel,
    StructuralMomentumTransportModel
>::read()
{
    if (BasicStructural::read())
    {
        Ck_.readIfPresent(this->coeffDict());

        return true;
    }
    else
    {
        return false;
    }
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace LESModels
} // End namespace Foam

// ************************************************************************* //
