/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2018-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

Class
    Foam::LESModels::Clark

Description
    General base class for Clark gradient models for all kinds of flows.

    Reference:
    \verbatim
        Clark, R.A., Ferziger, J. H., & Reynolds, W. C. (1979).
        Evaluation of subgrid-scale models using an accurately simulated
        turbulent flow.
        Journal of Fluid Mechanics, 91(1), 1-16.
    \endverbatim

    The modeled SGS term is formulated as follows:
    \f[
        B_{ij} = \overline{\alpha}\ \overline{\rho}_\alpha
            \frac{\overline{\Delta}^2}{12}
            \frac{\partial \overline{u}_i}{\partial x_k}
            \frac{\partial \overline{u}_j}{\partial x_k} ,
    \f]
    where the overbar denotes quantities, which are (implicitly) filtered in
    the context of large eddy simulations. The Delta denotes the (implicit)
    filter width.

    Depending on the specific context, the model is used in
    (e.g. compressible/incompressible, single/multi fluid), the phase volume
    fraction and/or the phase specific density might be set equal to one. In
    this way, the formula collapses to the formulation, which can be found in
    the literature.

    Such models are not strictly dissipative. Therefore they either have to be
    combined (mixed) with an eddy viscosity model of some form, or they have to
    be switched off in regions of anti-dissipative behaviour (see Kobayashi's
    methodology).

SourceFiles
    Clark.C

\*---------------------------------------------------------------------------*/

#ifndef Clark_H
#define Clark_H

#include "structuralModel.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{
namespace LESModels
{

/*---------------------------------------------------------------------------*\
                            Class Clark Declaration
\*---------------------------------------------------------------------------*/

template<class BasicMomentumTransportModel>
class Clark
:
    public structuralModel<BasicMomentumTransportModel>
{
public:

    typedef typename BasicMomentumTransportModel::alphaField alphaField;
    typedef typename BasicMomentumTransportModel::rhoField rhoField;


private:

    // Private Data

        //- Threshold value, below which gradient of U is clipped
        const dimensionedScalar thresholdGradU_;


    // Private Member Functions

        // Disallow default bitwise copy construct and assignment
        Clark(const Clark&);
        void operator=(const Clark&);


protected:

        //- Return the sub-grid stress tensor
        virtual tmp<volSymmTensorField> B
        (
            const rhoField& rho,
            const volVectorField& U
        ) const;


public:

    // Constructors

        //- Construct from components, inherit from another turbulence model
        Clark
        (
            const alphaField& alpha,
            const rhoField& rho,
            const volVectorField& U,
            const surfaceScalarField& alphaRhoRhi,
            const surfaceScalarField& phi,
            const viscosity& viscosity,
            const word& type
        );

        //- Construct from components, inherit from abstract model class
        Clark
        (
            const word& type,
            const alphaField& alpha,
            const rhoField& rho,
            const volVectorField& U,
            const surfaceScalarField& alphaRhoRhi,
            const surfaceScalarField& phi,
            const viscosity& viscosity
        );


    //- Destructor
    virtual ~Clark()
    {}
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace LESModels
} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#ifdef NoRepository
    #include "Clark.C"
#endif

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
