/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2019-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

Class
    Foam::LESModels::KleinStructuralModel

Description
    Klein's second modification of Kobayashi's regularised structural sub-grid
    scale model. This model implementation is referred to as KKK2 in the
    reference.

    Reference:
    \verbatim
        Klein, M., Ketterl, S., Engelmann, L., Kempf, A., Kobayashi, H. (2020).
        Regularized, parameter free scale similarity type models for Large Eddy
        Simulation.
        International Journal of Heat and Fluid Flow, 81, 108496.
    \endverbatim

    The modeled sub-grid stress tensor is modelled as follows:
    \f[
        \mathbf{B} = \mathbf{B}^\mathrm{struct} - 2 \nu_t \mathbf{D},
    \f]
    \f[
        \mathbf{D} =
          \frac{1}{2}
          \left(\nabla\cdot\mathbf{u} + (\nabla\cdot\mathbf{u})^T\right) .
    \f]
    with sub-grid viscosity
    \f[
        \nu_t
      = \frac{1}{2} \mathit{max}
        \left(
            \frac{\mathbf{B}^\mathrm{struct}:\mathbf{D}}{\mathbf{D}:\mathbf{D}},
            0
        \right).
    \f]

SourceFiles
    KleinStructuralModel.C

\*---------------------------------------------------------------------------*/

#ifndef KleinStructuralModel_H
#define KleinStructuralModel_H

#include "typeInfo.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{
namespace LESModels
{

/*---------------------------------------------------------------------------*\
                    Class KleinStructuralModel Declaration
\*---------------------------------------------------------------------------*/

template
<
    class BasicMomentumTransportModel,
    template<class> class StructuralMomentumTransportModel
>
class KleinStructuralModel
:
    public StructuralMomentumTransportModel<BasicMomentumTransportModel>
{
    // Private Data

        // Model constant for calculating nuSs
        dimensionedScalar Ck_;


    // Private Member Functions

        // Disallow default bitwise copy construct and assignment
        KleinStructuralModel(const KleinStructuralModel&);
        void operator=(const KleinStructuralModel&);

        //- Return model coefficient
        virtual dimensionedScalar Ck() const
        {
            return Ck_;
        }


public:

    typedef BasicMomentumTransportModel BasicLinearViscousStress;
    typedef StructuralMomentumTransportModel<BasicLinearViscousStress>
        BasicStructural;

    typedef typename BasicMomentumTransportModel::alphaField alphaField;
    typedef typename BasicMomentumTransportModel::rhoField rhoField;

    //- Runtime type information
    TypeName("KleinStructuralModel");


    // Constructors

        //- Construct from components
        KleinStructuralModel
        (
            const alphaField& alpha,
            const rhoField& rho,
            const volVectorField& U,
            const surfaceScalarField& alphaRhoRhi,
            const surfaceScalarField& phi,
            const viscosity& viscosity,
            const word& type = typeName
        );


    //- Destructor
    virtual ~KleinStructuralModel()
    {}


    // Member Functions

        //- Return the turbulence viscosity
        virtual tmp<volScalarField> nut() const;

        //- Return the turbulence viscosity on patch
        virtual tmp<scalarField> nut(const label patchi) const
        {
            return this->nut()().boundaryField()[patchi];
        }

        //- Return the effective stress tensor
        virtual tmp<volSymmTensorField> devTau() const;

        //- Return source term for momentum equation
        virtual tmp<fvVectorMatrix> divDevTau(volVectorField& U) const;

        //- Return source term for momentum equation
        virtual tmp<fvVectorMatrix> divDevTau
        (
            const volScalarField& rho,
            volVectorField& U
        ) const;

        //- Correct eddy-viscosity and related properties
        virtual void correct();

        //- Read model coefficient
        virtual bool read();
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace LESModels
} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //
#ifdef NoRepository
    #include "KleinStructuralModel.C"
#endif

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
