/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2018-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "NicoudSigma.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{
namespace LESModels
{

// * * * * * * * * * * * * Private Member Functions  * * * * * * * * * * * * //

template<class BasicMomentumTransportModel>
tmp<volSymmTensorField> NicoudSigma<BasicMomentumTransportModel>::G
(
    const tmp<volTensorField>& gradU
) const
{
    return symm(gradU() & gradU().T());
}


template<class BasicMomentumTransportModel>
tmp<volScalarField> NicoudSigma<BasicMomentumTransportModel>::Dsigma
(
    const tmp<volSymmTensorField>& tG
) const
{
    volVectorField sigmaSqr(eigenValues(tG()));

    volScalarField sigma3(sqrt(posPart(sigmaSqr.component(vector::X))));
    volScalarField sigma2(sqrt(posPart(sigmaSqr.component(vector::Y))));
    volScalarField sigma1(sqrt(posPart(sigmaSqr.component(vector::Z))));

    return tmp<volScalarField>
    (
        sigma3*(sigma1 - sigma2)*(sigma2 - sigma3)
       /max
        (
            sigmaSqr.component(vector::Z),
            dimensionedScalar(dimensionSet(0, 0, -2, 0, 0), small)
        )
    );
}


// * * * * * * * * * * * * Protected Member Functions  * * * * * * * * * * * //

template<class BasicMomentumTransportModel>
void NicoudSigma<BasicMomentumTransportModel>::correctNut()
{
    this->nut_ = sqr(Csigma_*this->delta())*Dsigma();
    this->nut_.correctBoundaryConditions();
}


// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

template<class BasicMomentumTransportModel>
NicoudSigma<BasicMomentumTransportModel>::NicoudSigma
(
    const alphaField& alpha,
    const rhoField& rho,
    const volVectorField& U,
    const surfaceScalarField& alphaRhoPhi,
    const surfaceScalarField& phi,
    const viscosity& viscosity,
    const word& type
)
:
    LESeddyViscosity<BasicMomentumTransportModel>
    (
        type,
        alpha,
        rho,
        U,
        alphaRhoPhi,
        phi,
        viscosity
    ),

    Csigma_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "Csigma",
            this->coeffDict_,
            1.5
        )
    )
{
    if (type == typeName)
    {
        this->printCoeffs(type);
    }
}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

template<class BasicMomentumTransportModel>
tmp<volScalarField> NicoudSigma<BasicMomentumTransportModel>::k() const
{
    return volScalarField::New
    (
        IOobject::groupName("k", this->alphaRhoPhi_.group()),
        sqr(1/this->Ck_*this->nut()/this->delta())
    );
}


template<class BasicMomentumTransportModel>
tmp<volScalarField> NicoudSigma<BasicMomentumTransportModel>::epsilon() const
{
    volSymmTensorField D(symm(fvc::grad(this->U())));

    return volScalarField::New
    (
        IOobject::groupName("epsilon", this->alphaRhoPhi_.group()),
        2*this->nut_*magSqr(D)
    );
}


template<class BasicMomentumTransportModel>
tmp<volScalarField> NicoudSigma<BasicMomentumTransportModel>::omega() const
{
    return volScalarField::New
    (
        IOobject::groupName("omega", this->alphaRhoPhi_.group()),
        epsilon()/(this->Ck_*k())
    );
}


template<class BasicMomentumTransportModel>
void NicoudSigma<BasicMomentumTransportModel>::correct()
{
    LESeddyViscosity<BasicMomentumTransportModel>::correct();
    correctNut();

    if (momentumTransportModel::debug && this->runTime_.outputTime())
    {
        tmp<volSymmTensorField> tauSgs
        (
            volSymmTensorField::New
            (
                IOobject::groupName("tauSgs", this->alphaRhoPhi_.group()),
               -this->alpha_*this->rho_*this->nut()
               *dev(twoSymm(fvc::grad(this->U_)))
            )
        );
        Info<< "Writing " << tauSgs->name() << endl;
        tauSgs().write();

        Info<< "Writing " << this->nut_.name() << endl;
        this->nut_.write();
    }
}


template<class BasicMomentumTransportModel>
bool NicoudSigma<BasicMomentumTransportModel>::read()
{
    if (LESeddyViscosity<BasicMomentumTransportModel>::read())
    {
        Csigma_.readIfPresent(this->coeffDict());

        return true;
    }
    else
    {
        return false;
    }
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace LESModels
} // End namespace Foam

// ************************************************************************* //
