/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2018-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

Class
    Foam::LESModels::scaleSimilarity

Description
    General base class for all scale similarity models for all kinds of flows.

    Reference:
    \verbatim
        Liu, S., Meneveau, C., & Katz, J. (1994).
        On the properties of similarity subgrid-scale models as deduced from
        measurements in a turbulent jet
        Journal of Fluid Mechanics, 275, 83-119.
    \endverbatim

    The modeled SGS term is formulated as follows:
    \f[
        B_{ij} =
            \widehat{\left(\overline{\alpha}\,\overline{\rho}_\alpha\,
                     \overline{u}_{\alpha,i}\,\overline{u}_{\alpha,j}\right)}
          - \widehat{\overline{\alpha}}\,
            \widehat{\overline{\rho}}_\alpha
            \widehat{\overline{u}}_{\alpha,i}\,
            \widehat{\overline{u}}_{\alpha,j} ,
    \f]
    where the overbar denotes quantities, which are (implicitly) filtered in
    the context of large eddy simulations. The hat denotes explicit filtering
    via a test filter with a width larger than the one of the implicit
    filtering.

    Depending on the specific context, the model is used in
    (e.g. compressible/incompressible, single/multi fluid), the phase volume
    fraction and/or the phase specific density might be set equal to one. In
    this way, the formula collapses to the formulation, which can be found in
    the literature.

    Such models are not strictly dissipative. Therefore they either have to be
    combined (mixed) with an eddy viscosity model of some form, or they have to
    be switched off in regions of anti-dissipative behaviour (see Kobayashi's
    methodology).

SourceFiles
    scaleSimilarity.C

\*---------------------------------------------------------------------------*/

#ifndef scaleSimilarity_H
#define scaleSimilarity_H

#include "simpleFilter.H"
#include "structuralModel.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{
namespace LESModels
{

/*---------------------------------------------------------------------------*\
                       Class scaleSimilarity Declaration
\*---------------------------------------------------------------------------*/

template<class BasicMomentumTransportModel>
class scaleSimilarity
:
    public structuralModel<BasicMomentumTransportModel>
{
public:

    typedef typename BasicMomentumTransportModel::alphaField alphaField;
    typedef typename BasicMomentumTransportModel::rhoField rhoField;


protected:

    // Protected Data

        autoPtr<LESfilter> filterPtr_;
        LESfilter& filter_;


private:

    // Private Data

        //- Threshold value, below which sub-grid stress tensor is clipped
        const dimensionedScalar threshold_;


    // Private Member Functions

        // Disallow default bitwise copy construct and assignment
        scaleSimilarity(const scaleSimilarity&);
        void operator=(const scaleSimilarity&);

        //- Wrapper filter function, decides according to template type
        template<class T>
        T filterWrapper(const T& field) const
        {
            return FilterWrapper(field);
        }

        //- Wrapper filter function, for volScalarFields
        //  basically filter_()
        const volScalarField FilterWrapper(const volScalarField& field) const
        {
            return filter_(field)();
        }

        //- Wrapper filter function for geometricOneFields
        //  returns the field itself
        const geometricOneField FilterWrapper
        (
            const geometricOneField& field
        ) const
        {
            return field;
        }


protected:

        //- Return the sub-grid stress tensor
        virtual tmp<volSymmTensorField> B
        (
            const rhoField& rho,
            const volVectorField& U
        ) const;


public:

    // Constructors

        //- Construct from components, inherit from another turbulence model
        scaleSimilarity
        (
            const alphaField& alpha,
            const rhoField& rho,
            const volVectorField& U,
            const surfaceScalarField& alphaRhoRhi,
            const surfaceScalarField& phi,
            const viscosity& viscosity,
            const word& type
        );

        //- Construct from components, inherit from abstract model class
        scaleSimilarity
        (
            const word& type,
            const alphaField& alpha,
            const rhoField& rho,
            const volVectorField& U,
            const surfaceScalarField& alphaRhoRhi,
            const surfaceScalarField& phi,
            const viscosity& viscosity
        );


    //- Destructor
    virtual ~scaleSimilarity()
    {}


    // Member Functions

        //- Read model coefficients if they have changed
        virtual bool read();
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace LESModels
} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#ifdef NoRepository
    #include "scaleSimilarity.C"
#endif

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
