/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2018-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

Class
    Foam::omegaMenterWallFunctionFvPatchScalarField

Description
    This boundary condition provides a wall function constraint on turbulence
    specific dissipation, omega.  The values are computed using:

        \f[
            \omega = sqrt(\omega_{vis}^2 + \omega_{log}^2)
        \f]

    where

    \vartable
        \omega_{vis} | omega in viscous region
        \omega_{log} | omega in logarithmic region
    \endvartable

    see eq.(15) of:
    \verbatim
        Menter, F., & Esch, T. (2001).
        Elements of Industrial Heat Transfer Prediction.
        16th Brazilian Congress of Mechanical Engineering (COBEM),
        Nov. 2001.
    \endverbatim

    Other than in the standard omegaWallFunction also the turbulent production
    term in the near wall cell is computed from the wall function as described
    in:
    \verbatim
        Rzehak, R., & Kriebitzsch, S. (2015).
        Multiphase CFD-simulation of bubbly pipe flow: A code comparison.
        International Journal of Multiphase Flow, 68, 135-152.
    \endverbatim
    using the compound velocity wall function of Menter (2001).

Usage
    \table
        Property     | Description             | Required    | Default value
        beta1        | Model coefficient       | no          | 0.075
    \endtable

    Example of the boundary condition specification:
    \verbatim
    <patchName>
    {
        type            omegaMenterWallFunction;
    }
    \endverbatim

See also
    Foam::omegaWallFunctionFvPatchScalarField

SourceFiles
    omegaMenterWallFunctionFvPatchScalarField.C

\*---------------------------------------------------------------------------*/

#ifndef omegaMenterWallFunctionFvPatchScalarField_H
#define omegaMenterWallFunctionFvPatchScalarField_H

#include "omegaWallFunctionFvPatchScalarField.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

class momentumTransportModel;

/*---------------------------------------------------------------------------*\
          Class omegaMenterWallFunctionFvPatchScalarField Declaration
\*---------------------------------------------------------------------------*/

class omegaMenterWallFunctionFvPatchScalarField
:
    public omegaWallFunctionFvPatchScalarField
{
protected:
    // Protected Member Functions

        //- Calculate the omega and G
        virtual void calculate
        (
            const momentumTransportModel& turbulence,
            const List<scalar>& cornerWeights,
            const fvPatch& patch,
            scalarField& G,
            scalarField& omega
        );

public:

    //- Runtime type information
    TypeName("omegaMenterWallFunction");


    // Constructors

        //- Construct from patch and internal field
        omegaMenterWallFunctionFvPatchScalarField
        (
            const fvPatch&,
            const DimensionedField<scalar, volMesh>&
        );

        //- Construct from patch, internal field and dictionary
        omegaMenterWallFunctionFvPatchScalarField
        (
            const fvPatch&,
            const DimensionedField<scalar, volMesh>&,
            const dictionary&
        );

        //- Construct by mapping given
        // omegaMenterWallFunctionFvPatchScalarField
        //  onto a new patch
        omegaMenterWallFunctionFvPatchScalarField
        (
            const omegaMenterWallFunctionFvPatchScalarField&,
            const fvPatch&,
            const DimensionedField<scalar, volMesh>&,
            const fvPatchFieldMapper&
        );

        //- Disallow copy without setting internal field reference
        omegaMenterWallFunctionFvPatchScalarField
        (
            const omegaMenterWallFunctionFvPatchScalarField&
        ) = delete;

        //- Copy constructor setting internal field reference
        omegaMenterWallFunctionFvPatchScalarField
        (
            const omegaMenterWallFunctionFvPatchScalarField&,
            const DimensionedField<scalar, volMesh>&
        );


    // Member Functions

        // I-O

            //- Write
            virtual void write(Ostream&) const;
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
