/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2018-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

Class
    Foam::functionObjects::HZDRCourantNo

Description
    Calculates the CourantNo in the same way, the original
    functionObject "CourantNo" does. Additionally to writing out at defined
    times, the CourantNo field is written out whenever the maximum CourantNo
    exceeds the criterion CourantNo.

    Example of function object specification:
    \verbatim
    ConditionalCourantNumber
    {
        type HZDRCourantNo;
        libs ("libHZDRfieldFunctionObjects.so");
        log             true;
        writeControl    none;
        ...
        field           "phi.air";
        critWrite       true;
        criterion       0.5;
    }
    \endverbatim

Usage
    \table
        Property     | Description                    | Required| Default value
        type         | type name: HZDRCourantNo       | yes     |
        field        | flux field: phi or phi.*       | no      | "phi"
        rho          | density field                  | no      | "rho"
        critWrite    | switch for conditional write   | no      | false
        criterion    | critical value for writing out | no      | 0.5
        writeControl | writeControl options           | no      |
        result       | Name of the resulting field    | no    | derived from phi
    \endtable

See also
    Foam::functionObjects::CourantNo
    Foam::functionObject::fieldExpression
    Foam::functionObject

SourceFiles
    HZDRCourantNo.C

\*---------------------------------------------------------------------------*/

#ifndef functionObjects_HZDRCourantNo_H
#define functionObjects_HZDRCourantNo_H

#include "fieldExpression.H"
#include "volFields.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{
namespace functionObjects
{

/*---------------------------------------------------------------------------*\
                        Class HZDRCourantNo Declaration
\*---------------------------------------------------------------------------*/

class HZDRCourantNo
:
    public fieldExpression
{
    // Private Data

        //- Name of density field (according to flux field)
        word rhoName_;

        //- Switch for autowrite according to Co criterion "critCo"
        bool critWrite_;

        //- CourantNo criterion, save conditionalCo if maximum is above critCo
        scalar critCo_;

        //- Dictionary for runTimeModification
        const dictionary *dict_;


    // Private Member Functions

        //- Divide the Courant number by rho if required
        tmp<volScalarField::Internal> byRho
        (
            const tmp<volScalarField::Internal>& Co
        ) const;

        //- Calculate the Courant number field and return true if successful
        virtual bool calc();


public:

    //- Runtime type information
    TypeName("HZDRCourantNo");


    // Constructors

        //- Construct from Time and dictionary
        HZDRCourantNo
        (
            const word& name,
            const Time&,
            const dictionary&
        );


    //- Destructor
    virtual ~HZDRCourantNo();


    // Member Functions

        //- Read the CourantNo data
        virtual bool read(const dictionary&);
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace functionObjects
} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
