/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2019-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf
    and Technische Universitaet Darmstadt
    and Technische Universitaet Muenchen

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

Class
    Foam::functionObjects::phaseStructureProperties

Description
    Calculates and outputs the centre of gravity, velocity, volume, surface area
    and sphericity of a phase and can be used to determine the rise velocity
    and shape of a gas structure in stagnant liquid for example.

    Surface area is calculated via continuous interface reconstruction,
    according to Kunkelmann (2011).

    The dimensionality of the case is determined automatically.

    Reference:
    \verbatim
        Kunkelmann, C. (2011).
        Numerical modeling and investigation of boiling phenomena.
        PhD Thesis, Technische Universit{\"a}t Darmstadt, 1-122.
    \endverbatim

    Example of function object specification:
    \verbatim
    bubble.
    {
        type            phaseStructureProperties;
        libs            ("libHZDRfieldFunctionObjects.so");
        writeControl    runTime;
        writeInterval   1;
        ...
        regionType      cellZone;
        cellZoneName    zone;
        phaseName       alpha.air;
        velocityName    U.air;
        threshold       0.5;
        calcSphericity  true;
    }
    \endverbatim

Usage
    \table
        Property      | Description                        | Required  | Default
        type          | type name: phaseStructureProperties| yes         |
        regionType    | Evaluate for 'cellZone' or 'all'   | yes         |
        cellZoneName  | Specify cellZone                   | if cell zone|
        phaseName     | Phase volume fraction              | yes         |
        velocityName  | Phase specific velocity            | no          | U
        threshold     | Value for definition of interface  | no          | 0.5
        calcSphericity| Switch to turn on/off Sphericity   | no          | true
    \endtable

See also
    Foam::functionObjects::fvMeshFunctionObject
    Foam::functionObjects::logFiles

SourceFiles
    phaseStructureProperties.C

\*---------------------------------------------------------------------------*/

#ifndef functionObjects_phaseStructureProperties_H
#define functionObjects_phaseStructureProperties_H

#include "fvMesh.H"
#include "fvMeshFunctionObject.H"
#include "logFiles.H"
#include "NamedEnum.H"
#include "dimensionedVector.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{
namespace functionObjects
{

/*---------------------------------------------------------------------------*\
                  Class phaseStructureProperties Declaration
\*---------------------------------------------------------------------------*/

class phaseStructureProperties
:
    public fvMeshFunctionObject,
    public logFiles
{
public:

    // Public data types

        //- region type enumeration
        enum regionTypes
        {
            stCellZone,
            stAll
        };

        //- region type names
        static const NamedEnum<regionTypes, 2> regionTypeNames_;


private:

    // Private Member Functions

        //- Disallow default bitwise copy construct
        phaseStructureProperties(const phaseStructureProperties&);

        //- Disallow default bitwise assignment
        void operator=(const phaseStructureProperties&);

        //- Set cells to evaluate based on a cell zone
        void setCellZoneCells();

        //- Interpolate cell centred values to node values by inverse distance
        tmp<scalarField> interpolateCellDataToPointData
        (
            const volScalarField&
        ) const;

        //- Calculate surface area of iso-surface of given threshold
        scalarField surfaceArea(const scalarField&, const scalar);

        //- Register a dimensioned<Type> as UniformDimensionedField<Type> with
        //  the given name
        template<class Type>
        void
        registerDimensionedObject(const word, const dimensioned<Type>&)const;


protected:

    // Protected data

        //- Construction dictionary
        const dictionary dict_;

        //- region type
        const regionTypes regionType_;

        //- Name of region (patch, zone, etc.)
        word regionName_;

        //- Global number of cells
        label nCells_;

        //- Local list of cell IDs
        labelList cellId_;

        //- Name of phase fraction field
        const word phaseName_;

        //- Name of phase velocity field
        const word velocityName_;

        //- Whether to calculate and write surface area and sphericity
        const Switch  calcSphericity_;

        //- Value for interface definition
        const scalar threshold_;

        //- Two dimensional domain
        bool twoDimensional_;

        //- For 2D cases: thickness of domain in 3rd direction
        dimensionedScalar thickness_;

        //- Centre of gravity
        dimensionedVector CG_;

        //- Velocity
        dimensionedVector velocity_;

        //- Volume of the phase region
        dimensionedScalar volume_;

        //- Surface area of the phase region
        dimensionedScalar area_;

        //- Sphericity
        dimensionedScalar sphericity_;


    // Protected Member Functions

        //- Initialise, e.g. cell addressing
        void initialise(const dictionary& dict);

        //- Combine fields from all processor domains into single field
        template<class Type>
        void combineFields(Field<Type>& field);

        //- Output file header information
        virtual void writeFileHeader(const label i);


public:

    //- Run-time type information
    TypeName("phaseStructureProperties");


    // Constructors

        //- Construct from name, time and dictionary
        phaseStructureProperties
        (
            const word& name,
            const Time& runTime,
            const dictionary& dict
        );


    //- Destructor
    virtual ~phaseStructureProperties();


    // Member Functions

        //- Return the reference to the construction dictionary
        inline const dictionary& dict() const;

        //- Helper function to return the reference to the mesh
        inline const fvMesh& mesh() const;

        //- Return the region type
        inline const regionTypes& regionType() const;

        //- Return the local list of cell IDs
        inline const labelList& cellId() const;

        //- Filter a field according to cellIds
        template<class Type>
        tmp<Field<Type>> filterField(const Field<Type>& field) const;

        //- Read from dictionary
        virtual bool read(const dictionary&);

        //- Return the list of fields required
        virtual wordList fields() const
        {
            return wordList::null();
        }

        //- Execute, currently does nothing
        virtual bool execute();

        //- Calculate and write
        virtual bool write();
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace functionObjects
} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#include "phaseStructurePropertiesI.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#ifdef NoRepository
    #include "phaseStructurePropertiesTemplates.C"
#endif

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
