/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is a derivative work of OpenFOAM and part of the
    HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2018-2022 OpenFOAM Foundation
    and Helmholtz-Zentrum Dresden-Rossendorf

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    OpenFOAM and the HZDR Multiphase Addon for OpenFOAM are distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM and the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

Class
    Foam::profiler

Description
    Class which can be used for profiling, i.e. measuring the computational time
    spent for a function and the number of calls to it. The result is printed
    upon destruction, i.e. at the end of the log file.

    Based on code developed by J. Kelling (FWCC) at the HZDR C++ Hackathon in
    January 2017.

Usage
    At the top of the source file:

    \verbatim
        #include "profiler.H"

        static Foam::autoPtr<Foam::profiler> profiling
        (
            new Foam::profiler("function")
        );
    \endverbatim

    Before and after the portion of code to be analysed:

    \verbatim
        profiling->proceed();

        function();

        profiling->pause();
    \endverbatim

    In Make/options of the corresponding application or library:

    \verbatim
        EXE_INC = \
            -I../../../../../src/performance/lnInclude
    \endverbatim

\*---------------------------------------------------------------------------*/

#ifndef profiler_H
#define profiler_H

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#include <chrono>

namespace Foam
{

/*---------------------------------------------------------------------------*\
                          Class profiler Declaration
\*---------------------------------------------------------------------------*/

class profiler
{
    //- Private Data

        //- Name of function to be analysed
        word name_;

        //- Start time in microseconds
        std::chrono::time_point<std::chrono::steady_clock> start_;

        //- Accumulated time in microseconds
        std::chrono::microseconds accumulatedTime_;

        //- Accumulated number of calls
        label numberOfCalls_;


public:

    //- Constructor

        //- Construct with name of function to be analysed
        profiler(const word& name)
        :
            name_(name),
            accumulatedTime_(0),
            numberOfCalls_(0)
        {}


    //- Destructor
    ~profiler()
    {
        Info<< "Required time for " << name_ << " = "
            << accumulatedTime_.count() << " µs, number of calls = "
            << numberOfCalls_ << endl;
    }


    // Member Functions

        //- Continue counting
        void proceed()
        {
            ++numberOfCalls_;

            start_ = std::chrono::steady_clock::now();
        }

        //- Pause counting
        void pause()
        {
            auto diff = std::chrono::steady_clock::now() - start_;

            accumulatedTime_ +=
                std::chrono::duration_cast
                <
                    std::chrono::microseconds
                >(diff);
        }
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
