#! /usr/bin/env python3
"""----------------------------------------------------------------------------
  ==  == ====== ====   ====    |
                    \\     ||  | HZDR Multiphase Addon for OpenFOAM
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is part of the HZDR Multiphase Addon for OpenFOAM.

    Copyright (C) 2020-2022 Helmholtz-Zentrum Dresden-Rossendorf

    HZDR Multiphase Addon for OpenFOAM is free software:
    you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    HZDR Multiphase Addon for OpenFOAM is distributed in the
    hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with the HZDR Multiphase Addon for OpenFOAM.
    If not, see <http://www.gnu.org/licenses/>.

Description
    Extracts the data from testLoopReport into a new file used for plotting.

----------------------------------------------------------------------------"""

from collections import defaultdict
import os


class BenchmarkGraphGenerator:
    """
    Class to process testLoopReport
    """
    def __init__(self, report_path):
        self.report_path = report_path
        self.cases = []
        self.__parse_file()

    def __construct_short_alias(self, full_name):
        """
        The full name of the case is shortened for display purposes
        """
        split_full_name = full_name.split('/')
        alias = split_full_name[-4:-1]
        alias = '-'.join([a[0].upper() + a[1:] for a in alias])
        shortened_alias = ''
        last_upper = 0
        for i, letter in enumerate(alias):
            if letter == '-':
                shortened_alias += '-'
                continue
            if letter.isupper():
                last_upper = i
            if i-last_upper <= 3:
                shortened_alias += letter
        return shortened_alias

    def __name_formater(self, case):
        """
        Extract alias, case type (cuda, non-cuda) and execution time.
        """
        full_name = case[0]
        execution_time = case[1]
        split_full_name = full_name.split('/')
        shortened_alias = self.__construct_short_alias(full_name)
        case_type = split_full_name[-1]
        return shortened_alias, case_type, execution_time

    def __parse_file(self):
        """
        Load file
        """
        with open(self.report_path, encoding="utf8") as file:
            for line in file:
                line = line.strip()
                if 'completed in' not in line:
                    continue
                split_line = line.split(':')
                execution_time = float(split_line[1].strip().split(' ')[2])
                self.cases.append((split_line[0], execution_time))

    def get_cases(self):
        """
        Return cases
        """
        return self.cases

    def write_formated_file(self, graph_file_path):
        """
        Write file
        """
        types = set()
        case_dict = defaultdict(lambda: defaultdict(lambda: 0))
        for case in self.cases:
            alias, case_type, time = self.__name_formater(case)
            types.add(case_type)
            case_dict[alias][case_type] = time

        sorted_types = sorted(list(types))

        # Check if all the necessary directories exist.
        # If not they are created before file is created
        if not os.path.exists(os.path.dirname(graph_file_path)):
            try:
                os.makedirs(os.path.dirname(graph_file_path))
            except OSError:
                pass

        # File creation
        with open(graph_file_path, "w", encoding="utf8") as f:
            f.writelines('Case\t'+'\t'.join(sorted_types) + '\n')

            # pylint: disable=consider-using-dict-items
            for case in case_dict.keys():
                write_str = case
                for case_type in sorted_types:
                    write_str += '\t' + str(case_dict[case][case_type])
                f.write(write_str + '\n')

if __name__ == "__main__":
    input_file = './../testLoopReport'
    output_file = './../postProcessing/totalExecutionTimes.txt'
    report_generator = BenchmarkGraphGenerator(input_file)
    report_generator.write_formated_file(output_file)

#------------------------------------------------------------------------------
